# -*- coding: utf-8 -*-
"""
usage:
    xdxsmi [<command>] [<args>...] [options]
    
commands:
   clock          Control and query clock information of GPUs
   temp           Control and query temperature information of GPUs
   power          Control and query power information of GPUs
   fan            Control and query fan information of GPUs
   thermal        Control and query thermal information of GPUs
   dvfs           Control and query dvfs infomation of GPUs
   list           List all GPUs with their pci info
   query          Query GPUs' attributes
   stats          show statistic information about GPUs
   out            Show gpu info by out of band manager way
   vpuinfo        show the vpu info of all or specified gpu
   vpuutil        show the vpu utilization of all or specified gpu
   query_gpu      show a part of gpu info for all or specified gpu

options:
    -f, --filename=file  redirect query output to a specified file
    -l, --loop=sec       set look up interval, whose unit is second
    -h, --help           print this help message
    -i, --index=val      target gpu to read/write attributes, start from 0
    --version


See 'xdxsmi <command> -h' for more information on a specific command.
Run 'xdxsmi' without any options or commands will print a general gpu status.
"""

from __future__ import print_function

# from typing import Dict
from sys import exit
import time
import sys
import traceback
import os
import atexit

try:
    from xdxsmi.api.smi import XdxSmi
    from xdxsmi.api.xdxgpu import XdxGpu
    from xdxsmi.api.out_of_band_gpu_info import XdxGpuInfo
    from xdxsmi.api.third_party_docopt import docopt
    from xdxsmi.api.pyxdxml import xdxml_shutdown
except ImportError:
    from os.path import abspath, join, dirname

    current_file = os.path.abspath(__file__)
    project_root = os.path.dirname(os.path.dirname(os.path.dirname(current_file)))
    sys.path.insert(0, project_root)
    
    from xdxsmi.api.smi import XdxSmi
    from xdxsmi.api.xdxgpu import XdxGpu
    from xdxsmi.api.out_of_band_gpu_info import XdxGpuInfo
    from xdxsmi.api.third_party_docopt import docopt
    from xdxsmi.api.pyxdxml import xdxml_shutdown

# (xdx_smi: XdxSmi, args: Dict)
def show_xdxgpu_status_table_header(xdx_smi, args):
    smi = xdx_smi.smi_ver
    ver = xdx_smi.vers
    xdxgpu_status_table_header = \
"""
+-----------------------------------------------------------------------------------------+
| XDXSMI {smi:16}               KMD Version {ver[2]:9}             UMD Version {ver[3]} |
+----------------------------------+-------------------------------+----------------------+
| GPU     Name        Persistence-M| Bus-Id                 Disp.A | Volatile Uncorr. ECC |
| Fan     Temp  Perf  Pwr:Usage/Cap|                  Memory-Usage |             GPU-Util |
|                   Vpu-Driver-Name|          Vpu-Encoder-Util(fps)| Vpu-Decoder-Util(fps)|
|                                  |                               |                      |
|==================================+===============================+======================|""".format(smi=smi, ver=ver)

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write(xdxgpu_status_table_header)
        fo.close()
    else:
        print(xdxgpu_status_table_header, end='')

# (xdxgpu: XdxGpu, args: Dict)
def show_xdxgpu_status_table_body(xdxgpu, args):
    idx = xdxgpu.user_index
    name = xdxgpu.name
    bsid = xdxgpu.pci_bdf
    dsp = xdxgpu.dsp_mode #connected or disconnected or N\A
    dsp_t = xdxgpu.dsp_type # HDMI or DP
    if xdxgpu.fan_spd=="0":        
        fan = 'N/A'
    else:
        fan = xdxgpu.fan_spd
    fan = str(fan) + "RPM"
    temp = xdxgpu.temp
    if sys.version_info[0] == 3:
        tmp_temp = ''
        for num in temp:
            if num.isdigit() == False:
                break
            tmp_temp += num
        temp = tmp_temp + str("°C")

    # processes in multi-card environments only exists on card 0
    used_GPU_memory = xdxgpu.used_GPU_memory if xdxgpu.user_index == 0 else xdxgpu.fb_used
    fb_total = xdxgpu.fb_total
    util_gpu = xdxgpu.util_gpu
    vpu_vendor = '{:>32}'.format(xdxgpu.vpu_vendor) if len(xdxgpu.vpu_vendor) <= 32 else xdxgpu.vpu_vendor[:29] + "..."
    #32
    enc_cmd_rev_fps = xdxgpu.enc_cmd_rev_fps
    dec_cmd_rev_fps = xdxgpu.dec_cmd_rev_fps

    NA = 'N/A'
    ellipsis = '...'
    space = ' '
    bsid_l = 14
    dsp_l = 10
    temp_l = 6
    dsp_t_l = 4
    xdxgpu_status_table_body = ""

    if (sys.version_info[0] == 2): # for python2
        temp_l += 1

    if (len(dsp) > 9): # dsp = 'disconnected'
        bsid_l = 12

    elif (len(dsp) < 9): # dsp = 'N\A'
        space_l = 6
        dsp_l = 4
        dsp_t_l += 8

    xdxgpu_status_table_body = \
"""
| {idx}       {name:6}{ellipsis:5} {NA:13}|{bsid:{bsid_l}} {dsp_t:{dsp_t_l}}{dsp:{dsp_l}}| {space:16} {NA} |
| {fan:7} {temp:{temp_l}}{NA:5} {NA:12} | {space:4} {used_GPU_memory:8} MiB / {fb_total:5} MiB | {space:8} {util_gpu: >11} |
| {vpu_vendor} | {enc_cmd_rev_fps:>29} | {dec_cmd_rev_fps:>20} |
|                                  |                               |                      |
+----------------------------------+-------------------------------+----------------------+\
""".format(
            idx = idx, name = name, ellipsis = ellipsis, NA = NA, bsid = bsid, bsid_l = bsid_l, dsp_l = dsp_l,
            dsp_t=dsp_t, dsp_t_l=dsp_t_l, dsp=dsp, space = space, fan = fan, temp = temp, temp_l=temp_l,
            used_GPU_memory=used_GPU_memory, fb_total=fb_total, util_gpu = util_gpu, vpu_vendor = vpu_vendor,
            enc_cmd_rev_fps = enc_cmd_rev_fps, dec_cmd_rev_fps = dec_cmd_rev_fps
          )

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n' + xdxgpu_status_table_body)
        fo.close()
    else:
        print(xdxgpu_status_table_body, end=' ')


# (xdx_smi: XdxSmi, args: Dict)
def show_xdxgpu_status_table_process_header(xdx_smi, args):
    xdxgpu_status_table_process_header = \
"""
+-----------------------------------------------------------------------------------------+
| top-5 Processes: [Wait to finish]                                                       |
|  GPU   GI   CI        PID   Type   Process name                              GPU Memory |
|        ID   ID                                                               Usage      |
|=========================================================================================|"""
    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n'+ xdxgpu_status_table_process_header)
        fo.close()
    else:
        print(xdxgpu_status_table_process_header),

# (xdxgpu: XdxGpu, args: Dict)
def show_xdxgpu_status_table_process_body(xdxgpu, args):
    NA = 'N/A'
    space = ' '
    blank =''

    process_number = xdxgpu.pronum + 1
    # default_body = f"""|   {xdxgpu.user_index: >2}{' ': <3}{'N/A': <5}{'N/A': <5}{'': <4}{'N/A': <8}{'N/A': <3}{'': <3}{'N/A': <11}{' ': <30}{'N/A': >8}{' ': <4}|"""
    default_body = \
"""|   {user_index: >2}{space: <3}{NA: <5}{NA: <5}{blank: <4}{NA: <8}{NA: <3}{blank: <3}{NA: <11}{space: <30}{NA: >8}{space: <4}|
""".format(user_index = xdxgpu.user_index, space = space, NA = NA, blank = blank)

    xdxgpu_status_table_process_body = default_body if process_number == 1 else ''

    # Count the number of processes with non-zero fb memory
    non_zero_process = 0
    for i in range(1, process_number):
        attr_name = "pro{}LMA".format(i)
        if(float(getattr(xdxgpu, attr_name)) != 0):
            non_zero_process += 1
    
    for i in range(1, process_number):
        # gpu_id = xdxgpu.user_index if xdxgpu.user_index != None else 'N/A'
        gpu_id = 0
        attr_name = "pro{}PID".format(i)
        pid = getattr(xdxgpu, attr_name) if getattr(xdxgpu, attr_name) != None else 'N/A'
        
        type = 'N/A'
        attr_name = "pro{}NAME".format(i)
        p_name = getattr(xdxgpu, attr_name) if getattr(xdxgpu, attr_name) != None else 'N/A'
        
        attr_name = "pro{}LMA".format(i)
        g_usage = getattr(xdxgpu, attr_name) if getattr(xdxgpu, attr_name) != None else 'N/A'

        if(float(g_usage) == 0):
            break

        line = \
"""|   {gpu_id: >2}{space: <3}{NA: <5}{NA: <5}{blank: <4}{pid: <8}{NA: <3}{blank: <3}{p_name: <25}{space: <9}{g_usage: >15}MiB{space: <1}|
""".format(gpu_id = gpu_id, space = space, NA = NA, blank = blank, pid = pid, p_name = p_name, g_usage = g_usage)
        if(non_zero_process != 1):
            line += ''
        xdxgpu_status_table_process_body += line
        non_zero_process -= 1

    # option scheme: seperate processes by gpu number, if so, do not use table foot
    # xdxgpu_status_table_process_body += "\n+-----------------------------------------------------------------------------------------+"
    # only show processes of gpu 0
    xdxgpu_status_table_process_body = default_body if xdxgpu.user_index != 0 else xdxgpu_status_table_process_body
    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write('\n' + xdxgpu_status_table_process_body + '\n')
        fo.close()
    else:
        print(xdxgpu_status_table_process_body, end='')

# (xdx_smi: XdxSmi, args: Dict)
def show_xdxgpu_status_table_foot(xdx_smi, args):
    xdxgpu_status_table_foot = "+-----------------------------------------------------------------------------------------+"

    if(args['--filename'] != None):
        fo = open(args['--filename'], "a")
        fo.write(xdxgpu_status_table_foot)
        fo.close()
    else:
        print(xdxgpu_status_table_foot)

# before split only supports 1 gpu
# def show_xdxgpu_status_table_foot(xdxgpu: XdxGpu, args: Dict):
#     xdxgpu_status_table_foot = \
# f"""\
# +-----------------------------------------------------------------------------------------+
# | Processes: [Wait to finish]                                                             |
# |  GPU   GI   CI        PID   Type   Process name                              GPU Memory |
# |        ID   ID                                                               Usage      |
# |=========================================================================================|\n"""

#     for i in range(1, xdxgpu.pronum + 1):
#         gpu_id = xdxgpu.user_index if xdxgpu.user_index != None else 'N/A'
#         pid = getattr(xdxgpu, f"pro{i}PID") if getattr(xdxgpu, f"pro{i}PID") != None else 'N/A'
#         type = 'N/A'
#         p_name = getattr(xdxgpu,f"pro{i}NAME") if getattr(xdxgpu,f"pro{i}NAME") != None else 'N/A'
#         g_usage = getattr(xdxgpu,f"pro{i}Total") if getattr(xdxgpu,f"pro{i}Total") != None else 'N/A'

#         line = \
# f"""|   {gpu_id: >2}{' ': <3}{'N/A': <5}{'N/A': <5}{'': <4}{pid: <8}{type: <3}{'': <3}{p_name: <11}{' ': <30}{g_usage: >8}MiB{' ': <1}|\n"""
#         xdxgpu_status_table_foot += line

#     xdxgpu_status_table_foot += "+-----------------------------------------------------------------------------------------+\n"

#     if(args['--filename'] != None):
#         fo = open(args['--filename'], "a")
#         fo.write('\n' + xdxgpu_status_table_foot + '\n')
#         fo.close()
#     else:
#         print(xdxgpu_status_table_foot)


if __name__ == '__main__': 
    
    args = docopt(__doc__, version="1.6", options_first=True)
    
    # for user, bus invisible
    try:
        # e.g. xdxsmi -l=1 clock get
        if(args['--loop'] != None):
            args['<args>'] += ['--loop=' + args['--loop']]

        # e.g. xdxsmi -f=a.txt clock get
        if(args['--filename'] != None):
            args['<args>'] += ['--filename=' + args['--filename']]

        if(args['--index'] != None):
            args['<args>'] += ['--index=' + args['--index']]

        if args['<command>'] == 'out':
            import xdxsmi_out_of_band
            # out command only supports '-h' and '--help' 
            if not(args['--filename'] == None and args['--help'] == False 
                    and args['--loop'] == None and args['--version'] == False 
                    and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
                print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                exit()
            argv = [args['<command>']] + args['<args>']
            args = docopt(xdxsmi_out_of_band.__doc__, argv=argv)

            xdx_smi = XdxSmi(True)
            # Execute the shutdown function to  when this programme exits
            if xdx_smi.is_init == 1:
                atexit.register(xdxml_shutdown)
            xdxsmi_out_of_band.handle(xdx_smi, args)
        else:
            xdx_smi =  XdxSmi()
            # Execute the shutdown function when this programme exits
            if xdx_smi.is_init == 1:
                atexit.register(xdxml_shutdown)
            # we move the '-f' '-l' before command to the tail of the processing logic
            # so args["<args>"] may not be None.
            if not args["<command>"]:
                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()

                # catch invalid value of '--loop', and print warning
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()    

                # catch invalid value of '--index'
                index = -1
                num_xdxgpus = len(xdx_smi.xdxgpus)
                try:
                    if(args['--index'] == None):
                        index = -1                    # default situation
                    elif(args['--index'] != None and int(args['--index']) >= 0
                        and int(args['--index']) < num_xdxgpus):
                        index = int(args['--index'])  # assign index
                    else:
                        raise ValueError("No devices were found")
                except:
                    print("No devices were found")
                    exit()

                try:
                    target_gpus = xdx_smi.xdxgpus
                    if(index >= 0):
                        target_gpus = [target_gpus[index]]
                    # The static information of the VPU can be obtained first.
                    # Moreover, the video team hasn't modified the UMD code yet, which will print redundant information and affect the display effect,
                    # especially when multiple cards are in use.
                    if not xdx_smi.isVpufunctionValid:
                        outStr = "Vpu query are not allowed because the xdxgpu driver is not installed correctly!\n"
                        if(args['--filename'] != None):
                            fo = open(args['--filename'], "a")
                            fo.write(outStr)
                            fo.close()
                        else:
                            print(outStr, end="")
                    else:
                        for xdxgpu in target_gpus:
                            xdxgpu.update_vpu_static_info()
                    while True:
                        show_xdxgpu_status_table_header(xdx_smi, args)
                        for xdxgpu in target_gpus:
                            if xdx_smi.isVpufunctionValid:  #vpu utilization depends on libva library
                                xdxgpu.update_vpu_ultilization_dynamic()
                            show_xdxgpu_status_table_body(xdxgpu, args)
                        # running processes get by ioctl are the whole processes
                        # and cannot disginguished by gpu, so here show process once
                        target_gpus = [target_gpus[0]]
                        show_xdxgpu_status_table_process_header(xdx_smi, args)
                        for xdxgpu in target_gpus:
                            # each addition run time of thie function, the memory of process python3 adds 0.001MiB
                            # till here, the process_memory has been computed, if update info again, the sum of the process fb memory is incorrect
                            # xdxgpu.update_dynamic_info()
                            show_xdxgpu_status_table_process_body(xdxgpu, args)
                        show_xdxgpu_status_table_foot(xdx_smi, args)
                        
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval) 
                        # update dynamic info if -l command is true
                        for xdxgpu in target_gpus:
                            xdxgpu.update_dynamic_info()
                except KeyboardInterrupt:
                    exit()
            argv = [args['<command>']] + args['<args>']
            
            # handle options
            if args['<command>'] == 'clock':
                import xdxsmi_clock
                args = docopt(xdxsmi_clock.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                
                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()   

                xdxsmi_clock.handle(xdx_smi, args) 
            
                try:
                    # loop_interval = float(args["--loop"])
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)

                        # xdx_smi =  XdxSmi()
                        xdxsmi_clock.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()   
            elif args['<command>'] == 'power':
                import xdxsmi_power
                args = docopt(xdxsmi_power.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()

                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()
                
                xdxsmi_power.handle(xdx_smi, args)

                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)
                        # xdx_smi =  XdxSmi()
                        xdxsmi_power.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()   
            elif args['<command>'] == 'temp':
                import xdxsmi_temp
                args = docopt(xdxsmi_temp.__doc__, argv=argv)
                
                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                
                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()

                xdxsmi_temp.handle(xdx_smi, args) 
                
                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)
                        # xdx_smi =  XdxSmi()
                        xdxsmi_temp.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit() 

            elif args['<command>'] == 'util':
                import xdxsmi_utility
                args = docopt(xdxsmi_utility.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                xdxsmi_utility.handle(xdx_smi, args)
            elif args['<command>'] == 'fan':
                import xdxsmi_fan
                args = docopt(xdxsmi_fan.__doc__, argv=argv)
                
                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                
                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()
                
                xdxsmi_fan.handle(xdx_smi, args)
                
                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)
                        # xdx_smi =  XdxSmi()
                        xdxsmi_fan.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()     
            elif args['<command>'] == 'list':
                import xdxsmi_list
                if not(args['--filename'] == None and args['--help'] == False 
                    and args['--loop'] == None and args['--version'] == False 
                    and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
                    print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    exit()
                
                # list command only supports '-h' and '--help'
                args = docopt(xdxsmi_list.__doc__, argv=argv)
                xdxsmi_list.handle(xdx_smi, args)
            elif args['<command>'] == 'set':
                import xdxsmi_set
                args = docopt(xdxsmi_set.__doc__, argv=argv)
                xdxsmi_set.handle(xdx_smi, args)
            elif args['<command>'] == 'query':
                import xdxsmi_query
                args = docopt(xdxsmi_query.__doc__, argv=argv)
                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()

                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()

                try:
                    while True:
                        # xdx_smi =  XdxSmi()
                        xdxsmi_query.handle(xdx_smi, args)

                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)
                except KeyboardInterrupt:
                    exit()        

            elif args['<command>'] == 'pmon':
                import xdxsmi_pmon
                args = docopt(xdxsmi_pmon.__doc__, argv=argv)
                xdxsmi_pmon.handle(xdx_smi, args)
            elif args['<command>'] == 'stats':
                import xdxsmi_stats
                args = docopt(xdxsmi_stats.__doc__, argv=argv)
                xdxsmi_stats.handle(xdx_smi, args)
            elif args['<command>'] == 'vpuinfo':
                if not xdx_smi.isVpufunctionValid:  # If there is an exception in the libva library, it will not be executed.
                    print("This function is invalid because the xdxgpu driver is not installed correctly!\n")
                else:
                    if(args['--filename'] != None):
                        try:
                            fo = open(args['--filename'], "a")
                            fo.close()
                        except:
                            print("failed to created such file: ", args['--filename'])
                            exit()
                    import xdxsmi_vpu_info
                    args = docopt(xdxsmi_vpu_info.__doc__, argv=argv)
                    xdxsmi_vpu_info.handle(xdx_smi, args)
            elif args['<command>'] == 'vpuutil':
                if not xdx_smi.isVpufunctionValid:  # If there is an exception in the libva library, it will not be executed.
                    print("This function is invalid because the xdxgpu driver is not installed correctly!\n")
                else:
                    if(args['--filename'] != None):
                        try:
                            fo = open(args['--filename'], "a")
                            fo.close()
                        except:
                            print("failed to created such file: ", args['--filename'])
                            exit()
                    import xdxsmi_vpu_utlz
                    args = docopt(xdxsmi_vpu_utlz.__doc__, argv=argv)
                    xdxsmi_vpu_utlz.handle(xdx_smi, args)
            elif args['<command>'] == 'thermal':
                import xdxsmi_thermal
                args = docopt(xdxsmi_thermal.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()

                # if the value of '--loop' is valid, loop regularly
                # otherwise, display warninig.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()
                
                xdxsmi_thermal.handle(xdx_smi, args)  
            
                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)

                        xdxsmi_thermal.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()
            elif args['<command>'] == 'dvfs':
                import xdxsmi_dvfs
                args = docopt(xdxsmi_dvfs.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                

                # if the value of '--loop' is valid, loop regularly
                # otherwise, display only once.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()   

                xdxsmi_dvfs.handle(xdx_smi, args)
            
                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)

                        xdxsmi_dvfs.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()
            
            elif args['<command>'] == 'query_gpu':
                import xdxsmi_query_gpu_info
                args = docopt(xdxsmi_query_gpu_info.__doc__, argv=argv)

                if(args['--filename'] != None):
                    try:
                        fo = open(args['--filename'], "a")
                        fo.close()
                    except:
                        print("failed to created such file: ", args['--filename'])
                        exit()
                

                # if the value of '--loop' is valid, loop regularly
                # otherwise, display only once.
                loop_interval = -1
                if(args['--loop'] != None):
                    try:
                        if(float(args['--loop']) > 0):
                            loop_interval = float(args["--loop"])
                        else:
                            raise ValueError("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                    except:
                        print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                        exit()   

                xdxsmi_query_gpu_info.handle(xdx_smi, args)
            
                try:
                    while True:
                        if(loop_interval < 0):
                            exit()
                        time.sleep(loop_interval)

                        xdxsmi_query_gpu_info.handle(xdx_smi, args)
                except KeyboardInterrupt:
                    exit()

            else:
                # same as the nvidia-smi prompt message 
                exit("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
    except Exception:
        with open('/tmp/smiError.log', 'w+') as f:
            f.write(traceback.format_exc())
        os.chmod('/tmp/smiError.log', 0o666)
        exit("An internal error has occured.")

    # for debug. Please annotate the above code and use below to get
    # the  error message
    '''
    # e.g. xdxsmi -l=1 clock get
    if(args['--loop'] != None and args['--loop'] != '-1'):
        args['<args>'] += ['--loop=' + args['--loop']]

    # e.g. xdxsmi -f=a.txt clock get
    if(args['--filename'] != None):
        args['<args>'] += ['--filename=' + args['--filename']]

    if(args['--index'] != None):
        args['<args>'] += ['--index=' + args['--index']]

    if args['<command>'] == 'out':
        import xdxsmi_out_of_band
        # out command only supports '-h' and '--help' 
        if not(args['--filename'] == None and args['--help'] == False 
               and args['--loop'] == '-1' and args['--version'] == False 
               and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
            print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
            exit()
        argv = [args['<command>']] + args['<args>']
        args = docopt(xdxsmi_out_of_band.__doc__, argv=argv)

        xdx_smi = XdxSmi(True)
        xdxsmi_out_of_band.handle(xdx_smi, args)
    else:
        xdx_smi =  XdxSmi()
        # we move the '-f' '-l' before command to the tail of the processing logic
        # so args["<args>"] may not be None.
        if not args["<command>"]:
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1     

            # catch invalid value of '--index'
            index = -1
            num_xdxgpus = len(xdx_smi.xdxgpus)
            try:
                if(args['--index'] == None):
                    index = -1                    # default situation
                elif(args['--index'] != None and int(args['--index']) >= 0
                    and int(args['--index']) < num_xdxgpus):
                    index = int(args['--index'])  # assign index
                else:
                    raise ValueError("No devices were found")
            except:
                print("No devices were found")
                exit()
        
            try:
                while True:
                    target_gpus = xdx_smi.xdxgpus
                    if(index >= 0):
                        target_gpus = [target_gpus[index]]

                    show_xdxgpu_status_table_header(xdx_smi, args)
                    for xdxgpu in target_gpus:
                        show_xdxgpu_status_table_body(xdxgpu, args)
                    # split table foot to three parts can be convenient for multi gpus
                    show_xdxgpu_status_table_process_header(xdx_smi, args)
                    
                    # running processes get by ioctl are the whole processes
                    # and cannot disginguished by gpu, so here show process once 
                    target_gpus = [target_gpus[0]]
                    for xdxgpu in target_gpus:
                        show_xdxgpu_status_table_process_body(xdxgpu, args)
                    show_xdxgpu_status_table_foot(xdx_smi, args)
                    
                    
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)                    
                    xdx_smi =  XdxSmi()
            except KeyboardInterrupt:
                exit()
        argv = [args['<command>']] + args['<args>']
        
        # handle options
        if args['<command>'] == 'clock':
            import xdxsmi_clock
            args = docopt(xdxsmi_clock.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            xdxsmi_clock.handle(xdx_smi, args)

            # if the value of '--loop' is valid, loop regularly
            # otherwise, display only once.
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1   
        
            try:
                # loop_interval = float(args["--loop"])
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)

                    xdx_smi =  XdxSmi()
                    xdxsmi_clock.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()   
        elif args['<command>'] == 'power':
            import xdxsmi_power
            args = docopt(xdxsmi_power.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            xdxsmi_power.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 

            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_power.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()   
        elif args['<command>'] == 'temp':
            import xdxsmi_temp
            args = docopt(xdxsmi_temp.__doc__, argv=argv)
            
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()

            xdxsmi_temp.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 
            
            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_temp.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit() 

        elif args['<command>'] == 'util':
            import xdxsmi_utility
            args = docopt(xdxsmi_utility.__doc__, argv=argv)

            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            xdxsmi_utility.handle(xdx_smi, args)
        elif args['<command>'] == 'fan':
            import xdxsmi_fan
            args = docopt(xdxsmi_fan.__doc__, argv=argv)
            
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            xdxsmi_fan.handle(xdx_smi, args)

            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 
            
            try:
                while True:
                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
                    xdx_smi =  XdxSmi()
                    xdxsmi_fan.handle(xdx_smi, args)
            except KeyboardInterrupt:
                exit()     
        elif args['<command>'] == 'list':
            import xdxsmi_list
            if not(args['--filename'] == None and args['--help'] == False 
                and args['--loop'] == '-1' and args['--version'] == False 
                and (args['<args>'] == [] or args['<args>'] == ['-h'] or args['<args>'] == ['--help'])):
                print("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
                exit()
            
            # list command only supports '-h' and '--help'
            args = docopt(xdxsmi_list.__doc__, argv=argv)
            xdxsmi_list.handle(xdx_smi, args)
        elif args['<command>'] == 'set':
            import xdxsmi_set
            args = docopt(xdxsmi_set.__doc__, argv=argv)
            xdxsmi_set.handle(xdx_smi, args)
        elif args['<command>'] == 'query':
            import xdxsmi_query
            args = docopt(xdxsmi_query.__doc__, argv=argv)
            if(args['--filename'] != None):
                try:
                    fo = open(args['--filename'], "a")
                    fo.close()
                except:
                    print("failed to created such file: ", args['--filename'])
                    exit()
            
            # catch invalid value of '--loop'
            loop_interval = -1
            try:
                if(float(args['--loop']) > 0):
                    loop_interval = float(args["--loop"])
            except:
                loop_interval = -1 

            try:
                while True:
                    xdx_smi =  XdxSmi()
                    xdxsmi_query.handle(xdx_smi, args)

                    if(loop_interval < 0):
                        exit()
                    time.sleep(loop_interval)
            except KeyboardInterrupt:
                exit()        

        elif args['<command>'] == 'pmon':
            import xdxsmi_pmon
            args = docopt(xdxsmi_pmon.__doc__, argv=argv)
            xdxsmi_pmon.handle(xdx_smi, args)
        elif args['<command>'] == 'stats':
            import xdxsmi_stats
            args = docopt(xdxsmi_stats.__doc__, argv=argv)
            xdxsmi_stats.handle(xdx_smi, args)
        else:
            # same as the nvidia-smi prompt message 
            exit("Invalid combination of input arguments. Please run 'xdxsmi -h' for help.")
'''        
