"""
usage:
   xdxsmi dvfs [options] get [<ip>...]
   xdxsmi dvfs [options] show [<bound>...]
   xdxsmi dvfs [options] set enable_dvfs <dim>

options:
   -h, --help            print this help message
   -f, --filename=file   redirect query output to a specified file
   -l, --loop=sec        set look up interval, whose unit is second
   -i, --index=val       target gpu to read/write attributes, start from 0

ip:
   enable_dvfs           enable to dynamically adjust the frequency of GPU. 0: disable, 1: enable

bound:
   lower_bound           the lower bound of gpu frequency during dvfs, whose unit is MHz
   upper_bound           the upper bound of gpu frequency during dvfs, whose unit is MHz

xdxsmi dvfs get: to query the enable status of dvfs
xdxsmi dvfs show: to see the gpu frequency range adjusted by dvfs
"""

# from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *
from xdxsmi.cli.render import Render

# map from the input ip to the attr name in dvfs structure
dvfs_map = {
    "enable_dvfs" : "xdx_enable_dvfs"
}

dvfs_unit_map = {
    "enable_dvfs" : "",
    "lower_bound" : "MHz",
    "upper_bound" : "MHz"
}

dvfs_gfx_map = {
    "XDX121" : [["500", "1000"], ["500", "1000"]],
    "XDX130" : [["500", "1000"], ["500", "1000"]],
    "XDX151" : [["500", "1000"], ["500", "1000"]],

    "XDX121 Super" : [["500", "1200"], ["500", "1200"]],
    "XDX151 Super" : [["500", "1200"], ["500", "1200"]],

    "XDX151 Ti" : [["500", "1200"], ["500", "1300"]],
    "XDX160 Ti" : [["500", "1200"], ["500", "1300"]], 
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    index = -1
    num_xdxgpus = len(xdx_smi.xdxgpus)
    try:
        if(args['--index'] == None):
            index = -1                    # default situation
        elif(args['--index'] != None and int(args['--index']) >= 0
            and int(args['--index']) < num_xdxgpus):
            index = int(args['--index'])  # assign index
        else:
            raise ValueError("No devices were found")
    except:
        print("No devices were found")
        exit()
    
    target_gpus = xdx_smi.xdxgpus
    if(index >= 0):
        target_gpus = [xdx_smi.xdxgpus[index]]
       
    if args["set"] and args["enable_dvfs"] and args["<dim>"]:
        try:
            for gpu in target_gpus:
                xdev = gpu.handle
                chip_model = gpu.chip_model
                if chip_model not in dvfs_gfx_map:
                    notify = [["card {} ({}) doesn't support dvfs at this time.".format(gpu.user_index, chip_model)]]
                    Render.render_table(notify, args, upper = False)
                    exit()

                # judge the value of enable_slowdown is valid
                newVal = str(int(args["<dim>"]))
                if(newVal not in ["0","1"]):
                    print ("[XDXSMI ERROR]: Invalid input enable_dvfs", args["<dim>"])
                    print ("[XDXSMI ERROR]: Please select one of the following values and reset it:")
                    print ("                0, 1")
                    exit()
                xdxml_device_set_dvfs_enable(xdev, newVal)
        except Exception:
            element_ip = args["<dim>"]
            exit('''xdxsmi dvfs set enable_dvfs {element_ip} is not a xdxsmi dvfs command.\
                See 'xdxsmi dvfs --help'.'''.format(element_ip=element_ip))
    #Get values for target attributes
    elif args["get"]:
        target_ip = ["enable_dvfs"]
        # filter the same element between target_ip and args["<ip"]
        if args["<ip>"]:
            target_ip = [ip for ip in target_ip if ip in args["<ip>"]]

        # judge the input target_ip is legal
        if len(target_ip) == 0:
            element_ip = next(iter(args["<ip>"]))
            exit('''xdxsmi dvfs get {element_ip} is not a xdxsmi dvfs command.\
                See 'xdxsmi dvfs --help'.'''.format(element_ip=element_ip))
        
        data = []
        # i = 0
        header = ["index", "gpu", "name"] + target_ip
        data.insert(0, header)
        # instead traverse handle to traverse gpus, cause gpus include handle and user_index
        for gpu in target_gpus:
            xdev = gpu.handle 

            dvfsInfo = xdxml_device_get_dvfs_info(xdev)
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
            for ip in target_ip:
                # unifiy the unit
                row.append(str(int(getattr(dvfsInfo, dvfs_map[ip]))) + dvfs_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        Render.render_table(data, args)
    elif args["show"]:
        target_ip = ["lower_bound", "upper_bound"]
        # filter the same element between target_ip and args["<ip"]
        if args["<bound>"]:
            target_ip = [ip for ip in target_ip if ip in args["<bound>"]]

        # judge the input target_ip is legal
        if len(target_ip) == 0:
            element_ip = next(iter(args["<bound>"]))
            exit('''xdxsmi dvfs show {element_ip} is not a xdxsmi dvfs command.\
                See 'xdxsmi dvfs --help'.'''.format(element_ip=element_ip))
        data = []
        header = ["index", "gpu", "name"] + target_ip
        data.insert(0, header)

        for gpu in target_gpus:
            xdev = gpu.handle 
            chip_model = gpu.chip_model
            lower_bound = 0
            upper_bound = 0
            if chip_model in dvfs_gfx_map:
                lower_bound = dvfs_gfx_map[chip_model][1][0] if gpu.gfx_clk == '1300MHz' else dvfs_gfx_map[chip_model][0][0]
                upper_bound = dvfs_gfx_map[chip_model][1][1] if gpu.gfx_clk == '1300MHz' else dvfs_gfx_map[chip_model][0][1]
            else:
                notify = [["card {} ({}) doesn't support dvfs at this time.".format(gpu.user_index, chip_model)]]
                Render.render_table(notify, args, upper = False)
                exit()
            
            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                xdxml_device_get_device_model(xdev).decode('utf-8')]
            for ip in target_ip:
                # unifiy the unit
                row.append(eval(ip) + dvfs_unit_map[ip] if gpu.is_host else 'N/A')
            data.append(row)
        Render.render_table(data, args)

    else:
        # no case falls into this situation?
        exit("%r is not a xdxsmi dvfs command."\
             "See 'xdxsmi dvfs --help'." % args['<command>'])

if __name__ == "__main__":
    print(docopt(__doc__))