"""
usage:
   xdxsmi fan [options] get [<ip>...]
   xdxsmi fan [options] set fan_ctrl_mode <dim>
   xdxsmi fan [options] set fan <percent>

options:
   -h, --help            print this help message
   -f, --filename=file   redirect query output to a specified file
   -l, --loop=sec        set look up interval, whose unit is second
   -i, --index=val       target gpu to read/write attributes, start from 0

ip:
   fan                   fan speed of GPU, whose unit is RPM
   fan_min               minimum fan speed of GPU, whose unit is RPM
   fan_max               maximum fan speed of GPU, whose unit is RPM
   fan_ctrl_mode         fan control mode of GPU [0: firmware control mode, 1: user control mode]
"""

# from typing import Dict
from xdxsmi.api.third_party_docopt import docopt
from xdxsmi.api.smi import XdxSmi
from xdxsmi.api.pyxdxml import *
from xdxsmi.cli.render import Render
from get_set_attrs import setAttrs, getAttrs
import sys

fan_map = {
    "gpu" : "xdx_dbdf",
    "name": "xdx_product_name",
    "spd":  "fan1_input",
    "tgt_spd": "fan1_target",
    "pwm":  "pwm1",
    "pwm_min": "pwm1_min",
    "pwm_max": "pwm1_max",
    "fan_ctrl_mode": "fan1_ctrl_mode",
    "fan": "fan1_target",
    "fan_min": "pwm1_min",
    "fan_max": "pwm1_max"
    #add switch between manual and automatic
}

# (xdx_smi: XdxSmi, args: Dict)
def handle(xdx_smi, args):
    #Collect target gpus
    #target_gpus = [xdx_smi.handle[0]]
    index = -1
    num_xdxgpus = len(xdx_smi.xdxgpus)
    try:
        if(args['--index'] == None):
            index = -1                    # default situation
        elif(args['--index'] != None and int(args['--index']) >= 0
            and int(args['--index']) < num_xdxgpus):
            index = int(args['--index'])  # assign index
        else:
            raise ValueError("No devices were found")
    except:
        print("No devices were found")
        exit()
    
    target_gpus = xdx_smi.xdxgpus
    if(index >= 0):
        target_gpus = [xdx_smi.xdxgpus[index]]

    #Set value for attribute 'pwm'
    if args["set"]:
        #newVal = str(int(args['<mhz>']))
        for gpu in target_gpus:
            xdev = gpu.handle
            is_support_fan = xdxml_device_get_support_fan_info(xdev)
            chip_model = xdxml_device_get_device_model(xdev).decode('utf-8')
            # vendor_name: XDX, BITLAND, GOODTIMES, OTHERS
            vendor_name = xdxml_device_get_vendor(xdev).decode('utf-8')
            if is_support_fan:
                if args["fan_ctrl_mode"]:
                    try:
                        newVal = str(int(args['<dim>']))
                        if(newVal not in ["0","1"]):
                            print ("[XDXSMI ERROR]: Invalid input fan_ctrl_mode", args["<dim>"])
                            print ("[XDXSMI ERROR]: Please select one of the following values and reset it:")
                            print ("                0, 1")
                            exit()
                        xdxml_device_set_fan_ctrl_mode(xdev, newVal)
                    except Exception:
                        element_ip = args['<dim>']
                        exit('''xdxsmi fan set fan_ctrl_mode {element_ip} is not a xdxsmi fan command.\
                            See 'xdxsmi fan --help'.'''.format(element_ip=element_ip))
                elif args["fan"]:
                    try:
                        newVal = str(int(args['<percent>']))
                        if(int(newVal) < 30 or int(newVal) > 100):
                            print ("[XDXSMI ERROR]: Invalid input fan", args["<percent>"])
                            print ("[XDXSMI ERROR]: Please enter an integer from 30 to 100, which represents the percentage of maximum fan speed.")
                            exit()
                        
                        # judge the ctrl mode is 0 or 1
                        mode = xdxml_device_get_fan_ctrl_mode(xdev)
                        if(mode == 0):
                            print ("Warning: the current ctrl mode of fan is 0 (firmware control mode), please set the ctrl mode to 1 (user control mode) first.")
                            exit()

                        xdxml_device_set_fan1_speed(xdev, newVal)
                    except Exception:
                        element_ip = args['<percent>']
                        exit('''xdxsmi fan set fan {element_ip} is not a xdxsmi fan command.\
                            See 'xdxsmi fan --help'.'''.format(element_ip=element_ip))
            else :
                warning = [["Warnning: index = {}, chip model = {}, vendor = {}, this gpu does not support fan command".format(gpu.user_index, chip_model, vendor_name)]]
                Render.render_table(warning, args, upper = False)


    elif args["get"]:
        target_ip = ["fan", "fan_min", "fan_max", "fan_ctrl_mode"]
        # li: filter the same element between target_ip and args["<ip"]
        if args["<ip>"]:
            target_ip = [ip for ip in target_ip if ip in args["<ip>"]]

        # li: judge the input target_ip is legal
        if len(target_ip) == 0:
            element_ip = next(iter(args["<ip>"]))
            exit('''xdxsmi fan get {element_ip} is not a xdxsmi fan command.\
                See 'xdxsmi fan --help'.'''.format(element_ip=element_ip))
        
        data = []
        # i = index
        header = ["index", "gpu", "name"] + target_ip
        data.insert(0, header)
        for gpu in target_gpus:
            xdev = gpu.handle
            is_support_fan = xdxml_device_get_support_fan_info(xdev)
            chip_model = xdxml_device_get_device_model(xdev).decode('utf-8')
            # vendor_name: XDX, BITLAND, GOODTIMES, OTHERS
            vendor_name = xdxml_device_get_vendor(xdev).decode('utf-8')
            if is_support_fan == 0:
                warning = [["Warning: index = {}, chip model = {}, vendor = {}, this gpu does not support fan command".format(gpu.user_index, chip_model, vendor_name)]]
                Render.render_table(warning, args, upper = False)
                exit()
            
            warning = [["Warning: statistical results of fan are subject to error and cannot be calculated strictly in accordance with the set percentage."]]
            Render.render_table(warning, args, upper = False)
            
            # no function to get all the information of fan, so no struct to store fan information
            # hide fan info from the VM
            fan_info_dict = {
            "fan": str(xdxml_device_get_fan_speed(xdev)) if is_support_fan and gpu.is_host else 'N/A',
            "fan_min": str(xdxml_device_get_min_max_fan_speed(xdev)["min_fan_speed"]) if is_support_fan and gpu.is_host else 'N/A',
            "fan_max": str(xdxml_device_get_min_max_fan_speed(xdev)["max_fan_speed"]) if is_support_fan and gpu.is_host else 'N/A',
            "fan_ctrl_mode": str(xdxml_device_get_fan_ctrl_mode(xdev)) if is_support_fan and gpu.is_host else 'N/A'
            }

            row = [
                str(gpu.user_index),
                xdxml_device_get_pci_info(xdev).pci_dbdf.decode("utf8"),
                chip_model]
            for ip in target_ip:
                row.append(fan_info_dict[ip])
            data.append(row)
            # i += 1
        Render.render_table(data, args)
    
    else:
        exit("%r is not a xdxsmi fan command."\
             "See 'xdxsmi fan --help'." % args['<command>'])

if __name__ == "__main__":
    print(docopt(__doc__))