/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include <asm/unaligned.h>

#include <drm/drm_atomic_helper.h>
#include "xdx_dp_drv.h"
#include "xdx_dp_reg.h"
#include "../pix.h"
#include "xdx_device.h"
#include "../../../audio/xdxaudio_hal.h"
#include "drm/xdxgpu_drm.h"
#include "kernel_compatibility.h"

#define LOOP 0x10

static void xdx_dptx_poweron(struct dp_dev *dptx);
void dptx_soft_reset(struct dp_dev *dptx, u32 bits);

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 17, 0))
#define DP_TRAINING_PATTERN_4                   7
#define DP_TRAINING_PATTERN_MASK_1_4            0xf
#define DP_TPS4_SUPPORTED                       (1 << 7)

static inline bool
drm_dp_tps4_supported(const u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
        return dpcd[DP_DPCD_REV] >= 0x14 &&
                dpcd[DP_MAX_DOWNSPREAD] & DP_TPS4_SUPPORTED;
}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 15, 0))
#define DP_DP13_DPCD_REV 0x2200
#define DP_ADJUST_REQUEST_POST_CURSOR2 0x20c
#endif

static const struct drm_display_mode dp_default_modes[] = {
    /* 16 - 1920x1080@60Hz 16:9 */
    { DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2008,
           2052, 2200, 0, 1080, 1084, 1089, 1125, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 2 - 720x480@60Hz 4:3 */
    { DRM_MODE("720x480", DRM_MODE_TYPE_DRIVER, 27000, 720, 736,
           798, 858, 0, 480, 489, 495, 525, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
    /* 4 - 1280x720@60Hz 16:9 */
    { DRM_MODE("1280x720", DRM_MODE_TYPE_DRIVER, 74250, 1280, 1390,
           1430, 1650, 0, 720, 725, 730, 750, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 31 - 1920x1080@50Hz 16:9 */
    { DRM_MODE("1920x1080", DRM_MODE_TYPE_DRIVER, 148500, 1920, 2448,
           2492, 2640, 0, 1080, 1084, 1089, 1125, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 19 - 1280x720@50Hz 16:9 */
    { DRM_MODE("1280x720", DRM_MODE_TYPE_DRIVER, 74250, 1280, 1720,
           1760, 1980, 0, 720, 725, 730, 750, 0,
           DRM_MODE_FLAG_PHSYNC | DRM_MODE_FLAG_PVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_16_9, },
    /* 17 - 720x576@50Hz 4:3 */
    { DRM_MODE("720x576", DRM_MODE_TYPE_DRIVER, 27000, 720, 732,
           796, 864, 0, 576, 581, 586, 625, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
    /* 2 - 720x480@60Hz 4:3 */
    { DRM_MODE("720x480", DRM_MODE_TYPE_DRIVER, 27000, 720, 736,
           798, 858, 0, 480, 489, 495, 525, 0,
           DRM_MODE_FLAG_NHSYNC | DRM_MODE_FLAG_NVSYNC),
      .picture_aspect_ratio = HDMI_PICTURE_ASPECT_4_3, },
};

static bool dptx_bandwidth_ok(struct dp_dev *dptx,
                   u32 clock, u32 bpp,
                   unsigned int lanes, unsigned int rate)
{
    u32 max_bw, req_bw;

    req_bw = clock * bpp / 8;
    max_bw = lanes * rate;

    if (req_bw > max_bw)
        return false;

    return true;
}

static int dptx_read_dpcd(struct dp_dev *dptx, u32 addr, u8 *byteptr)
{
	int ret;
	//struct dp_dev *dp = dptx->dp;

    ret = drm_dp_dpcd_read(&(dptx->aux), addr, byteptr, 1);
	if (ret != 1)
		return ret;

	return 0;
}

static int dptx_write_dpcd(struct dp_dev *dptx, u32 addr, u8 byte)
{
	int ret;
	//struct dp_dev *dp = dptx->dp;

	ret = drm_dp_dpcd_write(&(dptx->aux), addr, &byte, 1);
	if (ret != 1)
		return ret;

	return 0;
}

static int dptx_read_bytes_from_dpcd(struct dp_dev *dptx, u32 reg_addr, u8 *bytes, u32 len)
{
	u32 i;

	for (i=0; i<len; i++) {
		dptx_read_dpcd(dptx, (reg_addr+i), (bytes+i));
	}
	return 0;
}


static u32 dptx_pcs_readl(struct dp_dev *dptx, u32 offset)
{
	struct xdx_dm *dm = output_to_dm(dptx->output);
	struct xdx_device *xdev = dm_to_xdev(dm);

	return xdx_mmio_readl(xdev, xdev->hw_info.dp_pcs_regs[dptx->id].base + (offset << 2));
}

static void dptx_pcs_writel(struct dp_dev *dptx, u32 offset, u16 value)
{
	struct xdx_dm *dm = output_to_dm(dptx->output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	u32 data;

	data = (value << 16) | value;
	xdx_mmio_writel(xdev, xdev->hw_info.dp_pcs_regs[dptx->id].base + (offset << 2), data);
}


static u32 dptx_readl(struct dp_dev *dptx, u32 offset)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev, xdev->hw_info.dp_regs[dptx->id].base + offset);
}

static void dptx_writel(struct dp_dev *dptx, u32 offset, u32 value)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev,
            xdev->hw_info.dp_regs[dptx->id].base + offset, value);
}

static void dptx_mask_writel(struct dp_dev *dptx, u32 offset, u32 mask, u32 value)
{
    u32 tmp = dptx_readl(dptx, offset) & ~mask;

    tmp |= (value & mask);
    dptx_writel(dptx, offset, tmp);
}

static u32 dptx_tile_readl(struct dp_dev *dptx, u32 offset)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev,
            xdev->hw_info.hdmi_tile_regs[dptx->id].base + offset);
}

static void dptx_tile_writel(struct dp_dev *dptx, u32 offset, u32 value)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev,
             xdev->hw_info.hdmi_tile_regs[dptx->id].base + offset, value);
}

static void dptx_tile_mask_writel(struct dp_dev *dptx,
        u32 offset, u32 mask, u32 value)
{
    u32 tmp = dptx_tile_readl(dptx, offset) & ~mask;

    tmp |= (value & mask);
    dptx_tile_writel(dptx, offset, tmp);
}

static inline void xdx_dp_h2a_doorbell_w(struct dp_dev *dptx, u64 value)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    xdx_mmio_writel(xdev,
            SYS_PANGU_REG_HDMIDOORBELL + 8,
            value & 0xffffffff);

    xdx_mmio_writel(xdev,
            SYS_PANGU_REG_HDMIDOORBELL + 0xc,
            (value >> 32) & 0xffffffff);
}

static inline u32 xdx_dp_h2a_doorbell_r(struct dp_dev *dptx)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    return xdx_mmio_readl(xdev,
            SYS_PANGU_REG_HDMIDOORBELL + 8);
}

static inline u64 xdx_dp_h2a_doorbell_r64(struct dp_dev *dptx)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u64 lret, hret;

    hret = xdx_mmio_readl(xdev,
            SYS_PANGU_REG_HDMIDOORBELL + 0xc);
    lret = xdx_mmio_readl(xdev,
            SYS_PANGU_REG_HDMIDOORBELL + 8);

    return ((0xffffffff & lret) | (0xffffffff & hret) << 32);
}

__attribute__((unused)) static void dptx_core_regs_dump(struct dp_dev *dptx)
{
    int i;
    const struct reg {
        u32 offset;
        const char *name;
    } regs[] = {
        { DPTX_CONFIG_REG1, "DPTX_CONFIG_REG1"},
        { DPTX_CONFIG_REG2, "DPTX_CONFIG_REG2"},
        { DPTX_CONFIG_REG3, "DPTX_CONFIG_REG3"},
        { DPTX_CCTL, "DPTX_CCTL"},
        { DPTX_SOFT_RESET_CTRL, "DPTX_SOFT_RESET_CTRL"},
        { DPTX_VSAMPLE_CTRL, "DPTX_VSAMPLE_CTRL"},
        { DPTX_VSAMPLE_STUFF_CTRL1, "DPTX_VSAMPLE_STUFF_CTRL1"},
        { DPTX_VSAMPLE_STUFF_CTRL2, "DPTX_VSAMPLE_STUFF_CTRL2"},
        { DPTX_VINPUT_POLARITY_CTRL, "DPTX_VINPUT_POLARITY_CTRL"},
        { DPTX_VIDEO_CONFIG1, "DPTX_VIDEO_CONFIG1"},
        { DPTX_VIDEO_CONFIG2, "DPTX_VIDEO_CONFIG2"},
        { DPTX_VIDEO_CONFIG3, "DPTX_VIDEO_CONFIG3"},
        { DPTX_VIDEO_CONFIG4, "DPTX_VIDEO_CONFIG4"},
        { DPTX_VIDEO_CONFIG5, "DPTX_VIDEO_CONFIG5"},
        { DPTX_VIDEO_MSA1, "DPTX_VIDEO_MSA1"},
        { DPTX_VIDEO_MSA2, "DPTX_VIDEO_MSA2"},
        { DPTX_VIDEO_MSA3, "DPTX_VIDEO_MSA3"},
        { DPTX_VIDEO_HBLANK_INTERVAL, "DPTX_VIDEO_HBLANK_INTERVAL"},
        { DPTX_PHYIF_CTRL, "DPTX_PHYIF_CTRL"},
        { DPTX_PHY_TX_EQ, "DPTX_PHY_TX_EQ"},
        { DPTX_CUSTOMPAT0, "DPTX_CUSTOMPAT0"},
        { DPTX_CUSTOMPAT1, "DPTX_CUSTOMPAT1"},
        { DPTX_CUSTOMPAT2, "DPTX_CUSTOMPAT2"},
        { DPTX_PHYIF_PWRDOWN_CTRL, "DPTX_PHYIF_PWRDOWN_CTRL"},
    };

    for (i = 0; i < ARRAY_SIZE(regs); i++) {
        dev_err(dptx->dev, "%s: 0x%x\n", regs[i].name,
            dptx_readl(dptx, regs[i].offset));
    }
}

__attribute__((unused)) static void dptx_tile_regs_dump(struct dp_dev *dptx)
{
    int i;
    const struct reg {
        u32 offset;
        const char *name;
    } regs[] = {
        { PHYREF_PLL_EN, "PHYREF_PLL_EN"},
        { DP_PHY_CTRL, "DP_PHY_CTRL"},
        { PHYREF_PLL_LOCK, "PHYREF_PLL_LOCK"},
        { DP_CTRL, "DP_CTRL"},
        { DP_PCS_CTRL, "DP_PCS_CTRL"},
    };

    for (i = 0; i < ARRAY_SIZE(regs); i++) {
        dev_err(dptx->dev, "%s: 0x%x\n", regs[i].name,
            dptx_tile_readl(dptx, regs[i].offset));
    }
}

static int dptx_phy_wait_busy(struct dp_dev *dptx)
{
    int timeout = 1000;
    u32 val;

    for (; timeout; timeout--) {
        val = dptx_readl(dptx, DPTX_PHYIF_CTRL);
        if (!(val & FIELD_PREP(PHY_BUSY, 0xf))) {
                break;
        }
        udelay(1000);
    }
    if (val & FIELD_PREP(PHY_BUSY, 0xf)) {
            dev_err(dptx->dev,"dptx phy wait busy fail\n");
            return -ETIMEDOUT;
    }

    return 0;
}

static int xdx_dp_preinit(struct dp_dev *dptx)
{
    int timeout = 1000;
    DPDBCmd_T_SINK PhyConfig;
    u32 val;

    PhyConfig.all = 0;
    PhyConfig.cmd = cH02ADisplayDBCmdDP;
    PhyConfig.subCmd = cH02ADPSubCmdCNTLPreConfig;
    PhyConfig.args0 = dptx->id;
    xdx_dp_h2a_doorbell_w(dptx,(u64)PhyConfig.all);

    for (; timeout; timeout--) {
        val = xdx_dp_h2a_doorbell_r(dptx);
        if(val & BIT(0))
            break;

        usleep_range(1000, 2000);
    }

    if (!(val & BIT(0))) {
        dev_err(dptx->dev,"dptx pll config fail\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static int xdx_dptx_tile_phy_init(struct dp_dev *dptx)
{
    struct video_params *params = &dptx->vparams;
    int timeout = 1000;
    u32 val;

    dptx_tile_mask_writel(dptx, PHYREF_PLL_EN,
                    PLL0_PHYREF_EN, PLL0_PHYREF_EN);
    for (; timeout; timeout--) {
        val = dptx_tile_readl(dptx, PHYREF_PLL_LOCK);
        if(val & PLL0_PHYREF_CORE_LOCK)
            break;

        usleep_range(1000, 2000);
    }

    if (!(val & PLL0_PHYREF_CORE_LOCK)) {
        dev_err(dptx->dev,"xdx hdmi wait PHYREF_PLL_LOCK timeout\n");
        return -ETIMEDOUT;
    }
    //SRAM bypass
    dptx_tile_mask_writel(dptx, DP_PHY_CTRL, SRAM_BYPASS, SRAM_BYPASS);
    // video mapping
    dptx_tile_writel(dptx,
    DP_CTRL, FIELD_PREP(TILE_VIDEO_MAPPING, params->bpc_mapping) | BIT(1));
    //phy mode config
    dptx_tile_mask_writel(dptx, DP_PCS_CTRL, DP_PCS_WIDTH | DP_LANE_RATE,
       FIELD_PREP(DP_PCS_WIDTH, 1) | FIELD_PREP(DP_LANE_RATE, LANE_RATE_HBR3));

    return 0;
}

static int xdx_drm_dp_read_extended_dpcd_caps(struct drm_dp_aux *aux,
                        u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    u8 dpcd_ext[6];
    int ret;

    if (!(dpcd[DP_TRAINING_AUX_RD_INTERVAL] &
        DP_EXTENDED_RECEIVER_CAP_FIELD_PRESENT))
        return 0;

    ret = drm_dp_dpcd_read(aux, DP_DP13_DPCD_REV, &dpcd_ext,
			       sizeof(dpcd_ext));
    if (ret < 0)
        return ret;
    if (ret != sizeof(dpcd_ext))
        return -EIO;

    if (dpcd[DP_DPCD_REV] > dpcd_ext[DP_DPCD_REV]) {
        DRM_DEBUG_KMS("%s: Extended DPCD rev less than base DPCD rev (%d > %d)\n",
                aux->name, dpcd[DP_DPCD_REV], dpcd_ext[DP_DPCD_REV]);
        return 0;
    }

    if (!memcmp(dpcd, dpcd_ext, sizeof(dpcd_ext)))
        return 0;

    DRM_DEBUG_KMS("%s: Base DPCD: %*ph\n",
                aux->name, DP_RECEIVER_CAP_SIZE, dpcd);

    memcpy(dpcd, dpcd_ext, sizeof(dpcd_ext));

    return 0;
}

static int xdx_drm_dp_read_dpcd_caps(struct drm_dp_aux *aux,
			  u8 dpcd[DP_RECEIVER_CAP_SIZE])
{
    int ret;

    ret = drm_dp_dpcd_read(aux, DP_DPCD_REV, dpcd, DP_RECEIVER_CAP_SIZE);
    if (ret < 0)
        return ret;
    if (ret != DP_RECEIVER_CAP_SIZE || dpcd[DP_DPCD_REV] == 0)
        return -EIO;

    ret = xdx_drm_dp_read_extended_dpcd_caps(aux, dpcd);
    if (ret < 0)
        return ret;

    return ret;
}

static void dptx_link_set_preemp_vswing(struct dp_dev *dptx)
{
    unsigned int i;
    u32 pes = 0,vss = 0;

    for (i = 0; i < dptx->link.lanes; i++) {
        u8 pe;
        u8 vs;

        pe = dptx->link.preemp_level[i];
        vs = dptx->link.vswing_level[i];
        if (i > 3)
            break;

        pes |= (pe << DPTX_PHY_TX_EQ_PREEMP_SHIFT(i)) &
        DPTX_PHY_TX_EQ_PREEMP_MASK(i);
        vss |= (vs << DPTX_PHY_TX_EQ_VSWING_SHIFT(i)) &
        DPTX_PHY_TX_EQ_VSWING_MASK(i);
        dev_dbg(dptx->dev, "loop:%d pe:%x vs:%x\n",i,pe,vs);
    }
    dptx_writel(dptx, DPTX_PHY_TX_EQ, pes | vss);
}

static void dptx_phy_enable_xmit(struct dp_dev *dptx,
        unsigned int lanes, bool enable)
{
    u32 phyifctrl = 0;

    switch (lanes) {
    case 4:
    case 2:
    case 1:
        phyifctrl = GENMASK(lanes - 1, 0);
        break;
    default:
        break;
    }

    if (!enable)
        phyifctrl = 0;

    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, XMIT_ENABLE,
               FIELD_PREP(XMIT_ENABLE, phyifctrl));
}

static int xdx_dptx_link_power_on(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    u8 value;
    int ret;

    if (link->rx_caps[DP_DPCD_REV] < 0x11)
        return 0;

    ret = drm_dp_dpcd_readb(&dptx->aux, DP_SET_POWER, &value);
    if (ret < 0)
        return ret;

    value &= ~DP_SET_POWER_MASK;
    value |= DP_SET_POWER_D0;

    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_SET_POWER, value);
    if (ret < 0)
        return ret;

    usleep_range(1000, 2000);

    return 0;
}

__attribute__((unused)) static int xdx_dptx_link_power_off(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    u8 value;
    int ret;

    if (link->rx_caps[DP_DPCD_REV] < 0x11)
        return 0;

    ret = drm_dp_dpcd_readb(&dptx->aux, DP_SET_POWER, &value);
    if (ret < 0)
        return ret;

    value &= ~DP_SET_POWER_MASK;
    value |= DP_SET_POWER_D3;

    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_SET_POWER, value);
    if (ret < 0)
        return ret;

    return 0;
}

static int dptx_set_training_pattern(struct dp_dev *dptx, u8 pattern)
{
    u8 buf = 0;
    int ret;

    if (pattern && pattern != DP_TRAINING_PATTERN_4) {
        buf |= DP_LINK_SCRAMBLING_DISABLE;

        dptx_mask_writel(dptx, DPTX_CCTL, SCRAMBLE_DIS, SCRAMBLE_DIS);
    } else {
        dptx_mask_writel(dptx, DPTX_CCTL, SCRAMBLE_DIS, 0);
    }

    switch (pattern) {
    case DP_TRAINING_PATTERN_DISABLE:
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, TPS_SEL,
               FIELD_PREP(TPS_SEL, DPTX_PHY_PATTERN_NONE));
        break;
    case DP_TRAINING_PATTERN_1:
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, TPS_SEL,
               FIELD_PREP(TPS_SEL, DPTX_PHY_PATTERN_TPS_1));
        break;
    case DP_TRAINING_PATTERN_2:
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, TPS_SEL,
               FIELD_PREP(TPS_SEL, DPTX_PHY_PATTERN_TPS_2));
        break;
    case DP_TRAINING_PATTERN_3:
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, TPS_SEL,
               FIELD_PREP(TPS_SEL, DPTX_PHY_PATTERN_TPS_3));
        break;
    case DP_TRAINING_PATTERN_4:
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, TPS_SEL,
               FIELD_PREP(TPS_SEL, DPTX_PHY_PATTERN_TPS_4));
        break;
    default:
        return -EINVAL;
    }

    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_TRAINING_PATTERN_SET,
                 buf | pattern);
    if (ret < 0)
        return ret;

    return 0;
}

static int dptx_link_training_lanes_set(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    int ret;
    unsigned int i;

    for (i = 0; i < link->lanes; i++) {
        u8 byte = 0;

        byte |= ((link->vswing_level[i] <<
                DP_TRAIN_VOLTAGE_SWING_SHIFT) &
                DP_TRAIN_VOLTAGE_SWING_MASK);

        if (link->vswing_level[i] == 3)
            byte |= DP_TRAIN_MAX_SWING_REACHED;

        byte |= ((link->preemp_level[i] <<
            DP_TRAIN_PRE_EMPHASIS_SHIFT) &
            DP_TRAIN_PRE_EMPHASIS_MASK);

        if (link->preemp_level[i] == 3)
            byte |= DP_TRAIN_MAX_PRE_EMPHASIS_REACHED;

        link->train_value[i] = byte;
    }

    ret = drm_dp_dpcd_write(&dptx->aux,
            DP_TRAINING_LANE0_SET, link->train_value, link->lanes);
    if (ret < 0)
        return ret;

    return 0;
}

static bool dptx_link_max_vswing_reached(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    int i;

    for (i = 0; i < link->lanes; i++)
        if ((link->train_value[i] & DP_TRAIN_MAX_SWING_REACHED) == 0)
            return false;

    return true;
}

static int dptx_link_reduce_rate(struct dp_dev *dptx)
{
    unsigned int rate = dptx->link.rate;

    switch (rate) {
    case 162000:
        return -EPROTO;
    case 270000:
        rate = 162000;
        break;
    case 540000:
        rate = 270000;
        break;
    case 810000:
        rate = 540000;
        break;
    }

    dev_dbg(dptx->dev, "%s: Reducing rate from %d to %d\n",
            __func__, dptx->link.rate, rate);

    dptx->link.rate = rate;

    return 0;
}

static u8 dptx_select_ch_eq_pattern(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;

    if (drm_dp_tps4_supported(link->rx_caps))
        return DP_TRAINING_PATTERN_4;
    else if (drm_dp_tps3_supported(link->rx_caps))
        return DP_TRAINING_PATTERN_3;
    else
        return DP_TRAINING_PATTERN_2;
}

static int dptx_link_clock_recovery(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    u8 status[DP_LINK_STATUS_SIZE];
    u8 cr_loops = 1;
    bool vswing_reached = false;
    int ret, i;

    ret = dptx_set_training_pattern(dptx, DP_TRAINING_PATTERN_1);
    if (ret < 0)
        return ret;

    for (;;) {
        ret = dptx_link_training_lanes_set(dptx);
        if (ret)
            return ret;
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)))
        drm_dp_link_train_clock_recovery_delay(&dptx->aux, link->rx_caps);
#else
        drm_dp_link_train_clock_recovery_delay(link->rx_caps);
#endif
        ret = drm_dp_dpcd_read_link_status(&dptx->aux, status);
        if (ret < 0) {
            dev_dbg(dptx->dev, "%s:dptx link status read error: %d\n", __func__, ret);
            return ret;
        }

        if (drm_dp_clock_recovery_ok(status, link->lanes)) {
            link->cr_state = true;
            link->link_state = CHANNEL_EQUALIZATION;
            dev_dbg(dptx->dev,"%s: in drm_dp_clock_recovery_ok\n", __func__);
            break;
        }

        if (cr_loops >= 5) {
            dev_err(dptx->dev, "vswing tried 5 times\n");
            link->cr_state = false;
            break;
        }

        if (vswing_reached) {
            link->cr_state = false;
            break;
        }

        for (i = 0; i < link->lanes; i++) {
            link->cr_vswing_level[i] =
                drm_dp_get_adjust_request_voltage(status, i) >>
                    DP_TRAIN_VOLTAGE_SWING_SHIFT;

            link->cr_preemp_level[i] =
                drm_dp_get_adjust_request_pre_emphasis(status, i) >>
                    DP_TRAIN_PRE_EMPHASIS_SHIFT;
        }

        if (link->vswing_level[0] != link->cr_vswing_level[0]) {
            for (i = 0; i < link->lanes; i++) {
                link->vswing_level[i] = link->cr_vswing_level[i];
                link->preemp_level[i] = link->cr_preemp_level[i];
            }
            cr_loops = 1;
        } else {
            cr_loops++;
        }

        dptx_link_set_preemp_vswing(dptx);

        if (dptx_link_max_vswing_reached(dptx))
            vswing_reached = true;
    }

    if (!link->cr_state) {
        ret = dptx_link_reduce_rate(dptx);
        if (ret < 0)
            return ret;
        else
            link->link_state = TRAIN_INIT;
    }

    return 0;
}

static int dptx_link_channel_equalization(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    u8 status[DP_LINK_STATUS_SIZE], pattern;
    u8 last_swing = link->vswing_level[0];
    unsigned int loops = 1;
    bool sidestep = false;
    int i, ret;

    pattern = dptx_select_ch_eq_pattern(dptx);

    ret = dptx_set_training_pattern(dptx, pattern);
    if (ret < 0)
        return ret;

    for (;;) {
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)))
        drm_dp_link_train_channel_eq_delay(&dptx->aux, link->rx_caps);
#else
        drm_dp_link_train_channel_eq_delay(link->rx_caps);
#endif
        ret = drm_dp_dpcd_read_link_status(&dptx->aux, status);
        if (ret < 0) {
            dev_dbg(dptx->dev,
              "drm_dp_dpcd_read_link_status failed when equalizer training\n");
            return ret;
        }

        if (!drm_dp_clock_recovery_ok(status, link->lanes)) {
            dev_dbg(dptx->dev,
           "clock recovery failed when equalizer training\n");
            link->cr_state = false;
        }

        if (drm_dp_channel_eq_ok(status, link->lanes)) {
            link->ch_eq_state = true;
            break;
        } else {
            link->ch_eq_state = false;
            dev_dbg(dptx->dev,
                "channel_eq failed when equalizer training\n");
        }

        if (loops >= 5) {
            dev_dbg(dptx->dev, "vswing tried 5 times\n");
            if (!sidestep) {
                for (i = 0; i < link->lanes; i++) {
                    link->cr_vswing_level[i] = last_swing;
                }
                ret = dptx_link_training_lanes_set(dptx);
                if (ret) {
                    dev_dbg(dptx->dev, "dptx_link_training_lanes_set fail\n");
                    return ret;
                }
                dptx_link_set_preemp_vswing(dptx);
                sidestep = true;
                continue;
            } else {
                dev_dbg(dptx->dev, "dptx link retry sidestep swing failed\n");
                break;
            }
        }

        for (i = 0; i < link->lanes; i++) {
            link->cr_vswing_level[i] =
                drm_dp_get_adjust_request_voltage(status, i) >>
                    DP_TRAIN_VOLTAGE_SWING_SHIFT;

            link->cr_preemp_level[i] =
                drm_dp_get_adjust_request_pre_emphasis(status, i) >>
                    DP_TRAIN_PRE_EMPHASIS_SHIFT;
        }

        if ((link->preemp_level[0] != link->cr_preemp_level[0]) ||
            (link->vswing_level[0] != link->cr_vswing_level[0])) {
            for (i = 0; i < link->lanes; i++) {
                link->vswing_level[i] = link->cr_vswing_level[i];
                link->preemp_level[i] = link->cr_preemp_level[i];
            }
            loops = 1;
        } else {
            loops++;
        }

        ret = dptx_link_training_lanes_set(dptx);
        if (ret) {
            dev_dbg(dptx->dev, "dptx_link_training_lanes_set fail\n");
            return ret;
        }

        dptx_link_set_preemp_vswing(dptx);
    }

    if (!link->ch_eq_state) {
        ret = dptx_link_reduce_rate(dptx);
        if (ret < 0) {
            dev_dbg(dptx->dev, "dptx_link_reduce_rate fail\n");
            return ret;
        } else {
            link->link_state = TRAIN_INIT;
        }
    } else {
        link->link_state = TRAIN_COMPLETED;
        dev_dbg(dptx->dev,"%s: dptx link channel equalization success\n",__func__);
    }

    return 0;
}

static void dptx_link_training_init(struct dptx_link *link)
{
    unsigned int i;

    for (i = 0; i < 4; i++) {
        link->vswing_level[i] = 0;
        link->cr_vswing_level[i] = 0;
        link->preemp_level[i] = 0;
        link->cr_preemp_level[i] = 0;
    }

    link->cr_state = false;
    link->ch_eq_state = false;
    link->link_state = TRAIN_START;
}

static void dptx_phy_set_powerdown(struct dp_dev *dptx, unsigned int lanes)
{
    u32 val;
    switch (lanes) {
        case 1:
            val = 0x13330;
            break;
        case 2:
            val = 0x13300;
            break;
        case 4:
            val = 0x0000;
            break;
        default:
            WARN(1,"pd Invalid number of lanes %d\n", lanes);
            return;
    }

    dptx_writel(dptx, DPTX_PHYIF_PWRDOWN_CTRL, val);
}

static int dptx_link_training_start(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    int ret;
    u8 val;
    bool ssc_en = false;

    ssc_en = !!(link->rx_caps[DP_MAX_DOWNSPREAD] & DP_MAX_DOWNSPREAD_0_5);
    /* Move PHY to P3 */
    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, PHY_POWERDOWN,
               FIELD_PREP(PHY_POWERDOWN, 0x3));

    ret = dptx_phy_wait_busy(dptx);
    if (ret < 0)
        goto out;

    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, PHY_LANES,
            FIELD_PREP(PHY_LANES, link->lanes / 2));

    val = drm_dp_link_rate_to_bw_code(link->rate) / 10;
    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, PHY_RATE,
            FIELD_PREP(PHY_RATE, val));

    dptx_phy_set_powerdown(dptx, link->lanes);

    if (ssc_en) {
        dptx_tile_mask_writel(dptx, DP_PCS_CTRL,
            DP_LANE_MPLLB_SSC_EN, DP_LANE_MPLLB_SSC_EN);
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, SSC_DIS, 0);
    } else {
        dptx_tile_mask_writel(dptx, DP_PCS_CTRL, DP_LANE_MPLLB_SSC_EN, 0);
        dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, SSC_DIS, SSC_DIS);
    }
    /* Move PHY to P0 */
    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, PHY_POWERDOWN,
               FIELD_PREP(PHY_POWERDOWN, 0x0));

    ret = dptx_phy_wait_busy(dptx);
    if (ret < 0)
        goto out;

    dptx_writel(dptx, DPTX_PHY_TX_EQ, 0);

    ret = dptx_set_training_pattern(dptx, DP_TRAINING_PATTERN_DISABLE);
    if (ret < 0) {
        return ret;
    }

    dptx_phy_enable_xmit(dptx, link->lanes, true);

    val = drm_dp_link_rate_to_bw_code(link->rate);

    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_LINK_BW_SET, val);
    if (ret < 0)
        goto out;

    val = link->lanes;
    if (drm_dp_enhanced_frame_cap(link->rx_caps)) {
        val |= DP_LANE_COUNT_ENHANCED_FRAME_EN;
        dptx_mask_writel(dptx, DPTX_CCTL, ENHANCE_FRAMING_EN,
                   FIELD_PREP(ENHANCE_FRAMING_EN, 1));
    } else {
        dptx_mask_writel(dptx, DPTX_CCTL, ENHANCE_FRAMING_EN,
                   FIELD_PREP(ENHANCE_FRAMING_EN, 0));
    }
    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_LANE_COUNT_SET, val);
    if (ret < 0)
        goto out;

    val = ssc_en ? DP_SPREAD_AMP_0_5 : 0;
    ret = drm_dp_dpcd_writeb(&dptx->aux, DP_DOWNSPREAD_CTRL, val);
    if (ret < 0)
        goto out;

    ret = drm_dp_dpcd_writeb(&dptx->aux,
            DP_MAIN_LINK_CHANNEL_CODING_SET, DP_SET_ANSI_8B10B);
    if (ret < 0)
        goto out;

    link->link_state = CLOCK_RECOVERY;

    return ret;
out:
    link->link_state = TRAIN_FAILED;

    return ret;
}

static int xdx_dptx_fast_link_training(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    u8 status[DP_LINK_STATUS_SIZE], pattern;
    int ret;

    dptx_link_training_init(link);

    dev_info(dptx->dev, "fast-training link: %u lanes at %u MHz\n",
            link->lanes, link->rate / 100);

    ret = dptx_link_training_start(dptx);
    if (ret < 0) {
        dev_err(dptx->dev, "dptx training start failed: %d\n", ret);
        return ret;
    }

    ret = dptx_set_training_pattern(dptx, DP_TRAINING_PATTERN_1);
    if (ret < 0)
        goto out;

    usleep_range(500, 1000);

    pattern = dptx_select_ch_eq_pattern(dptx);
    ret = dptx_set_training_pattern(dptx, pattern);
    if (ret < 0)
        goto out;

    usleep_range(500, 1000);

    ret = drm_dp_dpcd_read_link_status(&dptx->aux, status);
    if (ret < 0) {
        dev_err(dptx->dev, "read link status error: %d\n", ret);
        goto out;
    }

    if (!drm_dp_clock_recovery_ok(status, link->lanes)) {
        dev_err(dptx->dev, "clock recovery error\n");
        ret = -EINVAL;
        goto out;
    }

    if (!drm_dp_channel_eq_ok(status, link->lanes)) {
        dev_err(dptx->dev, "channel equalization error\n");
        ret = -EINVAL;
        goto out;
    }

out:
    dptx_set_training_pattern(dptx, DP_TRAINING_PATTERN_DISABLE);
    return ret;
}

static int xdx_dptx_full_link_training(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    int ret = 0;
    bool trainning_completed = false;

    link->link_state = TRAIN_INIT;

    while (!trainning_completed) {
        switch (link->link_state) {
        case TRAIN_INIT:
            dptx_link_training_init(link);
            break;
        case TRAIN_START:
            ret = dptx_link_training_start(dptx);
            if (ret < 0) {
                dev_err(dptx->dev, "dptx training start failed: %d\n", ret);
                goto out;
            }
            break;
        case CLOCK_RECOVERY:
            ret = dptx_link_clock_recovery(dptx);
            if (ret < 0) {
                dev_err(dptx->dev, "clock recovery failed: %d\n", ret);
                goto out;
            }
            break;
        case CHANNEL_EQUALIZATION:
            ret = dptx_link_channel_equalization(dptx);
            if (ret < 0) {
                dev_err(dptx->dev, "channel equalization failed: %d\n", ret);
                goto out;
            }
            break;
        case TRAIN_COMPLETED:
            trainning_completed = true;
            break;
        case TRAIN_FAILED:
            goto out;
        }
    }

    dev_dbg(dptx->dev, "channel equalization succeeded\n");
out:
    dptx_set_training_pattern(dptx, DP_TRAINING_PATTERN_DISABLE);

    return ret;
}

static int xdx_dptx_link_training(struct dp_dev *dptx)
{
    struct dptx_link *link = &dptx->link;
    int ret;

    if (xdx_drm_dp_fast_training_cap(link->rx_caps)) {
        if (link->cr_state && link->ch_eq_state) {
            ret = xdx_dptx_fast_link_training(dptx);
            if (!ret)
                dev_dbg(dptx->dev, "fast train error, retry full-train\n");
            else
                return 0;
        }
    }

    ret = xdx_dptx_full_link_training(dptx);
    if (ret < 0) {
        dev_err(dptx->dev, "DP tx full link train error: %d\n", ret);
        return ret;
    }

    dptx->link_trained = true;

    return 0;
}

static bool xdx_dptx_needs_link_retrain(struct dp_dev *dptx)
{
    u8 link_status[DP_LINK_STATUS_SIZE];

    if (!dptx->encoder_is_on)
        return false;

    if (!dptx->link_trained)
        return true;

    if (drm_dp_dpcd_read_link_status(&dptx->aux, link_status) < 0)
        return false;

    return !drm_dp_channel_eq_ok(link_status, dptx->link.lanes) ||
           !drm_dp_clock_recovery_ok(link_status, dptx->link.lanes);
}

static int xdx_dptx_phy_power_down(struct dp_dev *dptx)
{
    int ret;

    dptx->link_trained = false;

    dptx_mask_writel(dptx, DPTX_PHYIF_CTRL, PHY_POWERDOWN,
               FIELD_PREP(PHY_POWERDOWN, 0x3));

    ret = dptx_phy_wait_busy(dptx);
    if (ret < 0) {
        dev_err(dptx->dev, "Timed out waiting for PHY BUSY\n");
        dptx_soft_reset(dptx, DPTX_SRST_CTRL_PHY);
        return ret;
    }

    return 0;
}

static void xdx_dptx_video_set_sink_msa(struct dp_dev *dptx,
        u16 v_start, u16 h_start)
{
    struct video_params *params = &dptx->vparams;
    u16 misc = 0;

    switch (params->pix_enc) {
    case DRM_COLOR_FORMAT_RGB444:
        misc |= DP_MSA_MISC_COLOR_RGB;
        break;
    case DRM_COLOR_FORMAT_YCRCB444:
        misc |= DP_MSA_MISC_COLOR_YCBCR_444_BT709;
        break;
    case DRM_COLOR_FORMAT_YCRCB422:
        misc |= DP_MSA_MISC_COLOR_YCBCR_422_BT709;
        break;
    case DRM_COLOR_FORMAT_YCRCB420:
        break;
    default:
        misc |= DP_MSA_MISC_COLOR_RGB;
    }

    switch (params->bpc) {
    case COLOR_DEPTH_6:
        misc |= DP_MSA_MISC_6_BPC;
        break;
    case COLOR_DEPTH_8:
        misc |= DP_MSA_MISC_8_BPC;
        break;
    case COLOR_DEPTH_10:
        misc |= DP_MSA_MISC_10_BPC;
        break;
    case COLOR_DEPTH_12:
        misc |= DP_MSA_MISC_12_BPC;
        break;
    case COLOR_DEPTH_16:
        misc |= DP_MSA_MISC_16_BPC;
        break;
    default:
        misc |= DP_MSA_MISC_8_BPC;
    }

    dptx_writel(dptx, DPTX_VIDEO_MSA1,
             FIELD_PREP(VSTART, v_start) | FIELD_PREP(HSTART, h_start));
    dptx_writel(dptx, DPTX_VIDEO_MSA2, FIELD_PREP(MISC0, misc));
    dptx_writel(dptx, DPTX_VIDEO_MSA3, FIELD_PREP(MISC1, misc >> 8));
}

static int dptx_video_ts_calculate_change(struct dp_dev *dptx)
{
    struct video_params *params = &dptx->vparams;
    struct drm_display_mode *mode = &dptx->dp_mode;
    struct dptx_link *link = &dptx->link;
    u32 h_blanking = mode->htotal - mode->hdisplay;
    u32 ts = 0, tu = 0, tu_frac = 0, T1 = 0, T2 = 0;
    u32 rate = link->rate / 1000;

    dev_dbg(dptx->dev,"%s in clock:%d bpp:%d link->lanes:%d link->rate:%d\n",
         __func__, mode->clock, params->bpp, link->lanes, link->rate);

    ts = (8 * mode->clock * params->bpp) / (link->lanes * rate);
    tu = ts / 1000;
    tu_frac = ts / 100 - tu * 10;

    if (params->pixel_mode == DPTX_MP_SINGLE_PIXEL) {
        if (tu < 6)
            params->init_threshold = 32;
        else if (h_blanking <= 40 &&
         params->pix_enc == DRM_COLOR_FORMAT_YCRCB420)
            params->init_threshold = 3;
        else if (h_blanking <= 80 &&
         params->pix_enc != DRM_COLOR_FORMAT_YCRCB420)
            params->init_threshold = 12;
        else
            params->init_threshold = 15;
    } else {
        switch (params->bpc) {
        case COLOR_DEPTH_6:
            T1 = (4 * 1000 / 9) * link->lanes;
            break;
        case COLOR_DEPTH_8:
            if (params->pix_enc == DRM_COLOR_FORMAT_YCRCB422) {
                T1 = (1000 / 2) * link->lanes;
            } else {
                if (params->pixel_mode == DPTX_MP_DUAL_PIXEL)
                    T1 = (1000 / 3) * link->lanes;
                else
                    T1 = (3000 / 16) * link->lanes;
            }
            break;
        case COLOR_DEPTH_10:
            if (params->pix_enc == DRM_COLOR_FORMAT_YCRCB422)
                T1 = (2000 / 5) * link->lanes;
            else
                T1 = (4000 / 15) * link->lanes;
            break;
        case COLOR_DEPTH_12:
            if (params->pix_enc == DRM_COLOR_FORMAT_YCRCB422) {
                if (params->pixel_mode == DPTX_MP_DUAL_PIXEL)
                    T1 = (1000 / 6) * link->lanes;
                else
                    T1 = (1000 / 3) * link->lanes;
            } else {
                T1 = (2000 / 9) * link->lanes;
            }
            break;
        case COLOR_DEPTH_16:
            if (params->pix_enc != DRM_COLOR_FORMAT_YCRCB422 &&
                params->pixel_mode == DPTX_MP_DUAL_PIXEL)
                T1 = (1000 / 6) * link->lanes;
            else
                T1 = (1000 / 4) * link->lanes;
            break;
        default:
            return -EINVAL;
        }

        if (params->pix_enc == DRM_COLOR_FORMAT_YCRCB420)
            T2 = (link->rate / 4) * 1000 / (mode->clock / 2);
        else
            T2 = (link->rate / 4) * 1000 / mode->clock;

        if (tu_frac)
            params->init_threshold = T1 * T2 * (tu + 1) / (1000 * 1000);
        else
            params->init_threshold = T1 * T2 * tu / (1000 * 1000);

        if ((params->init_threshold <= 16) || (tu < 10))
            params->init_threshold = 40;
    }

    dptx_writel(dptx, DPTX_VIDEO_CONFIG5,
             FIELD_PREP(INIT_THRESHOLD_HI, params->init_threshold >> 6) |
             FIELD_PREP(AVERAGE_BYTES_PER_TU_FRAC, tu_frac) |
             FIELD_PREP(INIT_THRESHOLD, params->init_threshold) |
             FIELD_PREP(AVERAGE_BYTES_PER_TU, tu));

    return 0;
}

static int xdx_dptx_video_config(struct dp_dev *dptx)
{
    struct video_params *params = &dptx->vparams;
    struct drm_display_mode *mode = &dptx->dp_mode;
    struct dptx_link *link = &dptx->link;
    u32 h_active = mode->hdisplay;
    u32 h_blanking = mode->htotal - mode->hdisplay;
    u32 v_blanking = mode->vtotal - mode->vdisplay;
    u32 v_active = mode->vdisplay;
    u32 v_start = mode->vtotal - mode->vsync_start;
    u32 h_start = mode->htotal - mode->hsync_start;
    u32 h_sync_pulse_width = mode->hsync_end - mode->hsync_start;
    u32 h_sync_offset = mode->hsync_start - mode->hdisplay;
    u32 v_sync_pulse_width = mode->vsync_end - mode->vsync_start;
    u32 v_sync_offset = mode->vsync_start - mode->vdisplay;
    u32 val = 0, hblank_interval = 0;
    int ret = 0;
    u8 vic;

    dptx_mask_writel(dptx, DPTX_VSAMPLE_CTRL, PIXEL_MODE_SELECT,
               FIELD_PREP(PIXEL_MODE_SELECT, params->pixel_mode));

    xdx_dptx_video_set_sink_msa(dptx, v_start, h_start);

    dptx_mask_writel(dptx, DPTX_VSAMPLE_CTRL, VIDEO_MAPPING,
               FIELD_PREP(VIDEO_MAPPING, params->bpc_mapping));

    if (mode->flags & DRM_MODE_FLAG_PHSYNC)
        val |= HSYNC_IN_POLARITY;
    if (mode->flags & DRM_MODE_FLAG_PVSYNC)
        val |= VSYNC_IN_POLARITY;
    dptx_writel(dptx, DPTX_VINPUT_POLARITY_CTRL, val);

    val = FIELD_PREP(HACTIVE, h_active) | FIELD_PREP(HBLANK, h_blanking);
    vic = drm_match_cea_mode(mode);
    if (vic == 5 || vic == 6 || vic == 7 ||
        vic == 10 || vic == 11 || vic == 20 ||
        vic == 21 || vic == 22 || vic == 39 ||
        vic == 25 || vic == 26 || vic == 40 ||
        vic == 44 || vic == 45 || vic == 46 ||
        vic == 50 || vic == 51 || vic == 54 ||
        vic == 55 || vic == 58 || vic  == 59)
        val |= R_V_BLANK_IN_OSC;
    if (mode->flags & DRM_MODE_FLAG_INTERLACE)
        val |= FIELD_PREP(I_P, 1);
    dptx_writel(dptx, DPTX_VIDEO_CONFIG1, val);

    val = FIELD_PREP(VBLANK, v_blanking) | FIELD_PREP(VACTIVE, v_active);
    dptx_writel(dptx, DPTX_VIDEO_CONFIG2, val);

    val = FIELD_PREP(H_SYNC_WIDTH,
          h_sync_pulse_width) | FIELD_PREP(H_FRONT_PORCH, h_sync_offset);
    dptx_writel(dptx, DPTX_VIDEO_CONFIG3, val);

    val = FIELD_PREP(V_SYNC_WIDTH,
          v_sync_pulse_width) | FIELD_PREP(V_FRONT_PORCH, v_sync_offset);
    dptx_writel(dptx, DPTX_VIDEO_CONFIG4, val);

    ret = dptx_video_ts_calculate_change(dptx);

    hblank_interval = h_blanking * (link->rate / 4) / mode->clock;
    dptx_writel(dptx, DPTX_VIDEO_HBLANK_INTERVAL,
            FIELD_PREP(HBLANK_INTERVAL, hblank_interval));

    dptx_mask_writel(dptx, DPTX_VSAMPLE_CTRL,
            VIDEO_STREAM_ENABLE, VIDEO_STREAM_ENABLE);

    return ret;
}

static int xdx_dptx_pll_config(struct dp_dev *dptx)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int timeout = 1000;
    DPDBCmd_T_NOSINK PixClockConf;
    u32 val;

    PixClockConf.cmd = cDPCmdPLLConfig;
    PixClockConf.dcpixclock = dptx->dp_mode.clock;

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL,
                (u64)PixClockConf.all & 0xffffffff);

    xdx_mmio_writel(xdev, SYS_PANGU_REG_HDMIDOORBELL + 4,
                ((u64)PixClockConf.all >> 32) & 0xffffffff);
    for (; timeout; timeout--) {
            val = xdx_mmio_readl(xdev, SYS_PANGU_REG_HDMIDOORBELL);
            if(val & BIT(0))
                break;

            usleep_range(1000, 2000);
    }

    if (!(val & BIT(0))) {
        dev_err(xdev->dev,"dptx pll config fail\n");
        return -ETIMEDOUT;
    }

    return 0;
}

static void xdx_dp_trans_para_fill(struct dp_dev *dptx, struct dp_trans_para *para)
{
    struct drm_display_mode *mode = &dptx->dp_mode;
    struct video_params *vparams;

    vparams = &dptx->vparams;
    para->mode = drm_match_cea_mode(mode);
    para->bpc = vparams->bpc;
    para->multipixel = vparams->pixel_mode;
    para->rate = dptx->link.rate;
    para->dc_idx = dptx->id;
    para->clock = mode->clock;
    para->hdisplay = mode->hdisplay;
    para->htotal = mode->htotal;
    para->hsync_start = mode->hsync_start;
    para->hsync_end = mode->hsync_end;
    para->flag = mode->flags;
    para->vdisplay = mode->vdisplay;
    para->vtotal = mode->vtotal;
    para->vsync_start = mode->vsync_start;
    para->vsync_end = mode->vsync_end;
    para->picture_aspect_ratio = mode->picture_aspect_ratio;
    para->mode_flag_pvsync = DRM_MODE_FLAG_PVSYNC;

    dev_dbg(dptx->dev,"clock=%d, hdisplay=%d, htotal=%d, hsync_start=%d, hsync_end=%d, flag=%d, vdisplay=%d, vtotal=%d, vsync_start=%d, vsync_end=%d, picture_aspect_ratio=%d\n",
        para->clock, para->hdisplay, para->htotal, para->hsync_start, para->hsync_end, para->flag, para->vdisplay, para->vtotal,
        para->vsync_start, para->vsync_end, para->picture_aspect_ratio);
}

static void dp_config_err_analyze(struct dp_dev* xdev, u32 err_code)
{
       int i;
       const struct err_vector{
           u32 err_bit;
           const char *type;
       }err_vectors[] = {
           {BIT(0), "DP_WAIT_HPD_TIMEOUT"},
           {BIT(1), "DP_TS_CALCULATE_FAIL"},
           {BIT(2), "DP_CALCULATE_TU_FAIL"},
           {BIT(3), "DP_CALCULATE_BPC_INVALID"},
           {BIT(4), "DP_TRAINING_START_FAIL"},
           {BIT(5), "DP_CR_REDUCE_LANES_FAIL"},
           {BIT(6), "DP_REDECE_LANE_CR_DONE_FAIL"},
           {BIT(7), "DP_LINK_CR_FAIL"},
           {BIT(8), "DP_EQ_REDUCE_RATE_FAIL"},
           {BIT(9), "DP_EQ_REDUCE_LANES_FAIL"},
           {BIT(10), "DP_LINK_EQ_FAIL"},
           {BIT(11), "DP_LANES_RATE_INVALID"},
           {BIT(12), "DP_TO_P3_PHY_BUSY"},
           {BIT(13), "DP_SET_RATE_PHY_BUSY"},
           {BIT(14), "DP_TO_P0_PHY_BUSY"},
           {BIT(15), "DP_PATTERN_SET_FAIL"},
           {BIT(16), "DP_BW_SET_FAIL"},
           {BIT(17), "DP_LANE_COUNT_SET_FAIL"},
           {BIT(18), "DP_DOWNSPREAD_CRTL_FAIL"},
           {BIT(19), "DP_CHANNEL_CODING_FAIL"},
           {BIT(20), "DP_LANE0_SET_FAIL"},
           {BIT(21), "DP_READ_STATUS_FAIL"},
           {BIT(22), "DP_REQUEST_READ_FAIL"},
           {BIT(23), "DP_TRAINING_LANES_SET_FAIL"},
           {BIT(24), "DP_LINK_AJUST_FAIL"},
           {BIT(25), "DP_DETECT_DEVICE_FAIL"},
           {BIT(26), "DP_ENABLE_PLL_TIMEOUT"},
       };

       dev_err(xdev->dev,"DP config err: err_code = 0x%0x \n", err_code);
       for (i = 0; i < ARRAY_SIZE(err_vectors); i++){
           if(err_vectors[i].err_bit & err_code)
                   dev_err(xdev->dev,"DP CONFIG: ERROR TYPE = %s,   ERROR BIT = BIT(%d) \n", err_vectors[i].type, i);
       }
}

static int xdx_dptx_mode_setup_sink(struct dp_dev *dptx)
{
    struct dp_trans_para para;
    DPDBCmd_T_SINK PhyConfig;
    //u32 DBValue;
    int i, timeout = 1000;
    int ret;

	if (dptx->cts) {
	    dptx_enable_ssc(dptx);
	}

    if(dptx->output->psDoorbellShareMemDesc[2 + dptx->id] == NULL){
        ret = DPDoorbellShareMemAllocate(dptx->output,2 + dptx->id);
        if(ret){
            dev_err(dptx->dev,"DP share ram allocate fail\n");
            return ret;
        }
    }
    xdx_dp_trans_para_fill(dptx, &para);
    memcpy_toio(dptx->output->psDoorbellVirAddr[2 + dptx->id], &para, sizeof(struct dp_trans_para));
    PhyConfig.all = 0;
    PhyConfig.cmd = cH02ADisplayDBCmdDP;
    PhyConfig.subCmd = cH02ADPSubCmdCNTLConfig;
    PhyConfig.dword = dptx->output->psDoorbellPhyAddr[2 + dptx->id].uiAddr;
    xdx_dp_h2a_doorbell_w(dptx,(u64)PhyConfig.all);

    for (i = timeout; i; i--) {
        PhyConfig.all = xdx_dp_h2a_doorbell_r64(dptx);
        if (PhyConfig.low32 & BIT(0)){
	    if(PhyConfig.ret & BIT(0))
                dp_config_err_analyze(dptx, PhyConfig.dword);
            break;
	}
        usleep_range(1000, 2000);
    }

    if (!(PhyConfig.low32 & BIT(0))) {
        dev_err(dptx->dev,"xdx dptx tx mode set fail\n");
        return -ETIMEDOUT;
    }

    dptx->link_trained = true;
    return 0;
}

static int xdx_dptx_mode_setup(struct dp_dev *dptx)
{
    int ret;

    ret = xdx_dptx_tile_phy_init(dptx);
    if (ret < 0)
        return ret;

    ret = xdx_dptx_link_power_on(dptx);
    if (ret < 0)
        return ret;

    ret = xdx_dptx_pll_config(dptx);
    if (ret < 0)
        return ret;

    ret = xdx_dptx_link_training(dptx);
    if (ret < 0){
        dev_err(dptx->dev, "dptx link trainning error: %d\n", ret);
        return ret;
    }

    ret = xdx_dptx_video_config(dptx);
    if (ret < 0) {
        dev_err(dptx->dev, "dptx video config error: %d\n", ret);
        return ret;
    }

    //dptx_core_regs_dump(dptx);
    //dptx_tile_regs_dump(dptx);

    return ret;
}

static int xdx_dptx_mode_valid(struct drm_connector *connector,
                        struct drm_display_mode *mode)
{
    struct drm_device *dev = connector->dev;
    struct dp_dev *dptx = connector_to_dp(connector);
    struct dptx_link *link = &dptx->link;
    struct drm_display_info info = connector->display_info;
    u32 min_bpp;

    if (info.color_formats & DRM_COLOR_FORMAT_YCRCB420)
        min_bpp = 12;
    else if (info.color_formats & DRM_COLOR_FORMAT_YCRCB422)
        min_bpp = 16;
    else if (info.color_formats & DRM_COLOR_FORMAT_RGB444)
        min_bpp = 18;
    else
        min_bpp = 24;

    if((mode->clock < 16000) || ( mode->clock > 600000)) {
        dev_dbg(dev->dev, "mode:%dx%d vrefresh:%d pix_clock:%d pll-un-supported",
                           mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
        return MODE_ERROR;
    }

    if (!dptx_bandwidth_ok(dptx,mode->clock, min_bpp, link->lanes, link->rate)) {
        dev_dbg(dev->dev, "mode:%dx%d vrefresh:%d pix_clock:%d bw-un-supported",
                           mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);
        return MODE_CLOCK_HIGH;
    }
    dev_dbg(dev->dev, "mode:%dx%d vrefresh:%d pix_clock:%d    supported",
                       mode->hdisplay, mode->vdisplay, drm_mode_vrefresh(mode), mode->clock);

    return MODE_OK;
}

static struct drm_encoder *xdx_dptx_connector_best_encoder(
                struct drm_connector *connector)
{
    struct dp_dev *dptx = connector_to_dp(connector);

    return &dptx->encoder;
}

static enum drm_connector_status
xdx_dptx_helper_connector_detect(struct drm_connector *connector, bool force)
{
    struct dp_dev *dptx = connector_to_dp(connector);
    enum drm_connector_status status;
    u32 hpd_status;

    hpd_status = dptx_readl(dptx, DPTX_HPD_STATUS);
    if (hpd_status & (HPD_STATUS)) {
        status = connector_status_connected;
    } else {
        status = connector_status_disconnected;
    }

    if ((status == connector_status_connected) && dptx->encoder_is_on) {
        int ret;
        struct edid *edid;
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 7, 0))
        const struct drm_edid *drm_edid;
#endif
        struct dptx_link *link = &dptx->link;
        struct video_params *vparams = &dptx->vparams;

        mutex_lock(&dptx->mutex);

        ret = xdx_drm_dp_read_dpcd_caps(&dptx->aux, link->rx_caps);
        if (ret < 0) {
            mutex_unlock(&dptx->mutex);
            return connector_status_disconnected;
        }

        link->rate = drm_dp_max_link_rate(link->rx_caps);
        link->lanes = drm_dp_max_lane_count(link->rx_caps);
        if (!dptx->edid) {
            edid = drm_get_edid(connector, &dptx->aux.ddc);
        } else {
            edid = dptx->edid;
        }
        if (edid) {
            dptx->edid = edid;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 10, 0))
            vparams->downstream_max_bpc = drm_dp_downstream_max_bpc(link->rx_caps, dptx->downstream_ports);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(6, 7, 0))
            vparams->downstream_max_bpc = drm_dp_downstream_max_bpc(link->rx_caps, dptx->downstream_ports, edid);
#else
            drm_edid = drm_edid_read_ddc(connector, &dptx->aux.ddc);
            vparams->downstream_max_bpc = drm_dp_downstream_max_bpc(link->rx_caps, dptx->downstream_ports, drm_edid);
#endif
            if (vparams->downstream_max_bpc == 0) {
                dev_err(dptx->dev, "drm_dp_downstream_max_bpc error.\n");
            }
        }
        if (xdx_dptx_needs_link_retrain(dptx) && dptx->edid) {
            xdx_dptx_poweron(dptx);
        }
        mutex_unlock(&dptx->mutex);
    }

    return status;
}

static void xdx_dptx_helper_connector_force(struct drm_connector *connector)
{
    connector->status = connector_status_connected;
}

static void
xdx_dptx_helper_connector_destroy(struct drm_connector *connector)
{
    struct dp_dev *dptx = connector_to_dp(connector);
    WARN(dptx == NULL, "container dp_dev err %s", __FUNCTION__);
    /* Disable all interrupts */
    drm_connector_unregister(connector);
    drm_connector_cleanup(connector);
}

static int xdx_dptx_connector_get_modes(struct drm_connector *connector)
{
    struct dp_dev *dptx = connector_to_dp(connector);
    struct video_params *vparams = &dptx->vparams;
    struct dptx_link *link = &dptx->link;
    struct drm_display_mode *mode;
    struct edid *edid;
    u8 max_bpc = 10;
    u8 bpc;
    int count = 0, i, ret;

    mutex_lock(&dptx->mutex);

    if (!dptx->edid) {
        edid = drm_get_edid(connector, &dptx->aux.ddc);
        ret = xdx_drm_dp_read_dpcd_caps(&dptx->aux, link->rx_caps);
        if (ret < 0) {
            mutex_unlock(&dptx->mutex);
            return count;
        }

        link->rate = drm_dp_max_link_rate(link->rx_caps);
        link->lanes = drm_dp_max_lane_count(link->rx_caps);
    } else {
        edid = dptx->edid;
    }

    if (edid) {
        dptx->edid = edid;
        drm_connector_update_edid_property(connector, edid);
        count = drm_add_edid_modes(connector, edid);
    } else {
        for (i = 0; i < ARRAY_SIZE(dp_default_modes); i++) {
            const struct drm_display_mode *ptr =
                &dp_default_modes[i];

            mode = drm_mode_duplicate(connector->dev, ptr);
            if (mode) {
                if (!i)
                    mode->type = DRM_MODE_TYPE_PREFERRED;
                drm_mode_probed_add(connector, mode);
                count++;
            }
        }
    }
    dptx->display_info = connector->display_info;
    bpc = dptx->display_info.bpc;

    if (vparams->downstream_max_bpc)
        dptx->display_info.bpc = min3(bpc, vparams->downstream_max_bpc, max_bpc);
    else
        dptx->display_info.bpc = min_t(u8, bpc, max_bpc);

    mutex_unlock(&dptx->mutex);

    if (edid)
        dev_dbg(dptx->dev, "\nDP:Version: %d\nRevision: %d\nExtensions: %d\n# of modes: %d\n",
                dptx->edid->version, dptx->edid->revision,
                dptx->edid->extensions, count);

    return count;
}

static void xdx_dptx_encoder_mode_set(struct drm_encoder *encoder,
            struct drm_display_mode *mode,
            struct drm_display_mode *adjusted_mode)
{
    struct dp_dev *dptx = container_of(encoder, struct dp_dev, encoder);
    struct xdx_dm* dm = output_to_dm(dptx->output);
    struct xdx_device* xdev = dm_to_xdev(dm);
    struct drm_device *ddev = xdev_to_drm(xdev);
    struct drm_display_info display_info = dptx->display_info;
    struct drm_crtc* crtc;
    struct dm_crtc_state* dm_state;
    const struct dp_output_format *format = &possible_output[1];
    struct video_params *vparams = &dptx->vparams;
    struct dptx_link *link = &dptx->link;
    int i;

    drm_for_each_crtc (crtc, ddev)
    {
        if (drm_crtc_index(crtc) == drm_encoder_index(encoder))
            break;
    }

    dm_state = to_dm_crtc_state(crtc->state);

    mutex_lock(&dptx->mutex);

    vparams->pix_enc = format->pix_enc;
    vparams->bpc_mapping = format->bpc_mapping;
    vparams->bpc = format->bpc;
    vparams->bpp = format->bpp;

    for (i = 0; i < ARRAY_SIZE(possible_output); i++) {
        format = &possible_output[i];
        if (format->bpc > display_info.bpc)
            continue;

        if (!(format->pix_enc & display_info.color_formats))
            continue;

        if (!dptx_bandwidth_ok(dptx,
            mode->clock, format->bpp, link->lanes, link->rate))
            continue;

        vparams->pix_enc = format->pix_enc;
        vparams->bpc_mapping = format->bpc_mapping;
        vparams->bpc = format->bpc;
        vparams->bpp = format->bpp;
        break;
    }

    dm_state->output_fmt = vparams->pix_enc;
    dm_state->bpc = vparams->bpc;

    memcpy(&dptx->dp_mode, mode, sizeof(dptx->dp_mode));

    mutex_unlock(&dptx->mutex);
}

static void xdx_dptx_poweron(struct dp_dev *dptx)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);

    if (xdev->asic_type == CHIP_PANGU_B0)
        xdx_dptx_mode_setup_sink(dptx);
    else
        xdx_dptx_mode_setup(dptx);

#ifdef SUPPORT_AUDIO
    if (dptx->id == 0) {
        if (xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP, XDX_AUDIO_MODULE_CONNECT)) {
            dev_dbg(xdev->dev, "%s: Call audio DP plug error!\n", __func__);
        }
    } else if (dptx->id == 1) {
        if(xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP1, XDX_AUDIO_MODULE_CONNECT)) {
            dev_dbg(xdev->dev, "%s: Call audio DP plug error!\n", __func__);
        }
    }
#endif
}

static void xdx_dptx_poweroff(struct dp_dev *dptx)
{
#ifdef SUPPORT_AUDIO
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);
#endif
    xdx_dptx_phy_power_down(dptx);

#ifdef SUPPORT_AUDIO
    if (dptx->id == 0) {
        if (xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP, XDX_AUDIO_MODULE_DISCONNECT)) {
            dev_dbg(xdev->dev, "%s: Call audio DP unplug error!\n", __func__);
        }
    } else if (dptx->id == 1) {
        if (xdx_audio_drv_hotPlug(xdev, XDX_AUDIO_MODULE_DP1, XDX_AUDIO_MODULE_DISCONNECT)) {
            dev_dbg(xdev->dev, "%s: Call audio DP unplug error!\n", __func__);
        }
    }
#endif
}

static void xdx_dptx_encoder_dpms(struct drm_encoder *encoder, int mode)
{
    struct dp_dev *dptx = encoder_to_dp(encoder);

    mutex_lock(&dptx->mutex);

    if (mode != DRM_MODE_DPMS_ON) {
        xdx_dptx_poweroff(dptx);
        dptx->encoder_is_on = false;
    }
    else {
        xdx_dptx_poweron(dptx);
        dptx->encoder_is_on = true;
    }

    mutex_unlock(&dptx->mutex);
}

static void xdx_dptx_encoder_commit(struct drm_encoder *encoder)
{
    const struct drm_encoder_helper_funcs * helper = encoder->helper_private;

    helper->dpms(encoder, DRM_MODE_DPMS_ON);
}

static bool xdx_dptx_encoder_mode_fixup(struct drm_encoder *encoder,
            const struct drm_display_mode *mode,
            struct drm_display_mode *adjusted_mode)
{
    return true;
}

static void xdx_dptx_encoder_destroy(struct drm_encoder *encoder)
{
    drm_encoder_cleanup(encoder);
}

static void xdx_dptx_encoder_prepare(struct drm_encoder *encoder)
{
    const struct drm_encoder_helper_funcs * helper = encoder->helper_private;

    helper->dpms(encoder, DRM_MODE_DPMS_OFF);
}

static ssize_t xdx_dptx_debugfs_connector_read(struct file *f, char __user *buf, size_t size, loff_t *pos)
{
	struct drm_connector *connector = file_inode(f)->i_private;
	u32 len, buffer_size = 4096;
	char *str;
	size_t ret;

	str = kzalloc(buffer_size, GFP_KERNEL);
	if(str == NULL)
		return -ENOMEM;

	len = scnprintf(str, buffer_size,  "connector:%s\n",connector->name);
	ret = simple_read_from_buffer(buf, size ,pos, str, len);
	kfree(str);
	return ret;
}

static ssize_t xdx_dptx_debugfs_connector_write(struct file *f, const char __user *buf, size_t size, loff_t *pos)
{
	struct drm_connector *connector = file_inode(f)->i_private;
	struct dp_dev *dptx = container_of(connector, struct dp_dev, connector);
	ssize_t retval;
	char my_buffer[512];
	int temp = 0;

	retval = simple_write_to_buffer(my_buffer, size, pos, buf, 4);
	if (retval == 0) {
			return -retval;
	}
	sscanf(my_buffer, "%d", &temp);
	dptx->cts = temp;

	return retval;
}

static const struct file_operations xdx_dptx_connector_debugfs_fops = {
	.read = xdx_dptx_debugfs_connector_read,
	.write = xdx_dptx_debugfs_connector_write,
};

static int xdx_dptx_connector_late_register(struct drm_connector *connector)
{
	debugfs_create_file("dp_cts", 0444, connector->debugfs_entry, connector,
		            &xdx_dptx_connector_debugfs_fops);
    return 0;
}

static const struct drm_encoder_funcs xdx_dptx_encoder_funcs = {
    .destroy = xdx_dptx_encoder_destroy,
};

static const struct drm_encoder_helper_funcs xdx_dptx_encoder_helper_funcs = {
    .dpms = xdx_dptx_encoder_dpms,
    .mode_set = xdx_dptx_encoder_mode_set,
    .commit = xdx_dptx_encoder_commit,
    .mode_fixup = xdx_dptx_encoder_mode_fixup,
    .prepare = xdx_dptx_encoder_prepare,
};

static const struct drm_connector_funcs xdx_dptx_connector_funcs = {
    .dpms = drm_helper_connector_dpms,
    .reset = drm_atomic_helper_connector_reset,
    .detect = xdx_dptx_helper_connector_detect,
    .force = xdx_dptx_helper_connector_force,
    .fill_modes = drm_helper_probe_single_connector_modes,
    .destroy = xdx_dptx_helper_connector_destroy,
    //.set_property = xdx_hdmi_helper_connector_set_property,
    .atomic_duplicate_state = drm_atomic_helper_connector_duplicate_state,
    .atomic_destroy_state = drm_atomic_helper_connector_destroy_state,
	.late_register = xdx_dptx_connector_late_register,
};

static const struct drm_connector_helper_funcs xdx_dptx_connector_helper_funcs = {
    .get_modes = xdx_dptx_connector_get_modes,
    .mode_valid = xdx_dptx_mode_valid,
    /*
     * For atomic, don't set atomic_best_encoder or best_encoder. This will
     * cause the DRM core to fallback to drm_atomic_helper_best_encoder().
     * This is fine as we only have a single connector and encoder.
     */
    .best_encoder = xdx_dptx_connector_best_encoder,
};

static int xdx_dptx_register(struct dp_dev *dptx, u32 crtc_mask)
{
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct drm_device *ddev = xdev_to_drm(xdev);
    int ret;

    dptx->connector.polled = DRM_CONNECTOR_POLL_HPD;
    dptx->connector.interlace_allowed = 0;
    dptx->connector.doublescan_allowed = 0;

    // check if DPx is DP-VGA ?
    if(((xdev->fw_st_info.vgaDetectFlag == HDMI_VGA_DVI) && (dptx->id == 0)) ||
            ((xdev->fw_st_info.vgaDetectFlag == HDMI_VGA) && (dptx->id == 0)) ||
            ((xdev->fw_st_info.vgaDetectFlag == HDMI_DP_HDMI_VGA) && (dptx->id == 1)))
        drm_connector_init(ddev, &dptx->connector, &xdx_dptx_connector_funcs,
                   DRM_MODE_CONNECTOR_VGA);
    else
        drm_connector_init(ddev, &dptx->connector, &xdx_dptx_connector_funcs,
                   DRM_MODE_CONNECTOR_DisplayPort);

    drm_connector_helper_add(&dptx->connector,
                 &xdx_dptx_connector_helper_funcs);


    if (!dptx->id) {
        dptx->encoder.possible_clones = 1 << 1;
        dptx->encoder.possible_crtcs = BIT(1);
    } else {
        dptx->encoder.possible_clones = 1 << 3;
        dptx->encoder.possible_crtcs = BIT(3);
    }

    ret = drm_encoder_init(ddev, &dptx->encoder, &xdx_dptx_encoder_funcs,
                    DRM_MODE_ENCODER_TMDS, NULL);
    if (ret) {
        dev_dbg(xdev->dev, "failed to init drm encoder\n");
        return ret;
    }

    drm_encoder_helper_add(&dptx->encoder, &xdx_dptx_encoder_helper_funcs);
    dev_dbg(xdev->dev, "DP_Connector_ID=%d\n",dptx->connector.base.id);
    ret = drm_connector_attach_encoder(&dptx->connector, &dptx->encoder);
    if (ret) {
        dev_err(xdev->dev,
            "can't attach [ENCODER:%d:%s] to [CONNECTOR:%d:%s] (err=%d)\n",
            dptx->encoder.base.id, dptx->encoder.name,
            dptx->connector.base.id, dptx->connector.name, ret);
    }

    return 0;
}

static int dptx_handle_aux_reply(struct dp_dev *dptx, u32 *auxsts)
{
    u32 status;
    int ret = -ETIMEDOUT, timeout = 500;

    for (timeout = 500; timeout; timeout--) {
        *auxsts = dptx_readl(dptx, DPTX_AUX_STATUS);

        if (!(*auxsts & AUX_RECEIVED)) {
            ret = 0;
            break;
        }

		udelay(10);
    };

    *auxsts = dptx_readl(dptx, DPTX_AUX_STATUS);
    status = FIELD_GET(AUX_STATUS, *auxsts);

    if (*auxsts & AUX_REPLY_ERR)
        dev_dbg(dptx->dev, "%s: AUX_REPLY_ERR\n", __func__);
    if (*auxsts & AUX_TIMEOUT)
        dev_dbg(dptx->dev, "%s: AUX_TIMEOUT\n", __func__);
    if (*auxsts & AUX_SINK_DWA)
        dev_dbg(dptx->dev, "%s: AUX_SINK_DWA\n", __func__);

    switch (status) {
    case AUX_STATUS_ACK:
        dev_dbg(dptx->dev, "%s: AUX_STATUS_ACK\n", __func__);
        break;
    case AUX_STATUS_NACK:
        dev_dbg(dptx->dev, "%s: AUX_STATUS_NACK\n", __func__);
        break;
    case AUX_STATUS_DEFER:
        dev_dbg(dptx->dev, "%s: AUX_STATUS_DEFER\n", __func__);
        break;
    case AUX_STATUS_I2C_NACK:
        dev_dbg(dptx->dev, "%s: AUX_STATUS_I2C_NACK\n",
            __func__);
        break;
    case AUX_STATUS_I2C_DEFER:
        dev_dbg(dptx->dev, "%s: AUX_STATUS_I2C_DEFER\n",
            __func__);
        break;
    default:
        dev_dbg(dptx->dev, "Invalid AUX status 0x%x\n", status);
        break;
    }

    return ret;
}

static int dptx_aux_write_data(struct dp_dev *dptx,
        const u8 *buffer, size_t size)
{
    size_t i, j;

    for (i = 0; i < DIV_ROUND_UP(size, 4); i++) {
        size_t num = min_t(size_t, size - i * 4, 4);
        u32 value = 0;

        for (j = 0; j < num; j++)
            value |= buffer[i * 4 + j] << (j * 8);

        dptx_writel(dptx, DPTX_AUX_DATA0 + i * 4, value);
    }

    return size;
}

static int dptx_aux_read_data(struct dp_dev *dptx, u8 *buffer, size_t size)
{
    size_t i, j;

    for (i = 0; i < DIV_ROUND_UP(size, 4); i++) {
        size_t num = min_t(size_t, size - i * 4, 4);
        u32 value;

        value = dptx_readl(dptx, DPTX_AUX_DATA0 + i * 4);

        for (j = 0; j < num; j++)
            buffer[i * 4 + j] = value >> (j * 8);
    }

    return size;
}

void dptx_soft_reset(struct dp_dev *dptx, u32 bits)
{
    u32 rst;

    bits &= (DPTX_SRST_CTRL_ALL);

    /* Set reset bits */
    rst = dptx_readl(dptx, DPTX_SOFT_RESET_CTRL);
    rst |= bits;
    dptx_writel(dptx, DPTX_SOFT_RESET_CTRL, rst);

    usleep_range(10, 20);

    /* Clear reset bits */
    rst = dptx_readl(dptx, DPTX_SOFT_RESET_CTRL);
    rst &= ~bits;
    dptx_writel(dptx, DPTX_SOFT_RESET_CTRL, rst);
}


static ssize_t xdx_dptx_aux_transfer(struct drm_dp_aux *aux,
                  struct drm_dp_aux_msg *msg)
{
    struct dp_dev *dptx = container_of(aux, struct dp_dev, aux);
    u32 status, value;
    ssize_t ret = 0;

    if (WARN_ON(msg->size > 16))
        return -E2BIG;

    switch (msg->request & ~DP_AUX_I2C_MOT) {
    case DP_AUX_NATIVE_WRITE:
    case DP_AUX_I2C_WRITE:
    case DP_AUX_I2C_WRITE_STATUS_UPDATE:
        ret = dptx_aux_write_data(dptx, msg->buffer, msg->size);
        if (ret < 0)
            return ret;
        break;
    case DP_AUX_NATIVE_READ:
    case DP_AUX_I2C_READ:
        break;
    default:
        return -EINVAL;
    }

    if (msg->size > 0)
        value = FIELD_PREP(AUX_LEN_REQ, msg->size - 1);
    else
        value = FIELD_PREP(I2C_ADDR_ONLY, 1);
    value |= FIELD_PREP(AUX_CMD_TYPE, msg->request);
    value |= FIELD_PREP(AUX_ADDR, msg->address);
    dptx_writel(dptx, DPTX_AUX_CMD, value);

    status = dptx_handle_aux_reply(dptx, &value);
    if ((status < 0) || (value & AUX_TIMEOUT)) {
        dev_err(dptx->dev, "timeout waiting for AUX reply\n");
        return -ETIMEDOUT;
    }

    msg->reply = FIELD_GET(AUX_STATUS, value);

    if (msg->size > 0 && msg->reply == DP_AUX_NATIVE_REPLY_ACK) {
        if (msg->request & DP_AUX_I2C_READ) {
            size_t count = FIELD_GET(AUX_BYTES_READ, value) - 1;

            if (count != msg->size) {
                dptx_soft_reset(dptx, AUX_RESET);
                return -EBUSY;
            }

        ret = dptx_aux_read_data(dptx, msg->buffer, count);
            if (ret < 0)
                return ret;
        }
    }

    return ret;
}

static void xdx_dp_hpd_work(struct work_struct *work)
{
    struct dp_dev *dptx = container_of(work, struct dp_dev, hpd_work);
    struct xdx_dm *dm = output_to_dm(dptx->output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct drm_device *ddev = xdev_to_drm(xdev);
    struct dptx_link *link = &dptx->link;
    int ret;
    u32 hpdsts;

    dev_dbg(dptx->dev, "dptx get hpd %s irq\n",
            dptx->long_hpd ? "long" : "short");

    hpdsts = dptx_readl(dptx, DPTX_HPD_STATUS);

    mutex_lock(&dptx->mutex);

    if (hpdsts & HPD_STATUS) {
        if (xdx_drm_dp_read_dpcd_caps(&dptx->aux, link->rx_caps) < 0) {
            dev_dbg(dptx->dev, "%s: drm dptx read dpcd caps failed\n", __func__);
            mutex_unlock(&dptx->mutex);
            return;
        }

        link->rate = drm_dp_max_link_rate(link->rx_caps);
        link->lanes = drm_dp_max_lane_count(link->rx_caps);

        if (!dptx->long_hpd) {
            if (xdx_dptx_needs_link_retrain(dptx) && dptx->edid) {
                ret = xdx_dptx_link_training(dptx);
                if (ret < 0)
                    goto out;

                ret = dptx_video_ts_calculate_change(dptx);
                if (ret < 0)
                    goto out;
            }
        }
    } else {
        xdx_dptx_phy_power_down(dptx);

	if(dptx->edid) {
            kfree(dptx->edid);
            dptx->edid = NULL;
        }
    }

out:
    mutex_unlock(&dptx->mutex);

    drm_helper_hpd_irq_event(ddev);
}


/**
* dptx_sink_enabled_ssc() - Returns true, if sink is enabled ssc
* @dptx: The dptx struct
*
*/
static bool dptx_sink_enabled_ssc(struct dp_dev *dptx)
{
    u8 byte;

    dptx_read_dpcd(dptx, DP_MAX_DOWNSPREAD, &byte);

    return byte & 1;
}

static int dptx_core_program_ssc(struct dp_dev *dptx, bool sink_ssc)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 tile_data;
    u32 phyifctrl;
    u8  retval;

    // tile
    tile_data = dptx_tile_readl(dptx, DP_PCS_CTRL);
    dev_dbg(xdev->dev, "tile DP_PCS_CTRL: 0x%x\n", tile_data);

    /* Enable 4 lanes, before programming SSC */
    dptx_phy_set_lanes(dptx, 4);

    // Move PHY to P3 to program SSC
    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl |= (3 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

    //retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
        dev_err(xdev->dev, "timed out waiting for phy busy\n");
        return retval;
    }

    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    if (dptx->ssc_en && sink_ssc) {
        dev_dbg(xdev->dev, "enable ssc\n");
        phyifctrl &= ~DPTX_PHYIF_CTRL_SSC_DIS;
        tile_data |= 0x8;
    } else {
        dev_dbg(xdev->dev, "disable ssc\n");
        phyifctrl |= DPTX_PHYIF_CTRL_SSC_DIS;
        tile_data &= ~0x8;
    }

    dev_dbg(xdev->dev, "write 0x%x to DP_PCS_CTRL\n", tile_data);
    dptx_tile_writel(dptx, DP_PCS_CTRL, tile_data);

    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

    //retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
            dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
            return retval;
     }

    // Move PHY to P0 to program SSC
    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl |= (0 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P0 state
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

    //retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
        dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
        return retval;
    }

    dev_dbg(xdev->dev, "SSC program done\n");
    return 0;
}

/**
* dptx_enable_ssc() - Enables SSC based on automation request,
*             if DPTX controller enables ssc
* @dptx: The dptx struct
*
*/
void dptx_enable_ssc(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    bool sink_ssc = dptx_sink_enabled_ssc(dptx);

    if (sink_ssc) {
        dev_dbg(xdev->dev, "%s: SSC enable on the sink side\n", __func__);
        dptx->ssc_en = true;
    } else {
        dev_dbg(xdev->dev, "%s: SSC disabled on the sink side\n", __func__);
        dptx->ssc_en = false;
    }
    dptx_core_program_ssc(dptx, sink_ssc);
}



static void dptx_phy_set_pre_emphasis(struct dp_dev *dptx, unsigned int lane, unsigned int level)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 phytxeq;

    dev_dbg(xdev->dev, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

    if (WARN(lane > 3, "Invalid lane %d", lane))
        return;

    if (WARN(level > 3, "Invalid pre-emphasis level %d, using 3", level))
        level = 3;

    phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
    phytxeq &= ~DPTX_PHY_TX_EQ_PREEMP_MASK(lane);
    phytxeq |= (level << DPTX_PHY_TX_EQ_PREEMP_SHIFT(lane)) &
                DPTX_PHY_TX_EQ_PREEMP_MASK(lane);

    dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}

static void dptx_phy_set_vswing(struct dp_dev *dptx, unsigned int lane, unsigned int level)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 phytxeq;

    dev_dbg(xdev->dev, "%s: lane=%d, level=0x%x\n", __func__, lane, level);

    if (WARN(lane > 3, "Invalid lane %d", lane))
        return;

    if (WARN(level > 3, "Invalid vswing level %d, using 3", level))
        level = 3;

    phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
    phytxeq &= ~DPTX_PHY_TX_EQ_VSWING_MASK(lane);
    phytxeq |= (level << DPTX_PHY_TX_EQ_VSWING_SHIFT(lane)) &
                DPTX_PHY_TX_EQ_VSWING_MASK(lane);

    dptx_writel(dptx, DPTX_PHY_TX_EQ, phytxeq);
}

static void dptx_phy_set_pattern(struct dp_dev *dptx, unsigned int pattern)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 phyifctrl = 0;

    dev_dbg(xdev->dev, "%s: Setting PHY pattern=0x%x\n", __func__, pattern);

    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl &= ~DPTX_PHYIF_CTRL_TPS_SEL_MASK;
    phyifctrl |= ((pattern << DPTX_PHYIF_CTRL_TPS_SEL_SHIFT) &
                DPTX_PHYIF_CTRL_TPS_SEL_MASK);
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

static void dptx_phy_set_rate(struct dp_dev *dptx, unsigned int rate)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 phyifctrl;

    dev_dbg(xdev->dev, "%s: rate=%d\n", __func__, rate);

    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);

//#ifdef DPTX_COMBO_PHY
//  switch (rate) {
//  case DPTX_PHYIF_CTRL_RATE_RBR:
//  case DPTX_PHYIF_CTRL_RATE_HBR:
//      /* Set 20-bit PHY width */
//      phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
//      phyifctrl &= ~DPTX_PHYIF_CTRL_WIDTH;
//      dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
//      break;
//  case DPTX_PHYIF_CTRL_RATE_HBR2:
//  case DPTX_PHYIF_CTRL_RATE_HBR3:
//      /* Set 40-bit PHY width */
//      phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
//      phyifctrl |= DPTX_PHYIF_CTRL_WIDTH;
//      dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
//      break;
//  default:
//      WARN(1, "Invalid PHY rate %d\n", rate);
//      break;
//  }
//#endif

    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl &= ~DPTX_PHYIF_CTRL_RATE_MASK;
    phyifctrl |= rate << DPTX_PHYIF_CTRL_RATE_SHIFT;
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);


    //tile register phy mode config
    //phyifctrl = dptx_tilereadl(dptx,DP_PCS_CTRL);
    //phyifctrl = (phyifctrl | (rate << 6));
    //dptx_tilewritel(dptx,DP_PCS_CTRL,phyifctrl);
}

void dptx_phy_set_lanes(struct dp_dev *dptx, unsigned int lanes)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u32 phyifctrl;
    u32 val;

    dev_dbg(xdev->dev, "%s: lanes=%d\n", __func__, lanes);

    switch (lanes) {
    case 1:
        val = 0;
        break;
    case 2:
        val = 1;
        break;
    case 4:
        val = 2;
        break;
    default:
        WARN(1, "Invalid number of lanes %d\n", lanes);
        return;
    }

    phyifctrl = 0;
    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl &= ~DPTX_PHYIF_CTRL_LANES_MASK;
    phyifctrl |= (val << DPTX_PHYIF_CTRL_LANES_SHIFT);
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);
}

static int dptx_link_read_status(struct dp_dev *dptx)
{
	return dptx_read_bytes_from_dpcd(dptx, DP_LANE0_1_STATUS,
                  dptx->link.status,
                  // DP_LINK_STATUS_SIZE);
                   4);
}

static int dptx_link_check_status(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    u8 bytes[2];

    retval = dptx_read_bytes_from_dpcd(dptx, DP_SINK_COUNT,
                        bytes, 2);
    if (retval)
        return retval;

    retval = dptx_link_read_status(dptx);
    if (retval)
        return retval;

    /* Check if need to retrain link */
    if ((!drm_dp_channel_eq_ok(dptx->link.status, dptx->link.lanes) ||
        !drm_dp_clock_recovery_ok(dptx->link.status, dptx->link.lanes))) {
        dev_dbg(xdev->dev, "%s: Retraining link\n", __func__);
        return true;
    }

    return false;
}

static int dptx_set_custom_pattern(struct dp_dev *dptx)
{
    int retval;
    u8 pattern0 , pattern1, pattern2, pattern3, pattern4, pattern5,
       pattern6, pattern7, pattern8, pattern9;

    u32 custompat0;
    u32 custompat1;
    u32 custompat2;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_0, &pattern0);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_1, &pattern1);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_2, &pattern2);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_3, &pattern3);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_4, &pattern4);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_5, &pattern5);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_6, &pattern6);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_7, &pattern7);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_8, &pattern8);
    if (retval)
        return retval;

    retval = dptx_read_dpcd(dptx, DP_TEST_80BIT_CUSTOM_PATTERN_9, &pattern9);
    if (retval)
        return retval;

    /*
     *  Calculate 30,30 and 20 bits custom patterns depending on TEST_80BIT_CUSTOM_PATTERN sequence
     */
    custompat0 = ((((((pattern3 & (0xff >> 2)) << 8) | pattern2) << 8) | pattern1) << 8) | pattern0;
    custompat1 = ((((((((pattern7 & (0xf)) << 8) | pattern6) << 8) | pattern5) << 8) | pattern4) << 2) | ((pattern3 >> 6) & 0x3);
    custompat2 = (((pattern9 << 8) | pattern8) << 4) | ((pattern7 >> 4) & 0xf);

    dptx_writel(dptx, DPTX_CUSTOMPAT0, custompat0);
    dptx_writel(dptx, DPTX_CUSTOMPAT1, custompat1);
    dptx_writel(dptx, DPTX_CUSTOMPAT2, custompat2);

    return 0;
}

static void adjust_pcs_tx_eq(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    int i;
    u32 test_link_rate = 0x219;
    u8 rate;
    u8 lane_01;
    u8 lane_23;
    u32 reg;
    u32 read_val = 0;
    u16 val_0_2 = 0;
    u16 val_1_3 = 0;
	int j;

    dev_dbg(xdev->dev, "%s()\n", __func__);

    // get rate
    dptx_read_dpcd(dptx, test_link_rate, &rate);
    dev_dbg(xdev->dev, "rate: 0x%x\n", rate);


#if 0
    if (rate >= 0x14) {
        dev_dbg(xdev->dev, "HBR2 or HBR3\n");
        return;
    }

    dev_dbg(xdev->dev, "RBR or HBR1\n");
#endif

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE0_1, &lane_01);
    if (retval)
        return;

    dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE0_1: 0x%x\n", lane_01);

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE2_3, &lane_23);
    if (retval)
        return;

    dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE2_3: 0x%x\n", lane_23);

    // pcs global register
    dptx_pcs_writel(dptx, 0xa, 0x400);

	// loop number should be real test lane number
    for (i = 0; i < dptx->link.lanes; i++) {
        u8 pe = 0;
        u8 vs = 0;

        switch (i) {
        case 0:
            pe = (lane_01 & DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_01 & DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            break;
        case 1:
            pe = (lane_01 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_01 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            break;
        case 2:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
            break;
        case 3:
            pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
                >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
            vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
                >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
            break;
        default:
            break;
        }

		dev_dbg(xdev->dev, "vs: 0x%x\n", vs);
		dev_dbg(xdev->dev, "pe: 0x%x\n", pe);

		// RBR or HBR
		if (rate <= 0xa) {
			if (0x0 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x70c;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x70e;
			        val_1_3 = 0x8;
			    }
			    if (0x2 == pe) {
					val_0_2 = 0x710;
			        val_1_3 = 0x10;
			    }
			    if (0x3 == pe) {
					val_0_2 = 0x714;
			        val_1_3 = 0x20;
			    }
			}

			if (0x1 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x710;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x713;
			        val_1_3 = 0xc;
			    }
			    if (0x2 == pe) {
					val_0_2 = 0x716;
			        val_1_3 = 0x18;
			    }
			}

			if (0x2 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x716;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x718;
			        val_1_3 = 0x10;
			    }
			}

			if (0x3 == vs) {
				if (0x0 == pe) {
					val_0_2 = 0x718;
					val_1_3 = 0;
				}
			}
		} else {
			if (0x0 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x708;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x70a;
			        val_1_3 = 0x8;
			    }
			    if (0x2 == pe) {
					val_0_2 = 0x70c;
			        val_1_3 = 0x10;
			    }
			    if (0x3 == pe) {
					val_0_2 = 0x710;
			        val_1_3 = 0x20;
			    }
			}

			if (0x1 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x70c;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x70f;
			        val_1_3 = 0xc;
			    }
			    if (0x2 == pe) {
					val_0_2 = 0x712;
			        val_1_3 = 0x18;
			    }
			}

			if (0x2 == vs) {
			    if (0x0 == pe) {
					val_0_2 = 0x710;
			        val_1_3 = 0x0;
			    }
			    if (0x1 == pe) {
					val_0_2 = 0x714;
			        val_1_3 = 0x10;
			    }
			}

			if (0x3 == vs) {
				if (0x0 == pe) {
					val_0_2 = 0x718;
					val_1_3 = 0;
				}
			}
		}

		dev_dbg(xdev->dev, "val_0_2: 0x%x\n", val_0_2);
		dev_dbg(xdev->dev, "val_1_3: 0x%x\n", val_1_3);

#if 0
        // set pcs eq
        reg = 0x5000 + i*0x100;
        dptx_pcs_writel(dptx, (reg+0), val_0_2);
        read_val = dptx_pcs_readl(dptx, (reg+0));
        dev_dbg(xdev->dev, "pcs write test: 0x%x\n", read_val);

        dptx_pcs_writel(dptx, (reg+1), val_1_3);
        dptx_pcs_writel(dptx, (reg+2), val_0_2);
        dptx_pcs_writel(dptx, (reg+3), val_1_3);
        dptx_pcs_writel(dptx, (reg+4), val_0_2);
        dptx_pcs_writel(dptx, (reg+5), val_1_3);
        dptx_pcs_writel(dptx, (reg+6), val_0_2);
        dptx_pcs_writel(dptx, (reg+7), val_1_3);
#endif

		reg = 0x5000+ i*0x100;

		// 1. read 5N0X
		for (j = 0; j < LOOP; j++) {
			read_val = dptx_pcs_readl(dptx, (reg+j));
			dev_dbg(xdev->dev, "read pcs, reg: 0x%x, val: 0x%x", reg+j, read_val);
		}

		// 2. write 5N0X bit10
		for (j = 0; j < LOOP; j = j +2) {
			dptx_pcs_writel(dptx, (reg+j), BIT(10));
		}

		// 3. write 5N0X
		dev_dbg(xdev->dev, "write pcs");
		dptx_pcs_writel(dptx, (reg + 0), val_0_2);
		dptx_pcs_writel(dptx, (reg + 1), val_1_3);
		dptx_pcs_writel(dptx, (reg + 2), val_0_2);
		dptx_pcs_writel(dptx, (reg + 3), val_1_3);
		dptx_pcs_writel(dptx, (reg + 4), val_0_2);
		dptx_pcs_writel(dptx, (reg + 5), val_1_3);
		dptx_pcs_writel(dptx, (reg + 6), val_0_2);
		dptx_pcs_writel(dptx, (reg + 7), val_1_3);
		dptx_pcs_writel(dptx, (reg + 8), val_0_2);
		dptx_pcs_writel(dptx, (reg + 9), val_1_3);
		dptx_pcs_writel(dptx, (reg + 0xa), val_0_2);
		dptx_pcs_writel(dptx, (reg + 0xb), val_1_3);
		dptx_pcs_writel(dptx, (reg + 0xc), val_0_2);
		dptx_pcs_writel(dptx, (reg + 0xd), val_1_3);
		dptx_pcs_writel(dptx, (reg + 0xe), val_0_2);
		dptx_pcs_writel(dptx, (reg + 0xf), val_1_3);

		// 4. read 5N0X again
		for (j = 0; j < LOOP; j++) {
			read_val = dptx_pcs_readl(dptx, (reg+j));
			dev_dbg(xdev->dev, "read pcs after write, reg: 0x%x, val: 0x%x", reg+j, read_val);
		}

        // set pcs deskew enable
        reg = 0x1010 + i*0x100;
        dptx_pcs_writel(dptx, reg, 0xc);
    }
}

static int adjust_vswing_and_preemphasis(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    int i;
    u8 lane_01;
    u8 lane_23;
    u8 post_cursor2;
    u32 phytxeq;

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE0_1, &lane_01);
    if (retval) {
        return retval;
    }

    dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE0_1: 0x%x\n", lane_01);

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_LANE2_3, &lane_23);
    if (retval) {
        return retval;
    }
    dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_LANE2_3: 0x%x\n", lane_23);

    retval = dptx_read_dpcd(dptx, DP_ADJUST_REQUEST_POST_CURSOR2, &post_cursor2);
    if (retval) {
        return retval;
    }
    dev_dbg(xdev->dev, "DP_ADJUST_REQUEST_POST_CURSOR2: 0x%x\n", post_cursor2);

    dev_dbg(xdev->dev, "%s(), lanes: %d\n", __func__, dptx->link.lanes);
    for (i = 0; i < dptx->link.lanes; i++) {
	    u8 pe = 0;
	    u8 vs = 0;
	    //u8 cursor2;

	    switch (i) {
	    case 0:
	        pe = (lane_01 &  DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
	            >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
	        vs = (lane_01 &  DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
	            >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
	        //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE0_MASK) >> DP_ADJUST_POST_CURSOR2_LANE0_SHIFT;
	        break;
	    case 1:
	        pe = (lane_01 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
	            >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
	        vs = (lane_01 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
	            >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
	        //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE1_MASK) >> DP_ADJUST_POST_CURSOR2_LANE1_SHIFT;
	        break;
	    case 2:
	        pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE0_MASK)
	            >> DP_ADJUST_PRE_EMPHASIS_LANE0_SHIFT;
	        vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE0_MASK)
	            >> DP_ADJUST_VOLTAGE_SWING_LANE0_SHIFT;
	        //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE2_MASK) >> DP_ADJUST_POST_CURSOR2_LANE2_SHIFT;
	        break;
	    case 3:
	        pe = (lane_23 & DP_ADJUST_PRE_EMPHASIS_LANE1_MASK)
	            >> DP_ADJUST_PRE_EMPHASIS_LANE1_SHIFT;
	        vs = (lane_23 & DP_ADJUST_VOLTAGE_SWING_LANE1_MASK)
	            >> DP_ADJUST_VOLTAGE_SWING_LANE1_SHIFT;
	        //cursor2 = (post_cursor2 & DP_ADJUST_POST_CURSOR2_LANE3_MASK) >> DP_ADJUST_POST_CURSOR2_LANE3_SHIFT;
	        break;
	    default:
	        break;
	    }

	    dptx_phy_set_pre_emphasis(dptx, i, pe);
	    dptx_phy_set_vswing(dptx, i, vs);
	    //dptx_phy_set_post_cursor2(dptx, i, cursor2);
	}

	phytxeq = dptx_readl(dptx, DPTX_PHY_TX_EQ);
	dev_dbg(xdev->dev, "PHY_TX_EQ: 0x%x\n", phytxeq);

	return 0;
}

static int handle_test_phy_pattern(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    u8 pattern;
    u32 test_link_rate = 0x219;
    u32 test_lane_count = 0x220;
    u8 rate1, lanes;
    u8 rate;
    //u8 downspread;
    u8 rev;
    u32 phyifctrl;
    int retval;

    // get pattern
    //retval = dptx_read_dpcd(dptx, DP_TEST_PHY_PATTERN, &pattern);
    retval = dptx_read_dpcd(dptx, 0x248, &pattern);
    if (retval)
        return retval;

    pattern &= DP_TEST_PHY_PATTERN_SEL_MASK;
    dev_dbg(xdev->dev, "PATTERN: %d\n", pattern);

    // get dpcd version
    dptx_read_dpcd(dptx, 0x0, &rev);
    dev_dbg(xdev->dev, "DPCD_REV: 0x%x\n", rev);

    // get lane
    dptx_read_dpcd(dptx, test_lane_count, &lanes);
    lanes = lanes & 0x1f;
    dev_dbg(xdev->dev, "lanes: %d\n", lanes);
	dptx->link.lanes = lanes;

    // get rate
    dptx_read_dpcd(dptx, test_link_rate, &rate1);
    switch (rate1) {
    // 1.62Gbps
    case 0x6:
        rate = DPTX_PHYIF_CTRL_RATE_RBR;
        break;

    // 2.7Gbps
    case 0xa:
        rate = DPTX_PHYIF_CTRL_RATE_HBR;
        break;

    // 5.4Gbps
    case 0x14:
        rate = DPTX_PHYIF_CTRL_RATE_HBR2;
        break;

    // 8.1Gbps
    case 0x1e:
        rate = DPTX_PHYIF_CTRL_RATE_HBR3;
        break;
    default:
    rate = DPTX_PHYIF_CTRL_RATE_RBR;
    dev_dbg(xdev->dev, "invalid dp rate\n");

    }

    dev_dbg(xdev->dev, "rate: %d\n", rate);

    // set pattern
    switch (pattern) {
    case DP_TEST_PHY_PATTERN_NONE:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "No test pattern selected\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_NONE);
        break;

    case DP_TEST_PHY_PATTERN_D10:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "D10.2 without scrambling test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_1);
        break;

    case DP_TEST_PHY_PATTERN_SEMC:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "Symbol error measurement count test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_SYM_ERM);
        break;

    case DP_TEST_PHY_PATTERN_PRBS7:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "PRBS7 test phy pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_PRBS7);
        break;

    case DP_TEST_PHY_PATTERN_CUSTOM:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "80-bit custom pattern transmitted test phy pattern\n");
        //dptx_writel(dptx, DPTX_CUSTOMPAT0, 0x3E0F83E0);
        //dptx_writel(dptx, DPTX_CUSTOMPAT1, 0x3E0F83E0);
        //dptx_writel(dptx, DPTX_CUSTOMPAT2, 0xF83E0);
        retval = dptx_set_custom_pattern(dptx);
        if (retval)
            return retval;
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CUSTOM80);
        break;

    case DP_TEST_PHY_PATTERN_CP2520_1:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "CP2520_1 - HBR2 Compliance EYE pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CP2520_1);
        break;

	case DP_TEST_PHY_PATTERN_CP2520_2:
        adjust_pcs_tx_eq(dptx);
        retval = adjust_vswing_and_preemphasis(dptx);
        if (retval)
            return retval;
        dev_dbg(xdev->dev, "CP2520_2 - pattern\n");
        dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_CP2520_2);
        break;

	case DP_TEST_PHY_PATTERN_CP2520_3_TPS4:
		adjust_pcs_tx_eq(dptx);
		retval = adjust_vswing_and_preemphasis(dptx);
		if (retval)
			return retval;
		dev_dbg(xdev->dev, "DP_TEST_PHY_PATTERN_CP2520_3_TPS4 - pattern\n");
		dptx_phy_set_pattern(dptx, DPTX_PHYIF_CTRL_TPS_4);
		break;

	default:
		dev_dbg(xdev->dev, "Invalid TEST_PHY_PATTERN\n");
	    return -EINVAL;
    }

    // Move PHY to P3
    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl |= (2 << DPTX_PHYIF_CTRL_LANE_PWRDOWN_SHIFT);// move phy to P3 state
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

    //retval = dptx_phy_wait_busy(dptx, DPTX_MAX_LINK_LANES);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
        dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
        return retval;
    }

    // set rate
    dptx_phy_set_rate(dptx, rate);

    //retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
        dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
        return retval;
    }

    // set lanes
    dptx_phy_set_lanes(dptx, lanes);

    // Move phy to P0
    phyifctrl = dptx_readl(dptx, DPTX_PHYIF_CTRL);
    phyifctrl &= ~DPTX_PHYIF_CTRL_LANE_PWRDOWN_MASK;// P0
    dptx_writel(dptx, DPTX_PHYIF_CTRL, phyifctrl);

    /* Wait for PHY busy */
    //retval = dptx_phy_wait_busy(dptx, dptx->link.lanes);
    retval = dptx_phy_wait_busy(dptx);
    if (retval) {
        dev_err(xdev->dev, "Timed out waiting for PHY BUSY\n");
        return retval;
    }

    dptx_phy_enable_xmit(dptx, lanes, true);

    return retval;
}

static int handle_automated_test_request(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    u8 test;

    retval = dptx_read_dpcd(dptx, DP_TEST_REQUEST, &test);
    if (retval)
        return retval;

    if (test & DP_TEST_LINK_TRAINING) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_TRAINING\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_training(dptx);
        if (retval)
            return retval;
#endif
    }

    if (test & DP_TEST_LINK_VIDEO_PATTERN) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_VIDEO_PATTERN\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_video_pattern(dptx, 0);
        if (retval)
            return retval;
#endif
    }
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 1, 0))
    if (test & DP_TEST_LINK_AUDIO_PATTERN) {
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_AUDIO_PATTERN\n", __func__);
#if 0
        retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_ACK);
        if (retval)
            return retval;

        retval = handle_test_link_audio_pattern(dptx);
        if (retval)
            return retval;
#endif
    }
#endif
    if (test & DP_TEST_LINK_EDID_READ) {
        /* Invalid, this should happen on HOTPLUG */
        dev_dbg(xdev->dev, "%s(): DP_TEST_LINK_EDID_READ\n", __func__);
        return -ENOTSUPP;
    }

    if (test & DP_TEST_LINK_PHY_TEST_PATTERN) {
        dev_dbg(xdev->dev, "DP_TEST_LINK_PHY_TEST_PATTERN\n");
        retval = handle_test_phy_pattern(dptx);
        if (retval)
            return retval;
    }

    return 0;
}

static int handle_sink_request_b0(struct dp_dev *dptx)
{
	struct xdx_output *output = dptx->output;
	struct xdx_dm   *dm = output_to_dm(output);
	struct xdx_device *xdev = dm_to_xdev(dm);
	int timeout = 10000;
	u32 val;
	DPDBCmd_T_SINK PhyConfig;

	dev_dbg(xdev->dev, "%s()\n", __func__);

	PhyConfig.all = 0;
	PhyConfig.cmd = cH02ADisplayDBCmdDP;
	PhyConfig.subCmd = 0x3;
	PhyConfig.args0 = dptx->id;
	xdx_dp_h2a_doorbell_w(dptx, (u64)PhyConfig.all);

	for (; timeout; timeout--) {
		val = xdx_dp_h2a_doorbell_r(dptx);
		if(val & BIT(0)) {
			break;
		}

		udelay(1000);
	}

	if (!(val & BIT(0))) {
		dev_err(dptx->dev,"dptx CTS fail\n");
		return -ETIMEDOUT;
	}

	return 0;
}


static int handle_sink_request(struct dp_dev *dptx)
{
    struct xdx_output *output = dptx->output;
    struct xdx_dm   *dm = output_to_dm(output);
    struct xdx_device *xdev = dm_to_xdev(dm);
    int retval;
    u8 vector;
    u8 bytes[1];

	if (xdev->asic_type == CHIP_PANGU_B0) {
		handle_sink_request_b0(dptx);
		return 0;
	} else {
	    retval = dptx_link_check_status(dptx);
	    if (retval) {
	        return retval;
		}

	    retval = dptx_read_bytes_from_dpcd(dptx, DP_DEVICE_SERVICE_IRQ_VECTOR_ESI0, bytes, sizeof(bytes)); //sahakyan
	    if (retval) {
	        return retval;
		}

	    retval = dptx_read_dpcd(dptx, DP_DEVICE_SERVICE_IRQ_VECTOR, &vector);
	    if (retval)
	        return retval;

	    dev_dbg(xdev->dev, "%s: IRQ_VECTOR: 0x%02x\n", __func__, vector);

	    if (!vector)
	        return 0;

	    if ((vector & DP_REMOTE_CONTROL_COMMAND_PENDING) || (bytes[0] & DP_REMOTE_CONTROL_COMMAND_PENDING)) {
	    /* TODO */
	        dev_dbg(xdev->dev, "DP_REMOTE_CONTROL_COMMAND_PENDING: Not yet implemented\n");
	    }

	    if ((vector & DP_AUTOMATED_TEST_REQUEST) || (bytes[0] & DP_AUTOMATED_TEST_REQUEST)) {
	        dev_dbg(xdev->dev, "DP_AUTOMATED_TEST_REQUEST\n");
	        retval = handle_automated_test_request(dptx);
	        if (retval) {
	            dev_dbg(xdev->dev, "Automated test request failed\n");
	            if (retval == -ENOTSUPP) {
	                retval = dptx_write_dpcd(dptx, DP_TEST_RESPONSE, DP_TEST_NAK);
	                if (retval)
	                    return retval;
	            }
	        }
	    }
		return retval;
	}
}

static void xdx_dp_hpd_handler(struct dp_dev *dptx)
{
    u32 value;

    value = dptx_readl(dptx, DPTX_GENERAL_INTERRUPT);

    if (value & HPD_EVENT) {
        u32 hpdsts;
        hpdsts = dptx_readl(dptx, DPTX_HPD_STATUS);
        dev_dbg(dptx->dev, "DPTX_HPD_STATUS:%x\n", hpdsts);
        if (hpdsts & HPD_IRQ) {
            dev_dbg(dptx->dev, "IRQ from the HPD\n");
			//dptx->long_hpd = false;
            dptx_writel(dptx, DPTX_HPD_STATUS, HPD_IRQ);
	        if (dptx->cts) {
                int retval;
                dev_dbg(dptx->dev, "handle_sink_request\n");
                retval = handle_sink_request(dptx);
                if (retval) {
                    dev_dbg(dptx->dev, "unable to handle sink request:%d\n", retval);
                }
            }
        }

        if (hpdsts & HPD_HOT_PLUG) {
            dev_dbg(dptx->dev, "Hot plug detected\n");
            dptx->long_hpd = true;
            dptx_writel(dptx, DPTX_HPD_STATUS, HPD_HOT_PLUG);
        }

        if (hpdsts & HPD_HOT_UNPLUG) {
            dev_dbg(dptx->dev, "Unplug detected\n");
            dptx->long_hpd = true;
            dptx_writel(dptx, DPTX_HPD_STATUS, HPD_HOT_UNPLUG);
        }

        if (hpdsts & 0x80) {
            //dptx_writel(dptx, DPTX_HPD_STATUS, 0x80 | HPD_HOT_UNPLUG);
            dptx_writel(dptx, DPTX_HPD_STATUS, 0x80);
        }

        queue_work(dptx->workqueue, &dptx->hpd_work);
    }
    dptx_writel(dptx, DPTX_GENERAL_INTERRUPT, value);
}

void dptx_interrupt_disable(struct dp_dev *dptx)
{
    dptx_writel(dptx, GENERAL_INTERRUPT_ENABLE, 0);
    dptx_writel(dptx, HPD_INTERRUPT_ENABLE, 0);
}

void dptx_interrupt_enable(struct dp_dev *dptx)
{
    u32 value;

    value = dptx_readl(dptx, DPTX_GENERAL_INTERRUPT);
    dptx_writel(dptx, DPTX_GENERAL_INTERRUPT, value);

    value = dptx_readl(dptx, DPTX_HPD_STATUS);
    dptx_writel(dptx, DPTX_HPD_STATUS, value);

    dptx_writel(dptx, GENERAL_INTERRUPT_ENABLE, HPD_EVENT_EN);
    dptx_writel(dptx, HPD_INTERRUPT_ENABLE,
            HPD_IRQ | HPD_HOT_PLUG | HPD_HOT_UNPLUG);
}

static int dptx_irq_set(struct xdx_device *xdev, struct xdx_irq_src *src,
            unsigned type, enum xdx_interrupt_state state)
{
    // TODO: enable/disable IRQ
    return -1;
}

static int dptx_irq_process(struct xdx_device *xdev, struct xdx_irq_src *src,
                struct xdx_irq_entry *entry)
{
    struct xdx_dm *dm = xdev_to_dm(xdev);
    struct xdx_output *output = dm_to_output(dm);
    struct dp_dev *dptx;

    if (src->id == XDX_IRQ_CLIENTID_DP0)
        dptx = output->dp[0];
    else if (src->id == XDX_IRQ_CLIENTID_DP1)
        dptx = output->dp[1];
    else
        return -EINVAL;

    xdx_dp_hpd_handler(dptx);

    return 0;
}

static struct xdx_irq_src_funcs dptx_src_funcs = {
    .set = dptx_irq_set,
    .process = dptx_irq_process,
};

static int dptx_isr_install(struct dp_dev *dptx)
{
    struct xdx_dm* dm = output_to_dm(dptx->output);
    struct xdx_device* xdev = dm_to_xdev(dm);
    struct xdx_irq_src *irq_src;

    if (dptx->id == 0) {
        irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
        if (!irq_src)
            return -ENOMEM;
        irq_src->id = XDX_IRQ_CLIENTID_DP0;
    } else if (dptx->id == 1 && xdev->asic_type == CHIP_PANGU_B0) {
        irq_src = kmalloc(sizeof(*irq_src), GFP_KERNEL);
        if (!irq_src)
            return -ENOMEM;
        irq_src->id = XDX_IRQ_CLIENTID_DP1;
    } else {
        return -EINVAL;
    }

    irq_src->funcs = &dptx_src_funcs;

    xdx_irq_register_src(xdev, irq_src);

    dptx->irq_src = irq_src;

    return 0;
}

static int dptx_isr_uninstall(struct dp_dev *dptx)
{
    struct xdx_dm* dm = output_to_dm(dptx->output);
    struct xdx_device* xdev = dm_to_xdev(dm);
    int ret;

    ret = xdx_irq_unregister_src(xdev, dptx->irq_src);
    if (ret) {
        dev_err(xdev->dev, "failed to unregister irq source\n");
        return ret;
    }

    kfree(dptx->irq_src);
    dptx->irq_src = NULL;

    return 0;
}

int xdx_dp_init(struct xdx_dm *dm, u32 id)
{
    struct xdx_device *xdev = dm_to_xdev(dm);
    struct xdx_output *output = dm_to_output(dm);
    struct dp_dev *dptx;
    int ret;

    dptx = kzalloc(sizeof(*dptx), GFP_KERNEL);
    if (!dptx) {
        return -ENOMEM;
    }

    output->psDoorbellShareMemDesc[2 + id] = NULL;
    output->psDoorbellVirAddr[2 + id] = NULL;
    output->dp[id] = dptx;
    dptx->id = id;
    dptx->ssc_en = false;
    dev_dbg(xdev->dev, "init CTS default status\n");
    dptx->cts = 0;
    dptx->output = output;
    dptx->dev = xdev->dev;

    memcpy(&dptx->dp_mode, &dp_default_modes[0], sizeof(dptx->dp_mode));
    dptx->encoder_is_on = false;
    dptx->link_trained = false;
    dptx->vparams.pixel_mode = DPTX_MP_SINGLE_PIXEL;

    mutex_init(&dptx->mutex);

    dptx->workqueue = create_workqueue("dptx_hpd_queue");
    INIT_WORK(&dptx->hpd_work, xdx_dp_hpd_work);

    dptx->aux.dev = xdev->dev;
    dptx->aux.name = dev_name(dptx->dev);
    dptx->aux.transfer = xdx_dptx_aux_transfer;
#if ((LINUX_VERSION_CODE >= KERNEL_VERSION(5, 14, 0)))
    dptx->aux.drm_dev = &xdev->ddev;
#endif
    ret = drm_dp_aux_register(&dptx->aux);
    if (ret) {
        dev_err(xdev->dev, "failed to register dp aux\n");
        goto err_alloc;
    }

    ret = xdx_dptx_register(dptx, dm_dc_get_crtc_mask(dm->dc[id]));
    if (ret) {
        goto err_put_adapter;
    }

    if (xdev->asic_type == CHIP_PANGU_B0)
        xdx_dp_preinit(dptx);

    dptx_interrupt_disable(dptx);

    ret = dptx_isr_install(dptx);
    if (ret) {
        goto err_cleanup_dptx;
    }

    dptx_interrupt_enable(dptx);
    return 0;
err_cleanup_dptx:
    dptx->connector.funcs->destroy(&dptx->connector);
    dptx->encoder.funcs->destroy(&dptx->encoder);
err_put_adapter:
    drm_dp_aux_unregister(&dptx->aux);
err_alloc:
    if (dptx->workqueue)
        destroy_workqueue(dptx->workqueue);
    kfree(dptx);

    return 0;
}

int xdx_dp_deinit(struct xdx_dm *dm, u32 id)
{
    struct xdx_output *output = dm_to_output(dm);
    struct dp_dev *dptx = output->dp[id];

    xdx_dptx_phy_power_down(dptx);

    dptx_interrupt_disable(dptx);

    dptx_isr_uninstall(dptx);

    if (dptx->workqueue) {
        flush_workqueue(dptx->workqueue);
        destroy_workqueue(dptx->workqueue);
    }

    drm_dp_aux_unregister(&dptx->aux);

    kfree(dptx);

    output->dp[id] = NULL;

    return 0;
}

int xdx_dp_hw_init(struct xdx_dm *dm, u32 id)
{
    struct xdx_output *output = dm_to_output(dm);
    struct xdx_device* xdev = dm_to_xdev(dm);
    struct dp_dev *dptx = output->dp[id];

    if (xdev->asic_type == CHIP_PANGU_B0)
        xdx_dp_preinit(dptx);
    else
        xdx_dptx_tile_phy_init(dptx);

    dptx_interrupt_disable(dptx);

    dptx_interrupt_enable(dptx);

    return 0;
}
