/*******************************************************************************
@File
@Title          Server bridge for dma
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@Description    Implements the server side of the bridge for dma
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*******************************************************************************/

#include <linux/uaccess.h>

#include "img_defs.h"

#include "dma_km.h"

#include "common_dma_bridge.h"

#include "allocmem.h"
#include "pvr_debug.h"
#include "connection_server.h"
#include "pvr_bridge.h"
#include "dma_flags.h"
#if defined(SUPPORT_RGX)
#include "rgx_bridge.h"
#endif
#include "srvcore.h"
#include "handle.h"

#include <linux/slab.h>

/* ***************************************************************************
 * Server-side bridge entry points
 */

static_assert(MAX_DMA_OPS <= IMG_UINT32_MAX, "MAX_DMA_OPS must not be larger than IMG_UINT32_MAX");

static IMG_INT
PVRSRVBridgeDmaTransfer(IMG_UINT32 ui32DispatchTableEntry,
			IMG_UINT8 * psDmaTransferIN_UI8,
			IMG_UINT8 * psDmaTransferOUT_UI8, CONNECTION_DATA * psConnection)
{
	PVRSRV_BRIDGE_IN_DMATRANSFER *psDmaTransferIN =
	    (PVRSRV_BRIDGE_IN_DMATRANSFER *) IMG_OFFSET_ADDR(psDmaTransferIN_UI8, 0);
	PVRSRV_BRIDGE_OUT_DMATRANSFER *psDmaTransferOUT =
	    (PVRSRV_BRIDGE_OUT_DMATRANSFER *) IMG_OFFSET_ADDR(psDmaTransferOUT_UI8, 0);

	PMR **psPMRInt = NULL;
	PMR **psPMRInt_src = NULL;
	DMA_OP* pDma_Op = NULL;
	IMG_UINT32 ui32PmrNum = 0, ui32VDMANum = 0;

	IMG_UINT32 ui32NextOffset = 0;
	IMG_BYTE *pArrayArgsBuffer = NULL;
#if !defined(INTEGRITY_OS)
	IMG_BOOL bHaveEnoughSpace = IMG_FALSE;
#endif
	IMG_UINT32 ui32BufferSize = 0;
	IMG_UINT64 ui64BufferSize = (psDmaTransferIN->ui32uiFlags & DMA_FLAG_DEV_TO_DEV) 
				? ((IMG_UINT64) psDmaTransferIN->ui32NumDMAs * (sizeof(DMA_OP) + 2 * sizeof(sizeof(PMR *)))) + 0
				: ((IMG_UINT64) psDmaTransferIN->ui32NumDMAs * (sizeof(DMA_OP) + sizeof(sizeof(PMR *)))) + 0;

	if (unlikely(psDmaTransferIN->ui32NumDMAs > MAX_DMA_OPS))
	{
		psDmaTransferOUT->eError = PVRSRV_ERROR_BRIDGE_ARRAY_SIZE_TOO_BIG;
		goto DmaTransfer_exit;
	}

	if (ui64BufferSize > IMG_UINT32_MAX)
	{
		psDmaTransferOUT->eError = PVRSRV_ERROR_BRIDGE_BUFFER_TOO_SMALL;
		goto DmaTransfer_exit;
	}

	ui32BufferSize = (IMG_UINT32) ui64BufferSize;

	if (ui32BufferSize != 0)
	{
#if !defined(INTEGRITY_OS)
		/* Try to use remainder of input buffer for copies if possible, word-aligned for safety. */
		IMG_UINT32 ui32InBufferOffset =
		    PVR_ALIGN(sizeof(*psDmaTransferIN), sizeof(unsigned long));
		IMG_UINT32 ui32InBufferExcessSize =
		    ui32InBufferOffset >=
		    PVRSRV_MAX_BRIDGE_IN_SIZE ? 0 : PVRSRV_MAX_BRIDGE_IN_SIZE - ui32InBufferOffset;

		bHaveEnoughSpace = ui32BufferSize <= ui32InBufferExcessSize;
		if (bHaveEnoughSpace)
		{
			IMG_BYTE *pInputBuffer = (IMG_BYTE *) (void *)psDmaTransferIN;

			pArrayArgsBuffer = &pInputBuffer[ui32InBufferOffset];
		}
		else
#endif
		{
			pArrayArgsBuffer = OSAllocMemNoStats(ui32BufferSize);

			if (!pArrayArgsBuffer)
			{
				psDmaTransferOUT->eError = PVRSRV_ERROR_OUT_OF_MEMORY;
				goto DmaTransfer_exit;
			}
		}
	}

	if (psDmaTransferIN->ui32NumDMAs != 0)
	{
		psPMRInt = (PMR **) IMG_OFFSET_ADDR(pArrayArgsBuffer, ui32NextOffset);
		//OSCachedMemSet(psPMRInt, 0, psDmaTransferIN->ui32NumDMAs * sizeof(PMR *));
		ui32NextOffset += psDmaTransferIN->ui32NumDMAs * sizeof(PMR *);
		if(psDmaTransferIN->ui32uiFlags & DMA_FLAG_DEV_TO_DEV)
		{
			psPMRInt_src = (PMR **) IMG_OFFSET_ADDR(pArrayArgsBuffer, ui32NextOffset);
			ui32NextOffset += psDmaTransferIN->ui32NumDMAs * sizeof(PMR *);
		}
		pDma_Op = (DMA_OP *) IMG_OFFSET_ADDR(pArrayArgsBuffer, ui32NextOffset);
		ui32NextOffset += psDmaTransferIN->ui32NumDMAs * sizeof(DMA_OP);
	}

	/* Copy the data over */
	if (psDmaTransferIN->ui32NumDMAs > 0)
	{
		if (OSCopyFromUser
		    (NULL, pDma_Op, (const void __user *)psDmaTransferIN->hDma_Op,
		     psDmaTransferIN->ui32NumDMAs * sizeof(DMA_OP)) != PVRSRV_OK)
		{
			psDmaTransferOUT->eError = PVRSRV_ERROR_INVALID_PARAMS;

			goto DmaTransfer_exit;
		}
	}

	/* Lock over handle lookup. */
	LockHandle(psConnection->psHandleBase);

	{
		IMG_UINT32 i;

		if(psDmaTransferIN->ui32uiFlags & DMA_FLAG_DEV_TO_DEV)
		{
			for (i = 0; i < psDmaTransferIN->ui32NumDMAs; i++)
			{
				/* Look up the address from the handle */
				if(pDma_Op[i].hDevPMR && !(pDma_Op[i].eCmdType & DMA_CMD_PHY_FLAG))
				{
					psPMRInt[ui32PmrNum] = pDma_Op[i].hDevPMR;
					psDmaTransferOUT->eError =
						PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
									(void **)&pDma_Op[i].hDevPMR,
									psPMRInt[ui32PmrNum],
									PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);

					if (unlikely(psDmaTransferOUT->eError != PVRSRV_OK))
					{
						UnlockHandle(psConnection->psHandleBase);
						goto DmaTransfer_exit;
					}

					psPMRInt_src[ui32PmrNum] = pDma_Op[i].hHostPMR;
					psDmaTransferOUT->eError =
						PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
									(void **)&pDma_Op[i].hHostPMR,
									psPMRInt_src[ui32PmrNum],
									PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);
					ui32PmrNum++;
				}

				if(pDma_Op[i].eCmdType & DMA_CMD_VDMA_FLAG)
				{
					ui32VDMANum++;
				}
			}
		}
		else
		{
			for (i = 0; i < psDmaTransferIN->ui32NumDMAs; i++)
			{
				/* Look up the address from the handle */
				if(pDma_Op[i].hDevPMR && !(pDma_Op[i].eCmdType & DMA_CMD_PHY_FLAG))
				{
					psPMRInt[ui32PmrNum] = pDma_Op[i].hDevPMR;
					psDmaTransferOUT->eError =
						PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
									(void **)&pDma_Op[i].hDevPMR,
									psPMRInt[ui32PmrNum],
									PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);

					if (unlikely(psDmaTransferOUT->eError != PVRSRV_OK))
					{
						UnlockHandle(psConnection->psHandleBase);
						goto DmaTransfer_exit;
					}
					ui32PmrNum++;
				}

				if(pDma_Op[i].eCmdType & (DMA_CMD_VDMA_FLAG | DMA_CMD_MEMSET))
				{
					ui32VDMANum++;
				}
			}
		}
	}
	/* Release now we have looked up handles. */
	UnlockHandle(psConnection->psHandleBase);

	psDmaTransferOUT->eError =
	DmaTransfer(psConnection, OSGetDevNode(psConnection),
		psDmaTransferIN->ui32NumDMAs, ui32VDMANum,
		pDma_Op, psDmaTransferIN->ui32uiFlags | DMA_FLAG_FROM_USER, psDmaTransferIN->hUpdateTimeline);

DmaTransfer_exit:

	/* Lock over handle lookup cleanup. */
	LockHandle(psConnection->psHandleBase);

	if(psDmaTransferIN->ui32uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		IMG_UINT32 i;
		for (i = 0; i < ui32PmrNum; i++)
		{
			/* Unreference the previously looked up handle */
			PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase,
							    psPMRInt[i],
							    PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);

			/* Unreference the previously looked up handle */
			PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase,
							    psPMRInt_src[i],
							    PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);
		}
	}
	else
	{
		IMG_UINT32 i;
		for (i = 0; i < ui32PmrNum; i++)
		{
			/* Unreference the previously looked up handle */
			PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase,
							    psPMRInt[i],
							    PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);
		}
	}
	/* Release now we have cleaned up look up handles. */
	UnlockHandle(psConnection->psHandleBase);

	/* Allocated space should be equal to the last updated offset */
#ifdef PVRSRV_NEED_PVR_ASSERT
	if (psDmaTransferOUT->eError == PVRSRV_OK)
		PVR_ASSERT(ui32BufferSize == ui32NextOffset);
#endif /* PVRSRV_NEED_PVR_ASSERT */

#if defined(INTEGRITY_OS)
	if (pArrayArgsBuffer)
#else
	if (!bHaveEnoughSpace && pArrayArgsBuffer)
#endif
		OSFreeMemNoStats(pArrayArgsBuffer);

	return 0;
}

static IMG_INT
PVRSRVBridgeDmaSparseMappingTable(IMG_UINT32 ui32DispatchTableEntry,
				  IMG_UINT8 * psDmaSparseMappingTableIN_UI8,
				  IMG_UINT8 * psDmaSparseMappingTableOUT_UI8,
				  CONNECTION_DATA * psConnection)
{
	PVRSRV_BRIDGE_IN_DMASPARSEMAPPINGTABLE *psDmaSparseMappingTableIN =
	    (PVRSRV_BRIDGE_IN_DMASPARSEMAPPINGTABLE *)
	    IMG_OFFSET_ADDR(psDmaSparseMappingTableIN_UI8, 0);
	PVRSRV_BRIDGE_OUT_DMASPARSEMAPPINGTABLE *psDmaSparseMappingTableOUT =
	    (PVRSRV_BRIDGE_OUT_DMASPARSEMAPPINGTABLE *)
	    IMG_OFFSET_ADDR(psDmaSparseMappingTableOUT_UI8, 0);

	IMG_HANDLE hPMR = psDmaSparseMappingTableIN->hPMR;
	PMR *psPMRInt = NULL;
	IMG_BOOL *pbTableInt = NULL;

	IMG_UINT32 ui32NextOffset = 0;
	IMG_BYTE *pArrayArgsBuffer = NULL;
#if !defined(INTEGRITY_OS)
	IMG_BOOL bHaveEnoughSpace = IMG_FALSE;
#endif

	IMG_UINT32 ui32BufferSize = 0;
	IMG_UINT64 ui64BufferSize =
	    ((IMG_UINT64) psDmaSparseMappingTableIN->ui32SizeInPages * sizeof(IMG_BOOL)) + 0;

	if (psDmaSparseMappingTableIN->ui32SizeInPages > 32)
	{
		psDmaSparseMappingTableOUT->eError = PVRSRV_ERROR_BRIDGE_ARRAY_SIZE_TOO_BIG;
		goto DmaSparseMappingTable_exit;
	}

	psDmaSparseMappingTableOUT->pbTable = psDmaSparseMappingTableIN->pbTable;

	if (ui64BufferSize > IMG_UINT32_MAX)
	{
		psDmaSparseMappingTableOUT->eError = PVRSRV_ERROR_BRIDGE_BUFFER_TOO_SMALL;
		goto DmaSparseMappingTable_exit;
	}

	ui32BufferSize = (IMG_UINT32) ui64BufferSize;

	if (ui32BufferSize != 0)
	{
#if !defined(INTEGRITY_OS)
		/* Try to use remainder of input buffer for copies if possible, word-aligned for safety. */
		IMG_UINT32 ui32InBufferOffset =
		    PVR_ALIGN(sizeof(*psDmaSparseMappingTableIN), sizeof(unsigned long));
		IMG_UINT32 ui32InBufferExcessSize =
		    ui32InBufferOffset >=
		    PVRSRV_MAX_BRIDGE_IN_SIZE ? 0 : PVRSRV_MAX_BRIDGE_IN_SIZE - ui32InBufferOffset;

		bHaveEnoughSpace = ui32BufferSize <= ui32InBufferExcessSize;
		if (bHaveEnoughSpace)
		{
			IMG_BYTE *pInputBuffer = (IMG_BYTE *) (void *)psDmaSparseMappingTableIN;

			pArrayArgsBuffer = &pInputBuffer[ui32InBufferOffset];
		}
		else
#endif
		{
			pArrayArgsBuffer = OSAllocMemNoStats(ui32BufferSize);

			if (!pArrayArgsBuffer)
			{
				psDmaSparseMappingTableOUT->eError = PVRSRV_ERROR_OUT_OF_MEMORY;
				goto DmaSparseMappingTable_exit;
			}
		}
	}

	if (psDmaSparseMappingTableIN->ui32SizeInPages != 0)
	{
		pbTableInt = (IMG_BOOL *) IMG_OFFSET_ADDR(pArrayArgsBuffer, ui32NextOffset);
		ui32NextOffset += psDmaSparseMappingTableIN->ui32SizeInPages * sizeof(IMG_BOOL);
	}

	/* Lock over handle lookup. */
	LockHandle(psConnection->psHandleBase);

	/* Look up the address from the handle */
	psDmaSparseMappingTableOUT->eError =
	    PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
				       (void **)&psPMRInt,
				       hPMR, PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);
	if (unlikely(psDmaSparseMappingTableOUT->eError != PVRSRV_OK))
	{
		UnlockHandle(psConnection->psHandleBase);
		goto DmaSparseMappingTable_exit;
	}
	/* Release now we have looked up handles. */
	UnlockHandle(psConnection->psHandleBase);

	psDmaSparseMappingTableOUT->eError =
	    DmaSparseMappingTable(psPMRInt,
				  psDmaSparseMappingTableIN->uiOffset,
				  psDmaSparseMappingTableIN->ui32SizeInPages, pbTableInt);
	/* Exit early if bridged call fails */
	if (unlikely(psDmaSparseMappingTableOUT->eError != PVRSRV_OK))
	{
		goto DmaSparseMappingTable_exit;
	}

	/* If dest ptr is non-null and we have data to copy */
	if ((pbTableInt) && ((psDmaSparseMappingTableIN->ui32SizeInPages * sizeof(IMG_BOOL)) > 0))
	{
		if (unlikely
		    (OSCopyToUser
		     (NULL, (void __user *)psDmaSparseMappingTableOUT->pbTable, pbTableInt,
		      (psDmaSparseMappingTableIN->ui32SizeInPages * sizeof(IMG_BOOL))) !=
		     PVRSRV_OK))
		{
			psDmaSparseMappingTableOUT->eError = PVRSRV_ERROR_INVALID_PARAMS;

			goto DmaSparseMappingTable_exit;
		}
	}

DmaSparseMappingTable_exit:

	/* Lock over handle lookup cleanup. */
	LockHandle(psConnection->psHandleBase);

	/* Unreference the previously looked up handle */
	if (psPMRInt)
	{
		PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase,
					    hPMR, PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);
	}
	/* Release now we have cleaned up look up handles. */
	UnlockHandle(psConnection->psHandleBase);

	/* Allocated space should be equal to the last updated offset */
#ifdef PVRSRV_NEED_PVR_ASSERT
	if (psDmaSparseMappingTableOUT->eError == PVRSRV_OK)
		PVR_ASSERT(ui32BufferSize == ui32NextOffset);
#endif /* PVRSRV_NEED_PVR_ASSERT */

#if defined(INTEGRITY_OS)
	if (pArrayArgsBuffer)
#else
	if (!bHaveEnoughSpace && pArrayArgsBuffer)
#endif
		OSFreeMemNoStats(pArrayArgsBuffer);

	return 0;
}

static IMG_INT
PVRSRVBridgeDmaDeviceParams(IMG_UINT32 ui32DispatchTableEntry,
			    IMG_UINT8 * psDmaDeviceParamsIN_UI8,
			    IMG_UINT8 * psDmaDeviceParamsOUT_UI8, CONNECTION_DATA * psConnection)
{
	PVRSRV_BRIDGE_IN_DMADEVICEPARAMS *psDmaDeviceParamsIN =
	    (PVRSRV_BRIDGE_IN_DMADEVICEPARAMS *) IMG_OFFSET_ADDR(psDmaDeviceParamsIN_UI8, 0);
	PVRSRV_BRIDGE_OUT_DMADEVICEPARAMS *psDmaDeviceParamsOUT =
	    (PVRSRV_BRIDGE_OUT_DMADEVICEPARAMS *) IMG_OFFSET_ADDR(psDmaDeviceParamsOUT_UI8, 0);

	PVR_UNREFERENCED_PARAMETER(psDmaDeviceParamsIN);

	psDmaDeviceParamsOUT->eError =
	    DmaDeviceParams(psConnection, OSGetDevNode(psConnection),
			    &psDmaDeviceParamsOUT->ui32DmaBuffAlign,
			    &psDmaDeviceParamsOUT->ui32DmaTransferMult);
			    //&psDmaDeviceParamsOUT->bIommuEnable);

	return 0;
}

static IMG_INT
PVRSRVBridgeDmaMemFill(IMG_UINT32 ui32DispatchTableEntry,
			IMG_UINT8 * psDmaMemFillIN_UI8,
			IMG_UINT8 * psDmaMemFillOUT_UI8, CONNECTION_DATA * psConnection)
{
	PVRSRV_BRIDGE_IN_DMAMEMFILL *psDmaMemFillIN =
	    (PVRSRV_BRIDGE_IN_DMAMEMFILL *) IMG_OFFSET_ADDR(psDmaMemFillIN_UI8, 0);
	PVRSRV_BRIDGE_OUT_DMAMEMFILL *psDmaMemFillOUT =
	    (PVRSRV_BRIDGE_OUT_DMAMEMFILL *) IMG_OFFSET_ADDR(psDmaMemFillOUT_UI8, 0);

	PMR *psPMRInt = NULL;

	/* Lock over handle lookup. */
	LockHandle(psConnection->psHandleBase);

	{
		/* Look up the address from the handle */
		psDmaMemFillOUT->eError =
		    PVRSRVLookupHandleUnlocked(psConnection->psHandleBase,
					       (void **)&psPMRInt,
					       psDmaMemFillIN->hPMR,
					       PVRSRV_HANDLE_TYPE_PHYSMEM_PMR, IMG_TRUE);
		if (unlikely(psDmaMemFillOUT->eError != PVRSRV_OK))
		{
			UnlockHandle(psConnection->psHandleBase);
			goto DmaTransfer_exit;
		}
	}
	
	/* Release now we have looked up handles. */
	UnlockHandle(psConnection->psHandleBase);

	psDmaMemFillOUT->eError =
	    DmaMemFill(psConnection, OSGetDevNode(psConnection),
			psPMRInt,
			psDmaMemFillIN->uiOffset,
			psDmaMemFillIN->uiSize, psDmaMemFillIN->ui32Pattern);

DmaTransfer_exit:

	/* Lock over handle lookup cleanup. */
	LockHandle(psConnection->psHandleBase);

	/* Unreference the previously looked up handle */
	if (psPMRInt)
	{
		PVRSRVReleaseHandleUnlocked(psConnection->psHandleBase,
					    psDmaMemFillIN->hPMR,
					    PVRSRV_HANDLE_TYPE_PHYSMEM_PMR);
	}
	
	/* Release now we have cleaned up look up handles. */
	UnlockHandle(psConnection->psHandleBase);

	return 0;
}


/* ***************************************************************************
 * Server bridge dispatch related glue
 */

PVRSRV_ERROR InitDMABridge(void);
void DeinitDMABridge(void);

/*
 * Register all DMA functions with services
 */
PVRSRV_ERROR InitDMABridge(void)
{

	SetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMATRANSFER,
			      PVRSRVBridgeDmaTransfer, NULL);

	SetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMASPARSEMAPPINGTABLE,
			      PVRSRVBridgeDmaSparseMappingTable, NULL);

	SetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMADEVICEPARAMS,
			      PVRSRVBridgeDmaDeviceParams, NULL);

	
	SetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMAMEMFILL,
			      PVRSRVBridgeDmaMemFill, NULL);

	return PVRSRV_OK;
}

/*
 * Unregister all dma functions with services
 */
void DeinitDMABridge(void)
{

	UnsetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMATRANSFER);

	UnsetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMASPARSEMAPPINGTABLE);

	UnsetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMADEVICEPARAMS);
	
	UnsetDispatchTableEntry(PVRSRV_BRIDGE_DMA, PVRSRV_BRIDGE_DMA_DMAMEMFILL);

}
