/*************************************************************************/ /*!
@File           dma_km.c
@Title          kernel side of dma transfer scheduling
@Copyright      Copyright (c) Imagination Technologies Ltd. All Rights Reserved
@Description    Implements server side code for allowing DMA transfers between
                cpu and device memory.
@License        Dual MIT/GPLv2

The contents of this file are subject to the MIT license as set out below.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to deal
in the Software without restriction, including without limitation the rights
to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies or substantial portions of the Software.

Alternatively, the contents of this file may be used under the terms of
the GNU General Public License Version 2 ("GPL") in which case the provisions
of GPL are applicable instead of those above.

If you wish to allow use of your version of this file only under the terms of
GPL, and not to allow others to use your version of this file under the terms
of the MIT license, indicate your decision by deleting the provisions above
and replace them with the notice and other provisions required by GPL as set
out in the file called "GPL-COPYING" included in this distribution. If you do
not delete the provisions above, a recipient may use your version of this file
under the terms of either the MIT license or GPL.

This License is also included in this distribution in the file called
"MIT-COPYING".

EXCEPT AS OTHERWISE STATED IN A NEGOTIATED AGREEMENT: (A) THE SOFTWARE IS
PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR IMPLIED, INCLUDING
BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY, FITNESS FOR A PARTICULAR
PURPOSE AND NONINFRINGEMENT; AND (B) IN NO EVENT SHALL THE AUTHORS OR
COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.
*/ /**************************************************************************/
#if defined(__linux__)
#include <linux/version.h>
#include <linux/uaccess.h>
#include <asm/current.h>
#include <linux/sched.h>
#include <linux/mm.h>
#include <linux/highmem.h>
#endif

#include "pmr.h"
#include "log2.h"
#include "device.h"
#include "pvrsrv.h"
#include "osfunc.h"
#include "dma_km.h"
#include "pvr_debug.h"
#include "lock_types.h"
#include "allocmem.h"
#include "process_stats.h"
#if defined(PVRSRV_ENABLE_GPU_MEMORY_INFO) && defined(DEBUG)
#include "ri_server.h"
#endif
#include "devicemem.h"
#include "pvrsrv_apphint.h"
#include "pvrsrv_sync_server.h"
#include "km_apphint_defs.h"
#include "di_server.h"
#include "dma_flags.h"
#include "rgx_fwif_km.h"
#include "dma_drv_common.h"
#include "hash.h"
#include "xdxacpu.h"
#include "xdx_gfx_convert.h"
#include "physmem_lma.h"

#if defined(XDXGPU_SMMU_SUPPORT)
#include "smmu_platform.h"
#endif

/* This header must always be included last */
#if defined(__linux__)
#include "kernel_compatibility.h"
#endif
#define DMA_ALIGN 32
#define LEN_ALIGN 8
#define DMA_MIN_LEN 0x0

#define DMA_MAX_TRANSLATION_STACK_ALLOC				(8)

typedef struct _SERVER_CLEANUP_DATA_
{
	PVRSRV_DEVICE_NODE *psDevNode;
	CONNECTION_DATA *psConnection;
	IMG_UINT32 uiCount;
	SYNC_TIMELINE_OBJ sTimelineObject;
	PMR** ppsPMR;
	IMG_BOOL bHasTdma;
} SERVER_CLEANUP_DATA;

#if !defined(NO_HARDWARE)
static void Cleanup(void* pvCleanupData, IMG_BOOL bAdvanceTimeline)
{
	PVRSRV_ERROR eError;
	SERVER_CLEANUP_DATA* psCleanupData = (SERVER_CLEANUP_DATA*) pvCleanupData;

#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_ERROR, "Server Cleanup thread entry (%p)", pvCleanupData));
#endif

	/* Advance timeline */
	if (psCleanupData->sTimelineObject.pvTlObj && bAdvanceTimeline)
	{
		eError = SyncSWTimelineAdvanceKM(psCleanupData->psDevNode, &psCleanupData->sTimelineObject);
		PVR_LOG_IF_ERROR(eError, "SyncSWTimelineAdvanceKM");
		eError = SyncSWTimelineReleaseKM(&psCleanupData->sTimelineObject);
		PVR_LOG_IF_ERROR(eError, "SyncSWTimelineReleaseKM");
	}

	OSAtomicDecrement(&psCleanupData->psConnection->ui32NumDmaTransfersInFlight);
#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_ERROR, "Decremented to %d", OSAtomicRead(&psCleanupData->psConnection->ui32NumDmaTransfersInFlight)));
#endif
	eError = OSEventObjectSignal(psCleanupData->psConnection->hDmaEventObject);
	if (eError != PVRSRV_OK)
	{
		PVR_DPF((PVR_DBG_ERROR, "%s: OSEventObjectSignal failed: %s",
		        __func__, PVRSRVGetErrorString(eError)));
	}
}
#endif /* !defined(NO_HARDWARE) */


IMG_EXPORT PVRSRV_ERROR
PVRSRVInitialiseDMA(PVRSRV_DEVICE_NODE *psDeviceNode)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDeviceNode->psDevConfig;
	struct xdx_device *xdev = gfx_to_xdev(psDeviceNode);

	if (psDevConfig->bHasDma)
	{
		if (!xdev->vdma.dma_dev)
		{
			PVR_DPF((PVR_DBG_ERROR,
				"%s: Dma device is not ready or some error happed.", __func__));
			return PVRSRV_ERROR_NOT_READY;
		}

		if (xdev->tdma.dma_dev)
		{
			psDeviceNode->hDmaTTxChan = (IMG_HANDLE)get_tdma_channel(xdev->tdma.dma_dev);
			if (!psDeviceNode->hDmaTTxChan)
			{
				psDeviceNode->bHasTwiddleDMA = IMG_FALSE;
				PVR_DPF((PVR_DBG_ERROR, "Tdma channel get error!"));
			}
			else
                        {
				psDeviceNode->bHasTwiddleDMA = IMG_TRUE;
                        }
		}
		else
		{
			psDeviceNode->bHasTwiddleDMA = IMG_FALSE;
			PVR_DPF((PVR_DBG_MESSAGE, "2D/3D channel not support!"));
		}

		psDeviceNode->hDmaTxChan = (IMG_HANDLE)get_dma_channel(xdev->vdma.dma_dev);
		if (!psDeviceNode->hDmaTxChan)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}
		psDeviceNode->bHasSystemDMA = true;
		psDeviceNode->cache = KMEM_CACHE(sg_table, 0);
		if (!psDeviceNode->cache)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}

		psDeviceNode->sg_pool = mempool_create_slab_pool(1, psDeviceNode->cache);
		if (!psDeviceNode->sg_pool)
		{
			return PVRSRV_ERROR_RESOURCE_UNAVAILABLE;
		}
	}

	return PVRSRV_OK;
}

IMG_EXPORT void
PVRSRVDeInitialiseDMA(PVRSRV_DEVICE_NODE *psDeviceNode)
{
	if (psDeviceNode->bHasSystemDMA)
	{
		mempool_destroy(psDeviceNode->sg_pool);
		kmem_cache_destroy(psDeviceNode->cache);
		if (psDeviceNode->hDmaTxChan)
			release_dma_channel((struct dma_chan *)psDeviceNode->hDmaTxChan);
		psDeviceNode->hDmaTxChan = NULL;
		if (psDeviceNode->hDmaTTxChan)
			release_dma_channel((struct dma_chan *)psDeviceNode->hDmaTTxChan);
		psDeviceNode->hDmaTTxChan = NULL;
	}
}

IMG_EXPORT PVRSRV_ERROR
DmaDeviceParams(CONNECTION_DATA *psConnection,
				PVRSRV_DEVICE_NODE *psDevNode,
				IMG_UINT32 *ui32DmaBuffAlign,
				IMG_UINT32 *ui32DmaTransferMult)
				//IMG_UINT32 *ui32DmaTransferMult,
				//IMG_BOOL   *iommuEnable)
{
	PVRSRV_DEVICE_CONFIG *psDevConfig = psDevNode->psDevConfig;

	*ui32DmaBuffAlign = psDevConfig->ui32DmaAlignment;
	*ui32DmaTransferMult = psDevConfig->ui32DmaTransferUnit;
	//*iommuEnable = psDevConfig->bIommuEnable;

	 return PVRSRV_OK;
}

IMG_EXPORT PVRSRV_ERROR
DmaSparseMappingTable(PMR *psPMR,
					  IMG_DEVMEM_OFFSET_T uiOffset,
					  IMG_UINT32 ui32SizeInPages,
					  IMG_BOOL *pbTable)
{
		PVRSRV_ERROR eError = PVRSRV_OK;
		IMG_DEV_PHYADDR *psDevPhyAddr;
		IMG_BOOL *pbValid;

		psDevPhyAddr = OSAllocZMem(ui32SizeInPages * sizeof(IMG_CPU_PHYADDR));
		PVR_LOG_GOTO_IF_NOMEM(psDevPhyAddr, eError, err1);

		pbValid = OSAllocZMem(ui32SizeInPages * sizeof(IMG_BOOL));
		PVR_LOG_GOTO_IF_NOMEM(pbValid, eError, err2);

		eError = PMRLockSysPhysAddresses(psPMR);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMRLockSysPhysAddresses", err3);

		eError = PMR_DevPhysAddr(psPMR,
								 OSGetPageShift(),
								 ui32SizeInPages,
								 uiOffset,
								 psDevPhyAddr,
								 pbValid);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr", err3);

		PMRUnlockSysPhysAddresses(psPMR);

		memcpy(pbTable, pbValid, ui32SizeInPages * sizeof(IMG_BOOL));

err3:
		OSFreeMem(pbValid);
err2:
		OSFreeMem(psDevPhyAddr);
err1:
		return eError;
}

#if defined(SUPPORT_PANGU_ACPU)
FORCE_INLINE
void Sharemem_ACPUTINYCOPY(AcpuMemCopy_T *AcpuMemCopyInfo, IMG_UINT64 devPAddr, IMG_UINT32 size, IMG_UINT32 dir)
{
	AcpuMemCopyInfo->DstAddr = devPAddr;
	AcpuMemCopyInfo->dataSize = size;
	AcpuMemCopyInfo->copyType = dir;
	AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_copy;
	PVR_ASSERT(size <= MAX_TINYBUFFER_MAX_SIZE);
}

FORCE_INLINE
void Sharemem_ACPUTINYCOPY_Sparse(AcpuMemCopy_T *AcpuMemCopyInfo, DDK_MEM_COPY_T *psddk_mem_copy, IMG_UINT32 size)
{
	memcpy(&AcpuMemCopyInfo->MemCopyParam, psddk_mem_copy, sizeof(DDK_MEM_COPY_T));
	AcpuMemCopyInfo->dataSize = size;
	AcpuMemCopyInfo->copyType = 3;
	AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_copy;
}
#if 0
FORCE_INLINE
void Sharemem_ACPUTINYCOPY_Smmu(AcpuMemCopy_T *AcpuMemCopyInfo, DDK_MEM_COPY_T *psddk_mem_copy, IMG_UINT32 size)
{
	memcpy(&AcpuMemCopyInfo->MemCopyParam, psddk_mem_copy, sizeof(DDK_MEM_COPY_T));
	AcpuMemCopyInfo->dataSize = size;
	AcpuMemCopyInfo->copyType = 3;
	AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_copy_smmu;
}
#endif
FORCE_INLINE
void Sharemem_ACPUREGCOPY(AcpuMemCopy_T *AcpuMemCopyInfo,
									IMG_UINT64 srcPADDR, IMG_UINT64 devPAddr,
										IMG_UINT32 mask, IMG_UINT32 value)
{
	AcpuMemCopyInfo->cmdType = acpu_dbcmd_memop_subtype_tiny_reg;
	AcpuMemCopyInfo->DstAddr = devPAddr;
	AcpuMemCopyInfo->SrcAddr = srcPADDR;
	AcpuMemCopyInfo->mask = mask;
	AcpuMemCopyInfo->value = value;
}

FORCE_INLINE
int Sharemem_ACPUTINYCOPY_D2D(PVRSRV_DEVICE_NODE * psDevNode, DMA_OP* psOp)
{
	int ret = 0;
	IMG_UINT32 ui32SendSize, ui32HostOffsetInPage, ui32HostPageIndex = 0;
	IMG_UINT32 ui32DevOffsetInPage, ui32DevPageIndex = 0,PageOffset = 0;
	IMG_UINT64 *pui64HostPhyTable = NULL;
	IMG_UINT64 *pui64DevPhyTable = NULL;
	IMG_UINT64 ui64HostPhyAddress,ui64DevPhyAddress,ui32DmaCmdSize;

	IMG_DEVMEM_SIZE_T uiLogicalSize;
	IMG_UINT32 ui32SizeInPages;

	ui32HostPageIndex = psOp->uiHOSTOffset >> PAGE_SHIFT;
	ui32HostOffsetInPage = psOp->uiHOSTOffset & (PAGE_SIZE - 1);

	ui32DevPageIndex = psOp->uiOffset >> PAGE_SHIFT;
	ui32DevOffsetInPage = psOp->uiOffset & (PAGE_SIZE - 1);

	PMR_LogicalSize(psOp->hHostPMR, &uiLogicalSize);
	ui32SizeInPages = uiLogicalSize >> PAGE_SHIFT;
	pui64HostPhyTable = OSAllocZMem(ui32SizeInPages * sizeof(IMG_UINT64));
	if (pui64HostPhyTable == NULL) {
		ret = -ENOMEM;
		goto err_no_mem1;
	}

	PMR_LogicalSize(psOp->hDevPMR, &uiLogicalSize);
	ui32SizeInPages = uiLogicalSize >> PAGE_SHIFT;
	pui64DevPhyTable = OSAllocZMem(ui32SizeInPages * sizeof(IMG_UINT64));
	if (pui64HostPhyTable == NULL) {
		ret = -ENOMEM;
		goto err_no_mem2;
	}

	ret = PMR_GetPages(psOp->hHostPMR, PAGE_SHIFT, &pui64HostPhyTable);
	if (ret) {
		PVR_DPF((PVR_DBG_ERROR, "Failed to get pages of host pmr \n"));
		goto err_phy_addr;
	}

	ret = PMR_GetPages(psOp->hDevPMR, PAGE_SHIFT, &pui64DevPhyTable);
	if (ret) {
		PVR_DPF((PVR_DBG_ERROR, "Failed to get pages of devide pmr \n"));
		goto err_phy_addr;
	}

	PVR_DPF((PVR_DBG_MESSAGE, "%s pui64HostPhyTable = 0x%llx\n",__func__,pui64HostPhyTable[0]));
	PVR_DPF((PVR_DBG_MESSAGE, "%s pui64DevPhyTable[0] = 0x%llx\n",__func__,pui64DevPhyTable[0]));

	ui32DmaCmdSize = 0;
	ui64DevPhyAddress = pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
	ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

	while (psOp->uiSize > 0) {
		ui32SendSize = MIN(PAGE_SIZE - ui32HostOffsetInPage, PAGE_SIZE - ui32DevOffsetInPage);

		if (psOp->uiSize > ui32SendSize) {
			psOp->uiSize -= ui32SendSize;
			ui32DmaCmdSize += ui32SendSize;

			if (ui32HostOffsetInPage > ui32DevOffsetInPage) {
				ui32HostOffsetInPage = 0;
				ui32DevOffsetInPage += ui32SendSize;
				if ((pui64HostPhyTable[ui32HostPageIndex + PageOffset] + PAGE_SIZE) ==  pui64HostPhyTable[ui32HostPageIndex + PageOffset+1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Host buffer continuous page %llx, %llx", pui64HostPhyTable[ui32HostPageIndex + PageOffset], pui64HostPhyTable[ui32HostPageIndex + PageOffset + 1]));
					ui32HostPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset -1] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

				}
			} else if (ui32HostOffsetInPage == ui32DevOffsetInPage) {
				ui32HostOffsetInPage = 0;
				ui32DevOffsetInPage = 0;

				if ((pui64HostPhyTable[ui32HostPageIndex + PageOffset] + PAGE_SIZE) == pui64HostPhyTable[ui32HostPageIndex + PageOffset + 1]
					&& (pui64DevPhyTable[ui32DevPageIndex + PageOffset] + PAGE_SIZE) ==  pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Host & Device buffer continuous page %llx, %llx", pui64HostPhyTable[ui32HostPageIndex + PageOffset],pui64HostPhyTable[ui32HostPageIndex + PageOffset +1]));
					ui32HostPageIndex++;
					ui32DevPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset] + ui32HostOffsetInPage;

				}
			} else {
				ui32DevOffsetInPage = 0;
				ui32HostOffsetInPage += ui32SendSize;
				if ((pui64DevPhyTable[ui32DevPageIndex + PageOffset] + PAGE_SIZE) ==  pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]) {
					PVR_DPF((PVR_DBG_VERBOSE,"Device buffer continuous page %llx, %llx", pui64DevPhyTable[ui32DevPageIndex + PageOffset], pui64DevPhyTable[ui32DevPageIndex + PageOffset + 1]));
					ui32DevPageIndex++;
				} else {
					acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
					PageOffset++;
					ui32DmaCmdSize = 0;
					ui64DevPhyAddress =  pui64DevPhyTable[ui32DevPageIndex + PageOffset] + ui32DevOffsetInPage;
					ui64HostPhyAddress = pui64HostPhyTable[ui32HostPageIndex + PageOffset -1] + ui32HostOffsetInPage;
				}
			}
		} else {
			ui32DmaCmdSize += psOp->uiSize;
			acpu_tiny_copy_package(psDevNode, ui64DevPhyAddress, ui64HostPhyAddress, ui32DmaCmdSize);
			break;
		}
	}

	OSFreeMem(pui64DevPhyTable);
	OSFreeMem(pui64HostPhyTable);

	return 0;

err_phy_addr:
	OSFreeMem(pui64DevPhyTable);

err_no_mem2:
	OSFreeMem(pui64HostPhyTable);

err_no_mem1:
	return ret;
}
#endif

#if defined(NO_HARDWARE)
DmaTransfer(CONNECTION_DATA *psConnection,
			PVRSRV_DEVICE_NODE *psDevNode,
			IMG_UINT32 uiNumDMAs,
			IMG_UINT32 uiNumVDMAs,
			DMA_OP *pDma_Op,
			IMG_UINT32 uiFlags,
			PVRSRV_TIMELINE iUpdateFenceTimeline)
{
/* On nohw the kernel call just advances the timeline to signal completion */
	PVRSRV_ERROR eError = PVRSRV_OK;
	SYNC_TIMELINE_OBJ sSwTimeline = {NULL, PVRSRV_NO_TIMELINE};

	if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
	{
		eError = SyncSWGetTimelineObj(iUpdateFenceTimeline, &sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWGetTimelineObj");

		eError = SyncSWTimelineAdvanceKM(psDevNode, &sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWTimelineAdvanceKM");

		eError = SyncSWTimelineReleaseKM(&sSwTimeline);
		PVR_LOG_RETURN_IF_ERROR(eError, "SyncSWTimelineReleaseKM");
	}
	return PVRSRV_OK;
}
#else
#define XDXGPU_TDMA_USE_SMMU IMG_TRUE
#define XDXGPU_VDMA_USE_SMMU IMG_FALSE
static PVRSRV_ERROR DmaGetPhyAddrs(
		IMG_HANDLE hDevPMR,
		IMG_UINT32 ui32Log2PageSize,
		IMG_UINT32 ui32AddrNums,
		IMG_UINT32 uiOffset,
		IMG_DEV_PHYADDR  *psDmaAddr,
		IMG_BOOL *bValids)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	PVRSRV_ERROR eError_lock = PVRSRV_OK;

	eError_lock = PMRLockSysPhysAddresses(hDevPMR);
	PVR_LOG_GOTO_IF_ERROR(eError_lock, "DmaGetPhyAddrs:Lock", e0);
	eError = PMR_DevPhysAddr(hDevPMR,
							 ui32Log2PageSize,
							 ui32AddrNums,
							 uiOffset,
							 psDmaAddr,
							 bValids);
	eError_lock = PMRUnlockSysPhysAddresses(hDevPMR);
	PVR_LOG_GOTO_IF_ERROR(eError, "DmaGetPhyAddrs", e1);
	PVR_LOG_GOTO_IF_ERROR(eError_lock, "DmaGetPhyAddrs:UnLock", e0);
e1:
	return eError;
e0:
	return eError_lock;
}

static PVRSRV_ERROR DmaPMRRequestDmaAddrs(
		PVRSRV_DEVICE_NODE *psDevNode,
		IMG_HANDLE hDevPMR,
		IMG_UINT32 uiSize,
		IMG_UINT32 uiOffset,
		DMADevAddrType *ptDevAddr,
		IMG_BOOL Contig,
		IMG_BOOL bSmmuMap
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	IMG_UINT32 uiOffsetInPage;
	IMG_UINT32 ui32AddrNums;
#if defined(XDXGPU_SMMU_SUPPORT)
	struct xdx_device *xdev = gfx_to_xdev(psDevNode);
#endif

	if (!ptDevAddr)
	{
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		PVR_LOG_GOTO_IF_ERROR(eError, "[ASSERT:ptDevAddr]",e2);
	}
	memset(ptDevAddr, 0, sizeof(DMADevAddrType));
	ptDevAddr->psSrcPMR = hDevPMR;
#if defined(XDXGPU_SMMU_SUPPORT)
	if (bSmmuMap)
	{
		ptDevAddr->psDmaAddrVals = (IMG_DEV_PHYADDR *)&ptDevAddr->asAddressArray[0];
		eError = XdxSMMUComAllocAndMap(xdev, hDevPMR, &ptDevAddr->psDmaAddrVals->uiAddr);
		if (eError == PVRSRV_OK) {
			ptDevAddr->bIsVaddr = IMG_TRUE;
			ptDevAddr->bIsContig = 1;
			ptDevAddr->pbValids = NULL;
			ptDevAddr->ui32AddrNums = 1;
			ptDevAddr->psDmaAddrVals->uiAddr += uiOffset;
			PVR_DPF((PVR_DBG_MESSAGE, "[%s] Smmu Map:%llx", __func__, ptDevAddr->psDmaAddrVals->uiAddr));
			return PVRSRV_OK;
		}
	}
#endif
	ptDevAddr->bIsVaddr = IMG_FALSE;
	if (!PMR_IsSparse(hDevPMR) && PMR_IsPhysContig(PMRGetFlavourData(hDevPMR)))
	{
		ui32AddrNums = 1;
		ptDevAddr->bIsContig = 1;
		ptDevAddr->pbValids = NULL;
	}
	else
	{
		if (Contig)
		{
			eError = PVRSRV_ERROR_INVALID_PARAMS;
			PVR_LOG_GOTO_IF_ERROR(eError, "[Request continuous address]",e2);
		}
		ptDevAddr->bIsContig = 0;
		uiOffsetInPage = uiOffset & (OSGetPageSize() - 1);
		ui32AddrNums = (uiSize + uiOffsetInPage + OSGetPageSize() - 1) >> OSGetPageShift();
	}
	if (ui32AddrNums <= DMA_MAX_TRANSLATION_STACK_ALLOC_H)
		ptDevAddr->psDmaAddrVals = (IMG_DEV_PHYADDR *)&ptDevAddr->asAddressArray[0];
	else
		ptDevAddr->psDmaAddrVals = OSAllocZMem(ui32AddrNums * (sizeof(IMG_DMA_ADDR) + sizeof(IMG_BOOL)));
	PVR_LOG_GOTO_IF_NOMEM(ptDevAddr->psDmaAddrVals, eError, e2);
	ptDevAddr->ui32AddrNums = ui32AddrNums;
	ptDevAddr->pbValids = (IMG_BOOL*)(&ptDevAddr->psDmaAddrVals[ui32AddrNums]);
	eError = DmaGetPhyAddrs(hDevPMR, OSGetPageShift(), ui32AddrNums,
					uiOffset, (IMG_DEV_PHYADDR *)ptDevAddr->psDmaAddrVals, ptDevAddr->pbValids);
	PVR_LOG_GOTO_IF_ERROR(eError, "DmaGetPhyAddrs",e1);
	return eError;
e1:
	if (ui32AddrNums > DMA_MAX_TRANSLATION_STACK_ALLOC_H)
		OSFreeMem(ptDevAddr->psDmaAddrVals);
e2:
	return eError;
}

static PVRSRV_ERROR DmaPMRReleaseDmaAddrs(
	IMG_HANDLE hDevPMR,
	DMADevAddrType *ptDevAddr
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;

	if (!ptDevAddr || !ptDevAddr->psDmaAddrVals)
	{
		eError = PVRSRV_ERROR_INVALID_PARAMS;
		goto e0;
//		PVR_LOG_GOTO_IF_ERROR(eError, "[ASSERT:ptDevAddr]",e0);
	}
	if (ptDevAddr->psDmaAddrVals) {
		if (ptDevAddr->ui32AddrNums > DMA_MAX_TRANSLATION_STACK_ALLOC_H)
			OSFreeMem(ptDevAddr->psDmaAddrVals);
		ptDevAddr->psDmaAddrVals = NULL;
	}
	memset(ptDevAddr, 0 ,sizeof(DMADevAddrType));
e0:
	return eError;
}
static PVRSRV_ERROR DmaPrepXmitCpy(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	DMADevAddrType *tDevAddr,
	IMG_UINT32 uiFlags,
	void* pvOSData
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	tXDXDmaDualAddrs tDualAddrs;
	DMADevAddrType tSrcAddr;

	CHECK_8BIT_ALIGN(pDma_Op->uiSize, "Size", e0);
	CHECK_8BIT_ALIGN((IMG_DMA_ADDR*)tDevAddr->psDmaAddrVals->uiAddr, "DevAddr",e0);

	if (uiFlags & DMA_FLAG_DEV_TO_DEV) {
		eError = DmaPMRRequestDmaAddrs(psDevNode, pDma_Op->hHostPMR, pDma_Op->uiSize,
					pDma_Op->uiHOSTOffset, &tSrcAddr, IMG_FALSE, XDXGPU_VDMA_USE_SMMU);
		PVR_LOG_GOTO_IF_ERROR(eError, "DmaPMRRequestDmaAddrs",e0);
		CHECK_8BIT_ALIGN((IMG_DMA_ADDR*)tSrcAddr.psDmaAddrVals->uiAddr, "tSrcAddr",e0);
		tDualAddrs.DevAddrs1.pDmaPageVailds = tSrcAddr.pbValids;
		tDualAddrs.DevAddrs1.psDmaAddrs = (IMG_DMA_ADDR*)tSrcAddr.psDmaAddrVals;
		tDualAddrs.DevAddrs1.ui32DmaPages = tSrcAddr.ui32AddrNums;
		tDualAddrs.DevAddrs1.uiDmaOffse = pDma_Op->uiHOSTOffset;
		tDualAddrs.DevAddrs1.uiSize = pDma_Op->uiSize;
	} else {
		CHECK_8BIT_ALIGN(pDma_Op->ui64HostAddr, "HostAddr", e0);
		tDualAddrs.HostAddr1.puiAddress = pDma_Op->ui64HostAddr;
		tDualAddrs.HostAddr1.uiSize = pDma_Op->uiSize;
	}
	tDualAddrs.DevAddrs2.pDmaPageVailds = tDevAddr->pbValids;
	tDualAddrs.DevAddrs2.psDmaAddrs = (IMG_DMA_ADDR*)tDevAddr->psDmaAddrVals;
	tDualAddrs.DevAddrs2.ui32DmaPages = tDevAddr->ui32AddrNums;
	tDualAddrs.DevAddrs2.uiDmaOffse = pDma_Op->uiOffset;
	eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTxChan,
			XdxDmaTypeCPY, &tDualAddrs, pDma_Op->uiSize, uiFlags, pvOSData);

	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
		DmaPMRReleaseDmaAddrs(pDma_Op->hHostPMR, &tSrcAddr);
	return eError;
e0:
	return PVRSRV_ERROR_INVALID_PARAMS;
}

static PVRSRV_ERROR DmaPrepMemset(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	DMADevAddrType *tDevAddr,
	IMG_UINT32 uiFlags,
	void* pvOSData
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	tXDXDDmaMamemset  ptMemsetparas;

	CHECK_8BIT_ALIGN(pDma_Op->uiSize, "Size", e0);
	if (uiFlags & DMA_FLAG_MEM_TO_DEV) {
		CHECK_8BIT_ALIGN((IMG_DMA_ADDR*)tDevAddr->psDmaAddrVals->uiAddr, "DevAddr",e0);
		ptMemsetparas.DevAddrs1.pDmaPageVailds = tDevAddr->pbValids;
		ptMemsetparas.DevAddrs1.psDmaAddrs = (IMG_DMA_ADDR*)tDevAddr->psDmaAddrVals;
		ptMemsetparas.DevAddrs1.ui32DmaPages = tDevAddr->ui32AddrNums;
		ptMemsetparas.DevAddrs1.uiDmaOffse = pDma_Op->uiOffset;
	}else {
		CHECK_8BIT_ALIGN(pDma_Op->ui64HostAddr, "HostAddr", e0);
		ptMemsetparas.HostAddr1.puiAddress = pDma_Op->ui64HostAddr;
		ptMemsetparas.HostAddr1.uiSize = pDma_Op->uiSize;
	}
	ptMemsetparas.uiSize = pDma_Op->uiSize;
	ptMemsetparas.value = pDma_Op->value;
	eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTxChan,
			XdxDmaTypeMemset, &ptMemsetparas, pDma_Op->uiSize, uiFlags, pvOSData);
	return eError;
e0:
	return PVRSRV_ERROR_INVALID_PARAMS;
}


#if defined(SUPPORT_DMA_TRANSFER)
#ifdef SUPPORT_DMA_TRANSFER_B0

static IMG_UINT32 Dma2DCalcSize(DMA_OP *pDma_Op)
{
	IMG_UINT32 u32Size;
	IMG_UINT32 u32UpPow2X;

	if (pDma_Op->T_Width > pDma_Op->T_XStride)
	{
		PVR_DPF((PVR_DBG_ERROR,"[%s] Para Error. T_Width > T_XStride(%d > %d)", __func__,
			pDma_Op->T_Width, pDma_Op->T_XStride));
		return 0;
	}

	if (pDma_Op->T_LenUnit < 8 || pDma_Op->T_LenUnit & 0x3)
	{
		PVR_DPF((PVR_DBG_ERROR,"[%s] Para Error.LenUnit Not support.%d",
			__func__, pDma_Op->T_LenUnit));
		return 0;
	}
	u32UpPow2X = RoundUpToNextPowerOfTwo(pDma_Op->Width);
	u32Size = pDma_Op->T_XStride * u32UpPow2X * pDma_Op->T_LenUnit;
	u32Size = (u32Size + 7) >> 3;

	return u32Size;
}

static IMG_UINT64 DmaTwiddle3DCalcSize(DMA_TWD_2D3D_PARA* p3DArgs)
{
	IMG_UINT64 u64Size;

	if (p3DArgs->Width > p3DArgs->XStride)
	{
		PVR_DPF((PVR_DBG_ERROR,"[%s] Para Error. Width > Xstride(%lld > %lld)", __func__,
			p3DArgs->Width, p3DArgs->XStride));
		return 0;
	}
	if (p3DArgs->Height * p3DArgs->XStride > p3DArgs->YStride)
	{
		PVR_DPF((PVR_DBG_ERROR,"[%s] Para Error. Height*Xstride > YStride(%lld > %lld)",
			__func__,
			p3DArgs->Height * p3DArgs->XStride, p3DArgs->YStride));
		return 0;
	}
	if (p3DArgs->LenUnit < 8 || p3DArgs->LenUnit & 0x3)
	{
		PVR_DPF((PVR_DBG_ERROR,"[%s] Para Error.LenUnit Not support.%d",
			__func__, p3DArgs->LenUnit));
		return 0;
	}
	u64Size = p3DArgs->YStride * p3DArgs->Depth * p3DArgs->LenUnit;
	u64Size = (u64Size + 7) >> 3;
	PVR_DPF((PVR_DBG_MESSAGE,
		"[%s] Width:%llx Height:%llx Depth:%llx XStride:%llx YStride:%llx LenUnit:%x Len:0x%llx",
		__func__, p3DArgs->Width, p3DArgs->Height, p3DArgs->Depth,
		p3DArgs->XStride, p3DArgs->YStride, p3DArgs->LenUnit, u64Size));
	return u64Size;
}

static PVRSRV_ERROR DMAPrepXmit2D(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	IMG_UINT32 uiFlags,
	DMADevAddrType *tDevAddr,
	void* pvOSData
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	DMA_XMIT_2D_PARA tDma2dPare;
	DMADevAddrType tSrcDevAddr = {0};

	//Check the address is contig or mapped as contig virt addr.
	if (!tDevAddr || (!tDevAddr->bIsContig && !tDevAddr->bIsVaddr))
	{
		PVR_LOG_GOTO_WITH_ERROR("DMAPrepXmit3D",
				eError, PVRSRV_ERROR_INVALID_PARAMS, e0);
	}
	memset(&tDma2dPare, 0, sizeof(DMA_XMIT_2D_PARA));
	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		//Check the address is contig or mapped as contig virt addr.
		eError = DmaPMRRequestDmaAddrs(psDevNode, pDma_Op->hHostPMR, pDma_Op->uiSize,
			pDma_Op->HOSTOffset, &tSrcDevAddr, IMG_TRUE, XDXGPU_TDMA_USE_SMMU);
		PVR_LOG_GOTO_IF_ERROR(eError, "[Invalid HostPMR]DmaPMRRequestDmaAddrs", e0);
		if (!tSrcDevAddr.bIsContig && !tSrcDevAddr.bIsVaddr)
		{
			PVR_LOG_GOTO_WITH_ERROR("DMAPrepXmit3D",
					eError, PVRSRV_ERROR_INVALID_PARAMS, e1);
		}
		tDma2dPare.SrcAddr = tSrcDevAddr.psDmaAddrVals->uiAddr;
		tDma2dPare.SrcAddrFlag |= tSrcDevAddr.bIsVaddr ? DMA_ADDR_FLAG_VA : 0;
		tDma2dPare.SrcAddrFlag |= DMA_ADDR_FLAG_GPU_PORT;
	}
	else if (uiFlags & DMA_FLAG_MEM_TO_DEV)
	{
		tDma2dPare.SrcAddr = pDma_Op->ui64HostAddr;
		tDma2dPare.SrcAddrFlag |= DMA_ADDR_FLAG_HOST_PORT;
	}
	else
	{
		// Not support M2M & D2M  now.
		return PVRSRV_ERROR_INVALID_PARAMS;
	}

	tDma2dPare.DstAddr = tDevAddr->psDmaAddrVals->uiAddr;
	tDma2dPare.DstAddrFlag |= tDevAddr->bIsVaddr ? DMA_ADDR_FLAG_VA : 0;
	tDma2dPare.DstAddrFlag |= DMA_ADDR_FLAG_GPU_PORT;
	tDma2dPare.Width = pDma_Op->T_Width;
	tDma2dPare.Height = pDma_Op->T_Height;
	tDma2dPare.XStride = pDma_Op->T_XStride;
	tDma2dPare.LenUnit = pDma_Op->T_LenUnit;
	tDma2dPare.Size = Dma2DCalcSize(pDma_Op);
	if (!tDma2dPare.Size)
	{
		goto e1;
	}
	PVR_DPF((PVR_DBG_MESSAGE, "src addr:%llx flag:%x, Dst addr:%llx flag:%x\n",
		tDma2dPare.SrcAddr, tDma2dPare.SrcAddrFlag ,
		tDma2dPare.DstAddr, tDma2dPare.DstAddrFlag));
	eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTTxChan,
			XdxDmaType2D, &tDma2dPare, tDma2dPare.Size, uiFlags, pvOSData);

	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		DmaPMRReleaseDmaAddrs(pDma_Op->hHostPMR, &tSrcDevAddr);
	}
e1:
	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		DmaPMRReleaseDmaAddrs(pDma_Op->hHostPMR, &tSrcDevAddr);
	}
e0:
	return eError;
}

static PVRSRV_ERROR DMAPrepTwiddle2D3D(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	IMG_UINT32 uiFlags,
	DMADevAddrType *tDevAddr,
	void* pvOSData
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	DMA_TWD_2D3D_PARA tDma2d3dPara;
	DMADevAddrType tSrcDevAddr = {0};

	//Check the address is contig or mapped as contig virt addr.
	if (!tDevAddr || (!tDevAddr->bIsContig && !tDevAddr->bIsVaddr))
	{
		PVR_LOG_GOTO_WITH_ERROR("DMAPrepXmit2D3D",
				eError, PVRSRV_ERROR_INVALID_PARAMS, e0);
	}
	memset(&tDma2d3dPara, 0, sizeof(DMA_TWD_2D3D_PARA));
	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		//Check the address is contig or mapped as contig virt addr.
		eError = DmaPMRRequestDmaAddrs(psDevNode, pDma_Op->hHostPMR, pDma_Op->uiSize,
			pDma_Op->T_HOSTOffset, &tSrcDevAddr, IMG_TRUE, XDXGPU_TDMA_USE_SMMU);
		PVR_LOG_GOTO_IF_ERROR(eError, "[Invalid HostPMR]DmaPMRRequestDmaAddrs", e0);
		if (!tSrcDevAddr.bIsContig && !tSrcDevAddr.bIsVaddr)
		{
			PVR_LOG_GOTO_WITH_ERROR("DMAPrepXmit2D3D",
					eError, PVRSRV_ERROR_INVALID_PARAMS, e1);
		}
		tDma2d3dPara.SrcAddr = tSrcDevAddr.psDmaAddrVals->uiAddr;
		tDma2d3dPara.SrcAddrFlag |= tSrcDevAddr.bIsVaddr ? DMA_ADDR_FLAG_VA : 0;
		tDma2d3dPara.SrcAddrFlag |= DMA_ADDR_FLAG_GPU_PORT;
	}
	else if (uiFlags & DMA_FLAG_MEM_TO_DEV)
	{
		tDma2d3dPara.SrcAddr = pDma_Op->ui64HostAddr;
		tDma2d3dPara.SrcAddrFlag |= DMA_ADDR_FLAG_HOST_PORT;
	}
	else
	{
		// Not support M2M & D2M now.
		return PVRSRV_ERROR_INVALID_PARAMS;
	}

	tDma2d3dPara.DstAddr = tDevAddr->psDmaAddrVals->uiAddr;
	tDma2d3dPara.DstAddrFlag |= tDevAddr->bIsVaddr ? DMA_ADDR_FLAG_VA : 0;
	tDma2d3dPara.DstAddrFlag |= DMA_ADDR_FLAG_GPU_PORT;
	tDma2d3dPara.Width = pDma_Op->T_Width;
	tDma2d3dPara.Height = pDma_Op->T_Height;
	tDma2d3dPara.XStride = pDma_Op->T_XStride;
	tDma2d3dPara.LenUnit = pDma_Op->T_LenUnit;
	if (pDma_Op->eCmdType == DMA_CMD_HW_TWD_2D)
	{
		tDma2d3dPara.Depth = 1;
		tDma2d3dPara.CmdType = XdxDmaTypeTWD2D;
		tDma2d3dPara.Size = Dma2DCalcSize(pDma_Op);
	}
	else
	{
		tDma2d3dPara.YStride = pDma_Op->T_YStride;
		tDma2d3dPara.Depth = pDma_Op->T_Depth;
		tDma2d3dPara.CmdType = XdxDmaTypeTWD3D;
		tDma2d3dPara.Size = DmaTwiddle3DCalcSize(&tDma2d3dPara);
	}
	if (!tDma2d3dPara.Size)
	{
		goto e1;
	}

	PVR_DPF((PVR_DBG_MESSAGE,"src addr:%llx flag:%x, Dst addr:%llx flag:%x",
		tDma2d3dPara.SrcAddr, tDma2d3dPara.SrcAddrFlag ,
		tDma2d3dPara.DstAddr, tDma2d3dPara.DstAddrFlag));
	eError = OSDmaPrepareTransfer(psDevNode, psDevNode->hDmaTTxChan,
			XdxDmaTypeTWD2D3D, &tDma2d3dPara, tDma2d3dPara.Size, uiFlags, pvOSData);

	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		DmaPMRReleaseDmaAddrs(pDma_Op->hHostPMR, &tSrcDevAddr);
	}
e1:
	if (uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		DmaPMRReleaseDmaAddrs(pDma_Op->hHostPMR, &tSrcDevAddr);
	}
e0:
	return eError;
}
#endif

#if defined(SUPPORT_PANGU_ACPU)
static PVRSRV_ERROR DMABitCopy (
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	IMG_UINT32 uiFlags,
	IMG_UINT64 uiDmaAddr,
	IMG_UINT32 *pBitmaps
)
{
	IMG_UINT32 ui32TinyIndex = 0;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	AcpuMemCopy_T *AcpuMemCopyInfo = (AcpuMemCopy_T *)psDevInfo->puiTinyCopyVirAddr;

	ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
	if(INVALID_DB_INDEX == ui32TinyIndex)
	{
		//TinyBuffer is not available, release have allocated
		Do_ACPUTINYCOPY(psDevInfo, *pBitmaps, (uiFlags & DMA_FLAG_FROM_USER));
		*pBitmaps = 0;
		//try it again
		ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
	}
	Sharemem_ACPUREGCOPY(&AcpuMemCopyInfo[ui32TinyIndex], uiDmaAddr,
				uiDmaAddr, pDma_Op->mask, pDma_Op->value);
	BIT_SET(*pBitmaps, ui32TinyIndex);
	return PVRSRV_OK;
}

static PVRSRV_ERROR DMATinyCopy(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	IMG_UINT32 uiFlags,
	IMG_UINT64 uiDmaAddr,
	IMG_UINT32 *pBitmaps
)
{
	IMG_UINT32 ui32TinyIndex = 0;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	AcpuMemCopy_T *AcpuMemCopyInfo = (AcpuMemCopy_T *)psDevInfo->puiTinyCopyVirAddr;
	int ret;

	ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
	if(INVALID_DB_INDEX == ui32TinyIndex)
	{
		//TinyBuffer is not available, release have allocated
		Do_ACPUTINYCOPY(psDevInfo, *pBitmaps, (uiFlags & DMA_FLAG_FROM_USER));
		*pBitmaps = 0;
		//try it again
		ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
	}

	if(uiFlags & DMA_FLAG_MEM_TO_DEV)
	{
		PVR_ASSERT(pDma_Op->uiSize <= MAX_TINYBUFFER_MAX_SIZE);
		if(uiFlags & DMA_FLAG_FROM_USER)
		{
			ret = copy_from_user(&(AcpuMemCopyInfo[ui32TinyIndex].Data[0]),
				(IMG_UINT64*)pDma_Op->ui64HostAddr, pDma_Op->uiSize);
			if (ret < 0)
				PVR_DPF((PVR_DBG_ERROR, "[%s] copy data error.%d", __func__, ret));
		}
		else
		{
			//if use kmalloc/Vmalloc,should use memcpy
			memcpy(&(AcpuMemCopyInfo[ui32TinyIndex].Data[0]), (IMG_UINT64*)pDma_Op->ui64HostAddr, pDma_Op->uiSize);
		}
		Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], uiDmaAddr, pDma_Op->uiSize, 0);
		BIT_SET(*pBitmaps, ui32TinyIndex);
	}
	else if(uiFlags & DMA_FLAG_DEV_TO_DEV)
	{
		PVR_ASSERT(0);
	}
	else //DMA_FLAG_DEV_TO_MEM
	{
		Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], uiDmaAddr, pDma_Op->uiSize, 1);
		psDevInfo->asArrayTinyCopyHostAddr[ui32TinyIndex] = pDma_Op->ui64HostAddr;
		BIT_SET(*pBitmaps, ui32TinyIndex);
	}
	return PVRSRV_OK;
}

static PVRSRV_ERROR DMATinyFill(
	PVRSRV_DEVICE_NODE *psDevNode,
	DMA_OP *pDma_Op,
	IMG_UINT32 uiFlags,
	IMG_UINT64 uiDmaAddr,
	IMG_UINT32 *pBitmaps
)
{
	IMG_UINT32 ui32TinyIndex = 0;
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
	AcpuMemCopy_T *AcpuMemCopyInfo = (AcpuMemCopy_T *)psDevInfo->puiTinyCopyVirAddr;

	ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_FALSE);
	if(INVALID_DB_INDEX == ui32TinyIndex)
	{
		//TinyBuffer is not available, release have allocated
		Do_ACPUTINYCOPY(psDevInfo, *pBitmaps, (uiFlags & DMA_FLAG_FROM_USER));
		*pBitmaps = 0;
		//try it again
		ui32TinyIndex = TinyBufferPoolGet(psDevInfo, IMG_TRUE);
	}

	PVR_ASSERT(uiFlags & DMA_FLAG_MEM_TO_DEV);
	PVR_ASSERT(pDma_Op->uiSize <= sizeof(pDma_Op->as32FillData));
	memcpy(&AcpuMemCopyInfo[ui32TinyIndex].Data, &(pDma_Op->as32FillData[0]), pDma_Op->uiSize);
	Sharemem_ACPUTINYCOPY(&AcpuMemCopyInfo[ui32TinyIndex], uiDmaAddr, pDma_Op->uiSize, 0);
	BIT_SET(*pBitmaps, ui32TinyIndex);
	return PVRSRV_OK;
}

#endif
#endif

static PVRSRV_ERROR DMAServerDataInit
(
	CONNECTION_DATA *psConnection,
	PVRSRV_DEVICE_NODE *psDevNode,
	SERVER_CLEANUP_DATA *asServerData,
	PVRSRV_TIMELINE iUpdateFenceTimeline
)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	OSAtomicIncrement(&psConnection->ui32NumDmaTransfersInFlight);
#if defined(DMA_VERBOSE)
	PVR_DPF((PVR_DBG_VERBOSE, "Incremented to %d",
		OSAtomicRead(&psConnection->ui32NumDmaTransfersInFlight)));
#endif
	OSCachedMemSet(asServerData, 0, sizeof(SERVER_CLEANUP_DATA));
	if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
	{
		eError = SyncSWGetTimelineObj(iUpdateFenceTimeline, &asServerData->sTimelineObject);
		PVR_LOG_GOTO_IF_ERROR(eError, "SyncSWGetTimelineObj", e0);
	}
	asServerData->psDevNode = psDevNode;
	asServerData->psConnection = psConnection;
	asServerData->bHasTdma = false;
e0:
	return eError;
}

#ifdef DEBUG
#define DMAKM_AVG_MUL_SHIFT 9
static IMG_UINT64 avg_get(IMG_UINT64 avr_v, IMG_UINT64 new_v, int mul_shift, int test_cnt)
{
	if (!test_cnt) {
		return new_v << mul_shift;
	}
	return (avr_v - (avr_v >> mul_shift)) + new_v;
}
#endif

IMG_EXPORT PVRSRV_ERROR
DmaTransfer(CONNECTION_DATA *psConnection,
			PVRSRV_DEVICE_NODE *psDevNode,
			IMG_UINT32 uiNumDMAs,
			IMG_UINT32 uiNumHwDMAs,
			DMA_OP *pDma_Op,
			IMG_UINT32 uiFlags,
			PVRSRV_TIMELINE iUpdateFenceTimeline)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
	int i;
	SERVER_CLEANUP_DATA asServerData;
	IMG_DMA_ADDR  sDmaAddr;
	void* pvOSData;
	DMADevAddrType tDevAddr;
//	struct xdx_device *xdev = gfx_to_xdev(psDevNode);
#if defined(SUPPORT_PANGU_ACPU)
	PVRSRV_RGXDEV_INFO *psDevInfo = psDevNode->pvDevice;
//	AcpuMemCopy_T *AcpuMemCopyInfo =  (AcpuMemCopy_T *)psDevInfo->puiTinyCopyVirAddr;
	IMG_UINT32 dbNum = INVALID_DB_INDEX, bitmap = 0;
#endif

#ifdef DEBUG
	static IMG_UINT64 prepcost_avg = 0;
	static IMG_UINT64 allcost_avg = 0;
	static int test_cnt = 0;
	IMG_UINT64 Starttimer = OSClockus64();
#endif

	if (!psDevNode->hDmaTxChan)
	{
		eError = PVRSRVInitialiseDMA(psDevNode);
		PVR_LOG_RETURN_IF_ERROR_VA(eError, "[%d]Dma device not ready.", __LINE__);
	}

	if(uiNumHwDMAs)
	{
		if (!psConnection->bAcceptDmaRequests)
		{
			return PVRSRV_OK;
		}
		eError = DMAServerDataInit(psConnection, psDevNode, &asServerData, iUpdateFenceTimeline);
		PVR_LOG_GOTO_IF_ERROR(eError, "DmaTransfer:DMAServerDataInit", e0);
		eError = OSDmaAllocData(psDevNode, uiNumHwDMAs, &asServerData, Cleanup, &pvOSData);
		PVR_LOG_GOTO_IF_ERROR(eError, "OSDmaAllocData failed", e1);
	}

	for (i = 0; i < uiNumDMAs; i++)
	{
		memset(&tDevAddr, 0, sizeof(DMADevAddrType));
		if (pDma_Op[i].eCmdType & DMA_CMD_PHY_FLAG)
		{
			sDmaAddr.uiAddr = pDma_Op[i].ui64DevPAddr;
			tDevAddr.ui32AddrNums = 1;
			tDevAddr.bIsContig = IMG_TRUE;
			tDevAddr.pbValids = NULL;
			tDevAddr.psDmaAddrVals = (IMG_DEV_PHYADDR *)&sDmaAddr;
		}
		else if (pDma_Op[i].hDevPMR)
		{
			if (pDma_Op[i].eCmdType == DMA_CMD_HW_TWD_2D ||
					pDma_Op[i].eCmdType == DMA_CMD_HW_TWD_3D)
			{
				eError = DmaPMRRequestDmaAddrs(psDevNode, pDma_Op[i].hDevPMR, pDma_Op[i].uiSize,
							pDma_Op[i].uiOffset, &tDevAddr, IMG_TRUE, XDXGPU_TDMA_USE_SMMU);
				PVR_LOG_GOTO_IF_ERROR(eError, "DmaPMRRequestDmaAddrs",e2);
			}
			else
			{
				eError = DmaPMRRequestDmaAddrs(psDevNode, pDma_Op[i].hDevPMR, pDma_Op[i].uiSize,
							pDma_Op[i].uiOffset, &tDevAddr, IMG_FALSE, XDXGPU_VDMA_USE_SMMU);
				PVR_LOG_GOTO_IF_ERROR(eError, "DmaPMRRequestDmaAddrs",e2);
			}
			sDmaAddr.uiAddr = tDevAddr.psDmaAddrVals[0].uiAddr + pDma_Op[i].uiOffset;
		}
		else if (!(pDma_Op[i].eCmdType & DMA_CMD_MEMSET))
		{
			PVR_DPF((PVR_DBG_ERROR, "i %d, uiNumVDMAs %d, uiNumDMAs %d type %d\n",
						i, uiNumHwDMAs, uiNumDMAs, pDma_Op[i].eCmdType));
			continue;
		}

		switch(pDma_Op[i].eCmdType)
		{
			case DMA_CMD_VDMA_COPY:
			case DMA_CMD_VDMA_COPY_PHY:
				eError = DmaPrepXmitCpy(psDevNode, &pDma_Op[i],
						&tDevAddr, uiFlags, pvOSData);
				PVR_LOG_GOTO_IF_ERROR(eError, "DmaPrepXmitCpy",e2);
				asServerData.uiCount++;
			break;
			case DMA_CMD_MEMSET:
				eError = DmaPrepMemset(psDevNode, &pDma_Op[i],
						&tDevAddr, uiFlags, pvOSData);
				PVR_LOG_GOTO_IF_ERROR(eError, "DmaPrepMemset",e2);
				asServerData.uiCount++;
			break;
#if defined(SUPPORT_PANGU_ACPU)
			case DMA_CMD_TINY_FILL:
				eError = DMATinyFill(psDevNode, &pDma_Op[i], uiFlags, sDmaAddr.uiAddr, &bitmap);
				PVR_LOG_GOTO_IF_ERROR(eError, "DMATinyFill",e2);
				break;
			case DMA_CMD_TINY_COPY_OPT:
				Sharemem_ACPUTINYCOPY_D2D(psDevNode, &pDma_Op[i]);
				break;

			case DMA_CMD_TINY_COPY:
				eError = DMATinyCopy(psDevNode, &pDma_Op[i], uiFlags, sDmaAddr.uiAddr, &bitmap);
				PVR_LOG_GOTO_IF_ERROR(eError, "DMATinyCopy",e2);
				break;
			case DMA_CMD_BIT_COPY:
				eError = DMABitCopy(psDevNode, &pDma_Op[i], uiFlags, sDmaAddr.uiAddr, &bitmap);
				PVR_LOG_GOTO_IF_ERROR(eError, "DMABitCopy",e2);
				break;

			case DMA_CMD_TINY_MEMSET:
				acpu_tiny_memset_package(psDevNode, sDmaAddr.uiAddr, pDma_Op[i].uiSize, &(pDma_Op[i].value));
				break;
			case DMA_CMD_VDMA_COPY_PMR:
				PVR_ASSERT(0);
				break;
#endif
#ifdef SUPPORT_DMA_TRANSFER_B0
			case DMA_CMD_HW_2D:
				eError = DMAPrepXmit2D(psDevNode,
							&pDma_Op[i], uiFlags, &tDevAddr, pvOSData);
				PVR_LOG_GOTO_IF_ERROR(eError, "DMAPrepXmit2D",e2);
				asServerData.uiCount++;
				break;
			case DMA_CMD_HW_TWD_2D:
			case DMA_CMD_HW_TWD_3D:
				if (!psDevNode->hDmaTTxChan)
				{
					PVR_DPF((PVR_DBG_ERROR,
						"not support transfer type %x", pDma_Op[i].eCmdType));
					PVR_ASSERT(0);
					PVR_LOG_GOTO_WITH_ERROR("DMAPrepTwiddle2D3D", eError,
								PVRSRV_ERROR_NOT_SUPPORTED, e2);
				}
				eError = DMAPrepTwiddle2D3D(psDevNode,
							&pDma_Op[i], uiFlags, &tDevAddr, pvOSData);
				PVR_LOG_GOTO_IF_ERROR(eError, "DMAPrepTwiddle2D3D",e2);
				asServerData.uiCount++;
				asServerData.bHasTdma = true;
				break;
#endif
			default:
				PVR_DPF((PVR_DBG_ERROR, "not support transfer type %x", pDma_Op[i].eCmdType));
				PVR_ASSERT(0);
				break;
		}

		if (pDma_Op[i].hDevPMR)
		{
			DmaPMRReleaseDmaAddrs(pDma_Op[i].hDevPMR, &tDevAddr);
		}
	}
#if defined(SUPPORT_PANGU_ACPU)
	if(bitmap)
	{
		dbNum = Do_ACPUTINYCOPYNonBlock(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
	}
#endif
#ifdef DEBUG
	prepcost_avg = avg_get(prepcost_avg, OSClockus64()-Starttimer, DMAKM_AVG_MUL_SHIFT, test_cnt);
#endif
	if(uiNumHwDMAs)
	{
		if (asServerData.uiCount == uiNumHwDMAs)
		{
			//TODO: Multi-channal transfer mixed.
			if (asServerData.bHasTdma)
			{
				OSDmaSubmitTransfer(pvOSData, psDevNode->hDmaTTxChan);
			}
			else
			{
				OSDmaSubmitTransfer(pvOSData, psDevNode->hDmaTxChan);
			}
		}
		else
		{
			PVR_DPF((PVR_DBG_ERROR, "%s Failed.", __func__));
			/* One of the transfers could not be programmed, roll back */
			if (asServerData.bHasTdma)
			{
				OSDmaForceCleanup(psDevNode->hDmaTTxChan, pvOSData);
			}
			else
			{
				OSDmaForceCleanup(psDevNode->hDmaTxChan, pvOSData);
			}
		}
		eError = OSDmaWaitDone(pvOSData);
	}
#if defined(SUPPORT_PANGU_ACPU)
	if(dbNum != INVALID_DB_INDEX)
	{
		Wait_ACPUTINYCOPY(psDevInfo, dbNum, bitmap);
		TinyBufferPoolPuts(psDevInfo, bitmap, (uiFlags & DMA_FLAG_FROM_USER));
	}
#endif
#ifdef DEBUG
	allcost_avg = avg_get(allcost_avg, OSClockus64()-Starttimer, DMAKM_AVG_MUL_SHIFT, test_cnt);
	test_cnt++;
	if(!(test_cnt & 0x1ff))
		PVR_DPF((PVR_DBG_ERROR, "%s allAVR %lld us  prepAVR %lld us",
				__func__, allcost_avg >> DMAKM_AVG_MUL_SHIFT, prepcost_avg >> DMAKM_AVG_MUL_SHIFT));
#endif
	return eError;
e2:
	if (pDma_Op[i].hDevPMR)
	{
		DmaPMRReleaseDmaAddrs(pDma_Op[i].hDevPMR, &tDevAddr);
	}

	if (asServerData.bHasTdma)
	{
		OSDmaForceCleanup(psDevNode->hDmaTTxChan, pvOSData);
	}
	else
	{
		OSDmaForceCleanup(psDevNode->hDmaTxChan, pvOSData);
	}
	OSDmaFreeData(pvOSData);
e1:
	if (iUpdateFenceTimeline != PVRSRV_NO_TIMELINE)
	{
		SyncSWTimelineReleaseKM(&asServerData.sTimelineObject);
	}
e0:
	//OSLockRelease(psConnection->hDmaReqLock);
	return eError;
}
#endif

IMG_EXPORT PVRSRV_ERROR
DmaMemFill(CONNECTION_DATA *psConnection,
		    PVRSRV_DEVICE_NODE *psDevNode,
			PMR* psPMR,
			IMG_DEVMEM_OFFSET_T uiOffset,
			IMG_DEVMEM_SIZE_T uiSize,
			IMG_UINT32	ui32Pattern)
{
	PVRSRV_ERROR eError = PVRSRV_OK;
#if defined(NO_HARDWARE)
	return PVRSRV_OK;
#else

#if defined(SUPPORT_PANGU_ACPU)
	IMG_DEV_PHYADDR sDevPhyAddr;
	//IMG_DMA_ADDR sDmaAddr;
	IMG_BOOL bValid;

	eError = PMRLockSysPhysAddresses(psPMR);
	PVR_LOG_GOTO_IF_ERROR(eError, "PMRLockSysPhysAddresses", loop_e3);

	if (!PMR_IsSparse(psPMR) && (PMR_IsPhysContig(PMRGetFlavourData(psPMR))))
	{
		eError = PMR_DevPhysAddr(psPMR,
							 OSGetPageShift(),
							 1,
							 uiOffset,
							 &sDevPhyAddr,
							 &bValid);
		PVR_LOG_GOTO_IF_ERROR(eError, "PMR_DevPhysAddr", loop_e4);

		PVR_ASSERT(bValid == IMG_TRUE);

		ACPUTINYMEMSET(psDevNode, sDevPhyAddr.uiAddr, uiSize, &ui32Pattern);

	}
	else
	{
	//	IMG_UINT32 ui32SizeInPages;
	//	ui32SizeInPages = (uiSize + uiOffsetInPage + OSGetPageSize() - 1) >> OSGetPageShift();
		PVR_DPF((PVR_DBG_ERROR,"Dma memfill %lld ui32Pattern %d \n", uiSize, ui32Pattern));
		PVR_ASSERT(0);
	}

loop_e4:
	PMRUnlockSysPhysAddresses(psPMR);

loop_e3:
#endif
	return eError;
#endif
}

IMG_EXPORT PVRSRV_ERROR DmaSparseMemFill_lma(PVRSRV_DEVICE_NODE * psDevNode, IMG_UINT64 * pages, IMG_UINT32 pageNum, IMG_UINT32 pageSize, IMG_BYTE ui32Pattern)
{
	PVRSRV_ERROR eError;
	IMG_UINT32 page_idx = 0;
	IMG_UINT32 dma_idx = 0;
	IMG_UINT32 valid_idx = 0;
	IMG_UINT64 *psDmaAddr = NULL;
	IMG_UINT64 *psDmaSize = NULL;
	IMG_UINT32 valid_size = pageSize;
	IMG_UINT32 ui32MemSize;
#if defined(SUPPORT_PANGU_ACPU)
	IMG_UINT32 pattern = ui32Pattern;
#endif

	if (pages != NULL) {
		psDmaAddr = OSAllocZMem(pageNum * (sizeof(IMG_UINT64)));
		PVR_GOTO_IF_ERROR(!psDmaAddr, err2);
		psDmaSize = OSAllocZMem(pageNum * (sizeof(IMG_UINT64)));
		PVR_GOTO_IF_ERROR(!psDmaSize, err1);

		psDmaAddr[0] = pages[0];
		psDmaSize[0] = pageSize;
		for (page_idx = 1; page_idx < pageNum; page_idx++) {
			if (pages[page_idx] == psDmaAddr[valid_idx] + valid_size) {
				valid_size += pageSize;
				if ((page_idx + 1) == pageNum)
					psDmaSize[valid_idx] = valid_size;
			} else {
				psDmaSize[valid_idx] = valid_size;
				valid_idx++;
				psDmaAddr[valid_idx] = pages[page_idx];
				valid_size = pageSize;
				if ((page_idx + 1) == pageNum) {
					psDmaSize[valid_idx] = pageSize;
				}
			}
		}

		for (dma_idx = 0; dma_idx <= valid_idx; dma_idx++) {
			while (psDmaSize[dma_idx] > 0) {
				ui32MemSize = MIN(MAX_ACPU_MEMSET_SIZE, psDmaSize[dma_idx]);
#if defined(SUPPORT_PANGU_ACPU)
				eError = ACPUTINYMEMSET(psDevNode, psDmaAddr[dma_idx], ui32MemSize, &pattern);
#else
				eError = PVRSRV_ERROR_NOT_SUPPORTED;
#endif
				if (eError != PVRSRV_OK) {
					PVR_DPF((PVR_DBG_ERROR, "Failed to memset by ACPUTINYMEMSET"));
					return  eError;
				}
				psDmaAddr[dma_idx] = psDmaAddr[dma_idx] + ui32MemSize;
				psDmaSize[dma_idx] -= ui32MemSize;
			}
		}
	}

	OSFreeMem(psDmaAddr);
	OSFreeMem(psDmaSize);

	return PVRSRV_OK;

err1:
	OSFreeMem(psDmaAddr);

err2:
	PVR_DPF((PVR_DBG_ERROR, "Failed to allocate system memory"));

	return PVRSRV_ERROR_FAILED_TO_ALLOC_USER_MEM;
}

