/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */
#include "xdx_device.h"
#include "pangu.h"

#include "rgxdevice.h"
#include "pvrsrv.h"
#include "pvrsrv_device.h"
#include "syscommon.h"
#include "vz_vmm_pvz.h"
#include "allocmem.h"
#include "sysinfo.h"
#include "physheap.h"
#include "pvr_debug.h"
#if defined(__linux__)
#include <linux/dma-mapping.h>
#endif
/*
 * In systems that support trusted device address protection, there are three
 * physical heaps from which pages should be allocated:
 * - one heap for normal allocations
 * - one heap for allocations holding META code memory
 * - one heap for allocations holding secured DRM data
 */

#if defined(SUPPORT_VALIDATION) && defined(PDUMP)
#include "validation_soc.h"
#endif

/* Change to test CPU_LOCAL sys layers*/
#define UMA_HEAP_USAGE_FLAG PHYS_HEAP_USAGE_GPU_LOCAL
#define UMA_DEFAULT_HEAP PVRSRV_PHYS_HEAP_GPU_LOCAL

#define RGX_NOHW_CORE_CLOCK_SPEED 100000000
#define SYS_RGX_ACTIVE_POWER_LATENCY_MS (0)

/*!< System specific poll/timeout details */
#define MAX_HW_TIME_US                           (500000)
#define DEVICES_WATCHDOG_POWER_ON_SLEEP_TIMEOUT  (1500)//(10000)
#define DEVICES_WATCHDOG_POWER_OFF_SLEEP_TIMEOUT (3600000)
#define WAIT_TRY_COUNT                           (10000)

static void _pvr_error_notify(IMG_HANDLE hSysData,
				 PVRSRV_ROBUSTNESS_NOTIFY_DATA *psErrorData)
{
	struct xdx_device *xdev = (struct xdx_device *)hSysData;
	uint32_t level;

	switch (psErrorData->eResetReason) {
	case RGX_CONTEXT_RESET_REASON_NONE:
	case RGX_CONTEXT_RESET_REASON_GUILTY_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_LOCKUP:
	case RGX_CONTEXT_RESET_REASON_GUILTY_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_INNOCENT_OVERRUNING:
	case RGX_CONTEXT_RESET_REASON_HARD_CONTEXT_SWITCH:
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_OK:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_OK:
		level = PVR_DBG_MESSAGE;
		break;
	case RGX_CONTEXT_RESET_REASON_GPU_ECC_HWR:
	case RGX_CONTEXT_RESET_REASON_FW_EXEC_ERR:
		level = PVR_DBG_WARNING;
		break;
	case RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM:
	case RGX_CONTEXT_RESET_REASON_FW_ECC_ERR:
	case RGX_CONTEXT_RESET_REASON_FW_WATCHDOG:
	case RGX_CONTEXT_RESET_REASON_FW_PAGEFAULT:
	case RGX_CONTEXT_RESET_REASON_HOST_WDG_FW_ERR:
	default:
		level = PVR_DBG_ERROR;
		break;
	}

	if (psErrorData->pid > 0) {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "PID %d experienced error: %d\n",
				 psErrorData->pid, psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "PID %d experienced error: %d\n",
				psErrorData->pid, psErrorData->eResetReason);
	} else {
		if (level == PVR_DBG_MESSAGE)
			dev_info(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_WARNING)
			dev_warn(xdev->dev, "Device experienced error: %d\n",
				 psErrorData->eResetReason);
		else if (level == PVR_DBG_ERROR)
			dev_err(xdev->dev, "Device experienced error: %d\n",
				psErrorData->eResetReason);
	}

	if ((psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_WGP_CHECKSUM) ||
	    (psErrorData->eResetReason ==
	     RGX_CONTEXT_RESET_REASON_TRP_CHECKSUM)) {
		dev_err(xdev->dev, "ExtJobRef 0x%x, DM %d",
			psErrorData->uErrData.sChecksumErrData.ui32ExtJobRef,
			psErrorData->uErrData.sChecksumErrData.eDM);
	}
}


/*
	CPU to Device physical address translation
*/
static
void UMAPhysHeapCpuPAddrToDevPAddr(IMG_HANDLE hPrivData,
								   IMG_UINT32 ui32NumOfAddr,
								   IMG_DEV_PHYADDR *psDevPAddr,
								   IMG_CPU_PHYADDR *psCpuPAddr)
{
	PVR_UNREFERENCED_PARAMETER(hPrivData);

	/* Optimise common case */
	psDevPAddr[0].uiAddr = psCpuPAddr[0].uiAddr;
	if (ui32NumOfAddr > 1)
	{
		IMG_UINT32 ui32Idx;
		for (ui32Idx = 1; ui32Idx < ui32NumOfAddr; ++ui32Idx)
		{
			psDevPAddr[ui32Idx].uiAddr = psCpuPAddr[ui32Idx].uiAddr;
		}
	}
}

/*
	Device to CPU physical address translation
*/
static
void UMAPhysHeapDevPAddrToCpuPAddr(IMG_HANDLE hPrivData,
								   IMG_UINT32 ui32NumOfAddr,
								   IMG_CPU_PHYADDR *psCpuPAddr,
								   IMG_DEV_PHYADDR *psDevPAddr)
{
	PVR_UNREFERENCED_PARAMETER(hPrivData);

	/* Optimise common case */
	psCpuPAddr[0].uiAddr = psDevPAddr[0].uiAddr;
	if (ui32NumOfAddr > 1)
	{
		IMG_UINT32 ui32Idx;
		for (ui32Idx = 1; ui32Idx < ui32NumOfAddr; ++ui32Idx)
		{
			psCpuPAddr[ui32Idx].uiAddr = psDevPAddr[ui32Idx].uiAddr;
		}
	}
}

static PHYS_HEAP_FUNCTIONS gsPhysHeapFuncs =
{
	/* pfnCpuPAddrToDevPAddr */
	UMAPhysHeapCpuPAddrToDevPAddr,
	/* pfnDevPAddrToCpuPAddr */
	UMAPhysHeapDevPAddrToCpuPAddr,
};

static PVRSRV_ERROR PhysHeapsCreate(PHYS_HEAP_CONFIG **ppasPhysHeapsOut,
									IMG_UINT32 *puiPhysHeapCountOut)
{
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	IMG_UINT32 ui32NextHeapID = 0;
	IMG_UINT32 uiHeapCount = 2;

	pasPhysHeaps = OSAllocZMem(sizeof(*pasPhysHeaps) * uiHeapCount);
	if (!pasPhysHeaps)
	{
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}

	pasPhysHeaps[ui32NextHeapID].pszPDumpMemspaceName = "SYSMEM";
	pasPhysHeaps[ui32NextHeapID].eType = PHYS_HEAP_TYPE_UMA;
	pasPhysHeaps[ui32NextHeapID].psMemFuncs = &gsPhysHeapFuncs;
	pasPhysHeaps[ui32NextHeapID].ui32UsageFlags = UMA_HEAP_USAGE_FLAG;
	ui32NextHeapID++;


    pasPhysHeaps[ui32NextHeapID].pszPDumpMemspaceName = "SYSMEM";
    pasPhysHeaps[ui32NextHeapID].eType = PHYS_HEAP_TYPE_UMA;
    pasPhysHeaps[ui32NextHeapID].psMemFuncs = &gsPhysHeapFuncs;
    pasPhysHeaps[ui32NextHeapID].ui32UsageFlags = PHYS_HEAP_USAGE_FW_MAIN;
    ui32NextHeapID++;

	*ppasPhysHeapsOut = pasPhysHeaps;
	*puiPhysHeapCountOut = uiHeapCount;

	return PVRSRV_OK;
}

static void PhysHeapsDestroy(PHYS_HEAP_CONFIG *pasPhysHeaps)
{
	OSFreeMem(pasPhysHeaps);
}

PVRSRV_ERROR SysDevInit(void *pvOSDevice, PVRSRV_DEVICE_CONFIG **ppsDevConfig)
{
	struct device *dev = pvOSDevice;
	struct drm_device *ddev = dev_get_drvdata(dev);
	struct xdx_device *xdev = drm_to_xdev(ddev);
    PVRSRV_DEVICE_CONFIG *psDevConfig;
	RGX_DATA *psRGXData;
	RGX_TIMING_INFORMATION *psRGXTimingInfo;
	PHYS_HEAP_CONFIG *pasPhysHeaps;
	IMG_UINT32 uiPhysHeapCount;
	PVRSRV_ERROR eError;

	dma_set_mask(pvOSDevice, DMA_BIT_MASK(40));

	psDevConfig = OSAllocZMem(sizeof(*psDevConfig) +
							  sizeof(*psRGXData) +
							  sizeof(*psRGXTimingInfo));
	if (!psDevConfig)
	{
		return PVRSRV_ERROR_OUT_OF_MEMORY;
	}

	psRGXData = (RGX_DATA *)((IMG_CHAR *)psDevConfig + sizeof(*psDevConfig));
	psRGXTimingInfo = (RGX_TIMING_INFORMATION *)((IMG_CHAR *)psRGXData + sizeof(*psRGXData));

	eError = PhysHeapsCreate(&pasPhysHeaps, &uiPhysHeapCount);
	if (eError)
	{
		goto ErrorFreeDevConfig;
	}

	/* Setup RGX specific timing data */
	psRGXTimingInfo->ui32CoreClockSpeed        = RGX_NOHW_CORE_CLOCK_SPEED;
	psRGXTimingInfo->bEnableActivePM           = IMG_FALSE;
	psRGXTimingInfo->bEnableRDPowIsland        = IMG_FALSE;
	psRGXTimingInfo->ui32ActivePMLatencyms     = SYS_RGX_ACTIVE_POWER_LATENCY_MS;

	/* Set up the RGX data */
	psRGXData->psRGXTimingInfo = psRGXTimingInfo;

	/* Setup the device config */
	psDevConfig->pvOSDevice				= pvOSDevice;
	psDevConfig->pszName                = "xdxgpu";
	psDevConfig->pszVersion             = NULL;

	/* Device setup information */
	psDevConfig->sRegsCpuPBase.uiAddr   = 0x00f00baa;
	psDevConfig->ui32RegsSize           = 0x4000;
	psDevConfig->ui32IRQ                = 0x00000bad;

	psDevConfig->pasPhysHeaps			= pasPhysHeaps;
	psDevConfig->ui32PhysHeapCount		= uiPhysHeapCount;

	psDevConfig->eDefaultHeap = UMA_DEFAULT_HEAP;

	/* No power management on no HW system */
	psDevConfig->pfnPrePowerState       = NULL;
	psDevConfig->pfnPostPowerState      = NULL;

	/* No clock frequency either */
	psDevConfig->pfnClockFreqGet        = NULL;

	psDevConfig->hDevData               = psRGXData;

	psDevConfig->bDevicePA0IsValid       = IMG_FALSE;
	psDevConfig->pfnSysDevFeatureDepInit = NULL;

	/* Pdump validation system registers */
#if defined(SUPPORT_VALIDATION) && defined(PDUMP)
	{
		PVRSRV_DEVICE_NODE* psDeviceNode;

		/* Pretend we are calling for device 0 for now
		   (sDeviceNode.sDevId.ui32InternalID is 0 above)
		   until this code can be moved elsewhere.
		 */
		psDeviceNode = OSAllocZMem(sizeof(*psDeviceNode));
		PVR_LOG_RETURN_IF_NOMEM(psDeviceNode, "OSAllocZMem");

		PVRSRVConfigureSysCtrl(psDeviceNode, NULL, PDUMP_FLAGS_CONTINUOUS);
#if defined(SUPPORT_SECURITY_VALIDATION)
		PVRSRVConfigureTrustedDevice(psDeviceNode, NULL, PDUMP_FLAGS_CONTINUOUS);
#endif
		OSFreeMem(psDeviceNode);
	}
#endif

	psDevConfig->bHasFBCDCVersion31 = IMG_TRUE;

	/* Set psDevConfig->pfnSysDevErrorNotify callback */
	psDevConfig->pfnSysDevErrorNotify = _pvr_error_notify;

	*ppsDevConfig = psDevConfig;

	return PVRSRV_OK;

ErrorFreeDevConfig:
	OSFreeMem(psDevConfig);
	return eError;
}

void SysDevDeInit(PVRSRV_DEVICE_CONFIG *psDevConfig)
{
	PhysHeapsDestroy(psDevConfig->pasPhysHeaps);
	OSFreeMem(psDevConfig);
}

PVRSRV_ERROR SysInstallDeviceLISR(IMG_HANDLE hSysData,
								  IMG_UINT32 ui32IRQ,
								  const IMG_CHAR *pszName,
								  PFN_LISR pfnLISR,
								  void *pvData,
								  IMG_HANDLE *phLISRData)
{
	PVR_UNREFERENCED_PARAMETER(hSysData);
	PVR_UNREFERENCED_PARAMETER(ui32IRQ);
	PVR_UNREFERENCED_PARAMETER(pszName);
	PVR_UNREFERENCED_PARAMETER(pfnLISR);
	PVR_UNREFERENCED_PARAMETER(pvData);
	PVR_UNREFERENCED_PARAMETER(phLISRData);

	return PVRSRV_OK;
}

PVRSRV_ERROR SysUninstallDeviceLISR(IMG_HANDLE hLISRData)
{
	PVR_UNREFERENCED_PARAMETER(hLISRData);

	return PVRSRV_OK;
}

PVRSRV_ERROR SysDebugInfo(PVRSRV_DEVICE_CONFIG *psDevConfig,
				DUMPDEBUG_PRINTF_FUNC *pfnDumpDebugPrintf,
				void *pvDumpDebugFile)
{
	PVR_UNREFERENCED_PARAMETER(psDevConfig);
	PVR_UNREFERENCED_PARAMETER(pfnDumpDebugPrintf);
	PVR_UNREFERENCED_PARAMETER(pvDumpDebugFile);
	return PVRSRV_OK;
}