/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <drm/drm_atomic.h>
#include <drm/drm_atomic_helper.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(5, 13, 0))
#include <drm/drm_gem_atomic_helper.h>
#endif

#include <drm/drm_fourcc.h>
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 14, 0))
#include <drm/drm_gem_framebuffer_helper.h>
#endif
#include <drm/drm_plane_helper.h>

#include "xdx_vkms.h"
#include "xdx_gem.h"
#include "xdx_device.h"
#include "kernel_compatibility.h"

static const u32 vkms_formats[] = {
	DRM_FORMAT_XRGB8888,
};

static const u32 vkms_cursor_formats[] = {
	DRM_FORMAT_ARGB8888,
};

static struct drm_plane_state *
vkms_plane_duplicate_state(struct drm_plane *plane)
{
	struct vkms_plane_state *vkms_state;
	struct vkms_composer *composer;

	vkms_state = kzalloc(sizeof(*vkms_state), GFP_KERNEL);
	if (!vkms_state)
		return NULL;

	composer = kzalloc(sizeof(*composer), GFP_KERNEL);
	if (!composer) {
		DRM_DEBUG_KMS("Couldn't allocate composer\n");
		kfree(vkms_state);
		return NULL;
	}

	vkms_state->composer = composer;

	__drm_atomic_helper_plane_duplicate_state(plane, &vkms_state->base);

	return &vkms_state->base;
}

static void vkms_plane_destroy_state(struct drm_plane *plane,
				     struct drm_plane_state *old_state)
{
	struct vkms_plane_state *vkms_state = to_vkms_plane_state(old_state);
	struct drm_crtc *crtc = vkms_state->base.crtc;

	if (crtc) {
		/* dropping the reference we acquired in
		 * vkms_primary_plane_update()
		 */
		if (drm_framebuffer_read_refcount(&vkms_state->composer->fb))
			drm_framebuffer_put(&vkms_state->composer->fb);
	}

	kfree(vkms_state->composer);
	vkms_state->composer = NULL;

	__drm_atomic_helper_plane_destroy_state(old_state);
	kfree(vkms_state);
}

static void vkms_plane_reset(struct drm_plane *plane)
{
	struct vkms_plane_state *vkms_state;

	if (plane->state)
		vkms_plane_destroy_state(plane, plane->state);

	vkms_state = kzalloc(sizeof(*vkms_state), GFP_KERNEL);
	if (!vkms_state) {
		DRM_ERROR("Cannot allocate vkms_plane_state\n");
		return;
	}

	plane->state = &vkms_state->base;
	plane->state->plane = plane;
}

static const struct drm_plane_funcs vkms_plane_funcs = {
	.update_plane = drm_atomic_helper_update_plane,
	.disable_plane = drm_atomic_helper_disable_plane,
	.destroy = drm_plane_cleanup,
	.reset = vkms_plane_reset,
	.atomic_duplicate_state = vkms_plane_duplicate_state,
	.atomic_destroy_state = vkms_plane_destroy_state,
};
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static void vkms_plane_atomic_update(struct drm_plane *plane,
				     struct drm_plane_state *old_state)
#else
static void vkms_plane_atomic_update(struct drm_plane *plane,
				     struct drm_atomic_state *old_state)
#endif

{
	struct vkms_plane_state *vkms_plane_state;
	struct drm_framebuffer *fb = plane->state->fb;
	struct vkms_composer *composer;

	if (!plane->state->crtc || !fb)
		return;

	vkms_plane_state = to_vkms_plane_state(plane->state);

	composer = vkms_plane_state->composer;
	memcpy(&composer->src, &plane->state->src, sizeof(struct drm_rect));
	memcpy(&composer->dst, &plane->state->dst, sizeof(struct drm_rect));
	memcpy(&composer->fb, fb, sizeof(struct drm_framebuffer));
	drm_framebuffer_get(&composer->fb);
	composer->offset = fb->offsets[0];
	composer->pitch = fb->pitches[0];
#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	composer->cpp = fb->format->cpp[0];
#else
	composer->cpp = fb->bits_per_pixel / 8;
#endif
}
#if (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
static int vkms_plane_atomic_check(struct drm_plane *plane,
				   struct drm_plane_state *state)
{
	struct drm_crtc_state *crtc_state;
	bool can_position = false;
	int ret;

	if (!state->fb | !state->crtc)
		return 0;

	crtc_state = drm_atomic_get_crtc_state(state->state, state->crtc);
	if (IS_ERR(crtc_state))
		return PTR_ERR(crtc_state);

	if (plane->type == DRM_PLANE_TYPE_CURSOR)
		can_position = true;

	ret = drm_atomic_helper_check_plane_state(state, crtc_state,
						  DRM_PLANE_HELPER_NO_SCALING,
						  DRM_PLANE_HELPER_NO_SCALING,
						  can_position, true);
	if (ret != 0)
		return ret;

	/* for now primary plane must be visible and full screen */
	if (!state->visible && !can_position)
		return -EINVAL;

	return 0;
}
#else
static int vkms_plane_atomic_check(struct drm_plane *plane,
				   struct drm_atomic_state *state_original)
{
	struct drm_crtc_state *crtc_state;
	bool can_position = false;
	int ret;
	struct drm_plane_state *state = drm_atomic_get_new_plane_state(state_original,plane);
	if (!state->fb | !state->crtc)
		return 0;

	crtc_state = drm_atomic_get_crtc_state(state->state, state->crtc);
	if (IS_ERR(crtc_state))
		return PTR_ERR(crtc_state);

	if (plane->type == DRM_PLANE_TYPE_CURSOR)
		can_position = true;

	ret = drm_atomic_helper_check_plane_state(state, crtc_state,
						#if (LINUX_VERSION_CODE >= KERNEL_VERSION(6, 1, 0))
						DRM_PLANE_NO_SCALING,
						DRM_PLANE_NO_SCALING,
						#else
						DRM_PLANE_HELPER_NO_SCALING,
						DRM_PLANE_HELPER_NO_SCALING,
						#endif
						can_position, true);
	if (ret != 0)
		return ret;

	/* for now primary plane must be visible and full screen */
	if (!state->visible && !can_position)
		return -EINVAL;

	return 0;
}
#endif

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0))

#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
extern void xdx_drm_atomic_set_fence_for_plane(struct drm_plane_state *plane_state,
                               struct fence *fence);
#endif

int xdx_drm_gem_fb_prepare_fb(struct drm_plane *plane,
                          struct drm_plane_state *state)
{
        struct drm_gem_object *obj;
        struct xdx_bo *bo;
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
        struct fence *fence = NULL;
#else
	struct dma_fence *fence = NULL;
#endif

        if (!state->fb)
                return 0;

        obj = xdx_fb_get_gem_obj(state->fb, 0);
        bo = to_xdx_gem(obj);

        fence = reservation_object_get_excl_rcu(bo->resv);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 10, 0))
        xdx_drm_atomic_set_fence_for_plane(state, fence);
#else
        drm_atomic_set_fence_for_plane(state, fence);
#endif

        return 0;
}

#endif

static int vkms_prepare_fb(struct drm_plane *plane,
			   struct drm_plane_state *state)
{
	struct vkms_plane_state *vkms_state = to_vkms_plane_state(state);
	struct vkms_output *output = NULL;
	struct xdx_bo *bo;
	int ret;
	struct xdx_device *xdev;
	struct vkms_device *vkmsdev;
	char *name;
	static char* type_str[] = {
		[DRM_PLANE_TYPE_OVERLAY] = "overlay",
		[DRM_PLANE_TYPE_PRIMARY] = "primary",
		[DRM_PLANE_TYPE_CURSOR] = "cursor",
	};

	if (!state || !state->crtc)
		return 0;

	output = drm_crtc_to_vkms_output(state->crtc);

	if (!state->fb || !output || !output->composer_enabled)
		return 0;

#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0))
	bo = to_xdx_gem(xdx_fb_get_gem_obj(state->fb, 0));
#else
	bo = to_xdx_gem(drm_gem_fb_get_obj(state->fb, 0));
#endif
	xdev = drm_to_xdev(bo->base.dev);
	vkmsdev = xdev_to_vkmsdev(xdev);

	/* make sure that physical pages are allocated */
	xdx_bo_get_pages(bo);

	ret = xdx_bo_vmap(bo);
	if (ret) {
		dev_err(xdev->dev, "vmap failed: %d\n", ret);
		return ret;
	}

	vkms_state->dump_blob.data = bo->vaddr;
	vkms_state->dump_blob.size = bo->base.size;

	name = kasprintf(GFP_KERNEL, "%s-%dx%d-XRGB-%d.raw",
			 type_str[plane->type], state->fb->width,
			 state->fb->height, 8);
	if (!name) {
		dev_err(xdev->dev, "failed to kasprintf\n");
		return -ENOMEM;
	}

	vkms_state->dump_debugfs = debugfs_create_blob(
		name, 0444, vkmsdev->output.connector.debugfs_entry,
		&vkms_state->dump_blob);
	kfree(name);
#if (LINUX_VERSION_CODE < KERNEL_VERSION(4, 14, 0))
	return xdx_drm_gem_fb_prepare_fb(plane, state);
#elif (LINUX_VERSION_CODE < KERNEL_VERSION(5, 13, 0))
	return drm_gem_fb_prepare_fb(plane, state);
#else
        return drm_gem_plane_helper_prepare_fb(plane, state);
#endif

}

static void vkms_cleanup_fb(struct drm_plane *plane,
			    struct drm_plane_state *old_state)
{
	struct vkms_plane_state *vkms_state = to_vkms_plane_state(old_state);
	struct vkms_output *output = NULL;
	struct xdx_bo *bo;
	struct xdx_device *xdev;
	struct vkms_device *vkmsdev;

	if (!old_state || !old_state->crtc)
                return;

	output = drm_crtc_to_vkms_output(old_state->crtc);

	if (!old_state->fb || !output || !output->composer_enabled)
		return;
#if (LINUX_VERSION_CODE <= KERNEL_VERSION(4, 13, 0))
	bo = to_xdx_gem(xdx_fb_get_gem_obj(old_state->fb, 0));
#else
	bo = to_xdx_gem(drm_gem_fb_get_obj(old_state->fb, 0));
#endif
	xdev = drm_to_xdev(bo->base.dev);
	vkmsdev = xdev_to_vkmsdev(xdev);

	xdx_bo_vunmap(bo);

	vkms_state->dump_blob.data = NULL;
	vkms_state->dump_blob.size = 0;
	debugfs_remove(vkms_state->dump_debugfs);
	vkms_state->dump_debugfs = NULL;
}

static const struct drm_plane_helper_funcs vkms_primary_helper_funcs = {
	.atomic_update = vkms_plane_atomic_update,
	.atomic_check = vkms_plane_atomic_check,
	.prepare_fb = vkms_prepare_fb,
	.cleanup_fb = vkms_cleanup_fb,
};

struct drm_plane *xdx_vkms_plane_init(struct xdx_device *xdev,
				      enum drm_plane_type type, int index)
{
	//struct vkms_device *vkmsdev = xdev_to_vkmsdev(xdev);
	struct drm_device *dev = xdev_to_drm(xdev);
	const struct drm_plane_helper_funcs *funcs;
	struct drm_plane *plane;
	const u32 *formats;
	int ret, nformats;

	plane = kzalloc(sizeof(*plane), GFP_KERNEL);
	if (!plane)
		return ERR_PTR(-ENOMEM);

	if (type == DRM_PLANE_TYPE_CURSOR) {
		formats = vkms_cursor_formats;
		nformats = ARRAY_SIZE(vkms_cursor_formats);
		funcs = &vkms_primary_helper_funcs;
	} else {
		formats = vkms_formats;
		nformats = ARRAY_SIZE(vkms_formats);
		funcs = &vkms_primary_helper_funcs;
	}

#if (LINUX_VERSION_CODE >= KERNEL_VERSION(4, 10, 0))
	ret = drm_universal_plane_init(dev, plane, 1 << index,
					&vkms_plane_funcs, formats, nformats,
					NULL, type, NULL);
#else
	ret = drm_universal_plane_init(dev, plane, 1 << index,
					&vkms_plane_funcs, formats, nformats,
					type, NULL);
#endif
	if (ret) {
		kfree(plane);
		return ERR_PTR(ret);
	}

	drm_plane_helper_add(plane, funcs);

	return plane;
}
