/*
 * @Copyright   Copyright (c) XIANGDIXIAN Computing Technology Co., Ltd. All Rights Reserved
 * @License     Dual MIT/GPLv2
 */

#include <asm/io.h>
#include <linux/uaccess.h>
#include <linux/errno.h>
#include <linux/fs.h>
#include <linux/init.h>
#include <linux/ioport.h>
#include <linux/kernel.h>
#include <linux/list.h>
#include <linux/mm.h>
#include <linux/module.h>
#include <linux/sched.h>
#include <linux/slab.h>
#include <linux/vmalloc.h>
#include <linux/pci.h>

#include "hantro_fpga.h"

#ifndef HLINA_START_ADDRESS
#define HLINA_START_ADDRESS 0x02000000
#endif

#ifndef HLINA_SIZE
#define HLINA_SIZE 96
#endif

#ifndef HLINA_TRANSL_OFFSET
#define HLINA_TRANSL_OFFSET 0x0
#endif

/* the size of chunk in MEMALLOC_DYNAMIC */
#define CHUNK_SIZE (PAGE_SIZE * 4)
struct hantro_fpga_t g_fpga = {
	.dev = NULL,
	.alloc_size = HLINA_SIZE,
	.alloc_base = HLINA_START_ADDRESS,
	.addr_transl = 0,
};

int hantro_fpga_release(void)
{
	struct hantro_fpga_t *fpga = &g_fpga;
	unsigned  int i = 0;

	for (i = 0; i < fpga->chunks; i++) {
		spin_lock(&fpga->mem_lock);
		if (fpga->hlina_chunks[i].filp != 0) {
			XDX_DEV_INFO("memalloc: Found unfreed memory at release time!\n");

			fpga->hlina_chunks[i].filp = 0;
			fpga->hlina_chunks[i].chunks_reserved = 0;
		}
		spin_unlock(&fpga->mem_lock);
	}
	vfree(fpga->hlina_chunks);
	XDX_DEV_INFO("dev closed\n");
	return 0;
}

/* Cycle through the buffers we have, give the first free one */
int hantro_fpga_memalloc(struct hantro_mem_handle *phandle, unsigned int size)
{
	struct hantro_fpga_t *fpga = &g_fpga;
	unsigned int i = 0;
	unsigned int j = 0;
	unsigned int skip_chunks = 0;
	unsigned long busaddr;

	/* calculate how many chunks we need; round up to chunk boundary */
	unsigned int alloc_chunks = (size + CHUNK_SIZE - 1) / CHUNK_SIZE;

	busaddr = 0;

	/* run through the chunk table */
	for (i = 0; i < fpga->chunks;) {
		skip_chunks = 0;
		/* if this chunk is available */
		if (!fpga->hlina_chunks[i].chunks_reserved) {
			/* check that there is enough memory left */
			if (i + alloc_chunks > fpga->chunks)
				break;

			/* check that there is enough consecutive chunks */
			for (j = i; j < i + alloc_chunks; j++) {
				if (fpga->hlina_chunks[j].chunks_reserved) {
					skip_chunks = 1;
					/* skip the used chunks */
					i = j + fpga->hlina_chunks[j]
							.chunks_reserved;
					break;
				}
			}

			/* if enough free memory found */
			if (!skip_chunks) {
				busaddr = fpga->hlina_chunks[i].bus_address;
				fpga->hlina_chunks[i].filp = 0x55aa;
				fpga->hlina_chunks[i].chunks_reserved =
					alloc_chunks;
				break;
			}
		} else {
			/* skip the used chunks */
			i += fpga->hlina_chunks[i].chunks_reserved;
		}
	}

	if (busaddr == 0) {
		XDX_DEV_ERROR("memalloc: Allocation FAILED: size = %d\n", size);
		return -EFAULT;
	}
	XDX_DEV_DEBUG("MEMALLOC OK: size: %d, reserved: %ld\n", size,
		alloc_chunks * CHUNK_SIZE);

	phandle->sliceidx = 0; /* no slice in fpga now */
	phandle->size = alloc_chunks * (CHUNK_SIZE);
	phandle->paddr = busaddr +0x20000000;
	phandle->mem_base = fpga->ddr_base;

	/* FIXME: */
	phandle->vaddr = ioremap_wc(busaddr, alloc_chunks * CHUNK_SIZE);
	memset(phandle->vaddr,0,phandle->size);
	return 0;
}

/**
 * Free a buffer based on bus address
 */
int hantro_fpga_memfree(struct hantro_mem_handle *phandle)
{
	struct hantro_fpga_t *fpga = &g_fpga;
	unsigned int i = 0;

	for (i = 0; i < fpga->chunks; i++) {
		/* user space SW has stored the translated bus address, add
		 * addr_transl to translate back to our address space.
		 */
		if (fpga->hlina_chunks[i].bus_address == phandle->paddr) {
			fpga->hlina_chunks[i].filp = 0;
			fpga->hlina_chunks[i].chunks_reserved = 0;
			break;
		}
	}
	if (i == fpga->chunks) {
		XDX_DEV_ERROR("memalloc: Owner mismatch while freeing memory!\n");
		return -1;
	}

	return 0;
}

/**
 * Reset "used" status
 */
static int hantro_fpga_memreset(void)
{
	struct hantro_fpga_t *fpga = &g_fpga;
	unsigned int i = 0;
	int result;
	unsigned long ba = fpga->alloc_base;

	XDX_DEV_INFO("memalloc: Linear Memory Allocator\n");
	XDX_DEV_INFO("memalloc: Linear memory base = %p\n",
		(void *)fpga->alloc_base);

	fpga->chunks = (fpga->alloc_size * 1024 * 1024) / CHUNK_SIZE;
	XDX_DEV_INFO("memalloc: Total size %d MB; %d chunks of size %lu\n",
		fpga->alloc_size, (int)fpga->chunks, CHUNK_SIZE);

	fpga->hlina_chunks = vmalloc(fpga->chunks * sizeof(struct hlinc));

	for (i = 0; i < fpga->chunks; i++) {
		fpga->hlina_chunks[i].bus_address = ba;
		fpga->hlina_chunks[i].filp = 0;
		fpga->hlina_chunks[i].chunks_reserved = 0;

		ba += CHUNK_SIZE;
	}
	return 0;

	if (fpga->hlina_chunks)
		vfree(fpga->hlina_chunks);

	return result;
}

/**
 * init hantro fpga pcie board
 *
 * \return -1 fail;
 * \return  0 success;
 */
int hantro_fpga_init(unsigned long ddr_base, unsigned long offset)
{
	struct hantro_fpga_t *fpga = &g_fpga;
	int ret;

	fpga->ddr_base = ddr_base;
	fpga->alloc_base = fpga->ddr_base + offset;
	//fulongdbg reserve 2G for VPU
	fpga->alloc_size = 500 - offset / (0x100000);
	fpga->addr_transl = fpga->ddr_base;



	ret = hantro_fpga_memreset();
	if (ret != 0)
		goto out_pci_disable_device;

	return 0;

out_pci_disable_device:
	pci_disable_device(fpga->dev);

	return -1;
}

int pcie_init(struct xdx_device* xdev,struct hantro_pci_t *pci_par)
{
#if 0
	struct pci_dev *dev = NULL; /* PCI device structure. */

	unsigned long
		base_hdwr; /* PCI base register address (Hardware address) */
	unsigned long base_ddr_hw; /* PCI base register address (memalloc) */
	u32 base_len; /* Base register address Length */
	u32 ddr_len; /* Base register address Length */

	dev = pci_get_device(PCI_VENDOR_ID_HANTRO, PCI_DEVICE_ID_HANTRO_PCI,
			     dev);
	if (!dev) {
		XDX_DEV_ERROR("Init: Hardware not found.\n");
		goto out;
	}
	if (pci_enable_device(dev) < 0) {
		XDX_DEV_ERROR("Init: Device not enabled.\n");
		goto out;
	}
	base_hdwr = pci_resource_start(dev, PCI_H2_BAR);
	if (base_hdwr < 0) {
		XDX_DEV_ERROR("Init: Base Address not set.\n");
		goto out_pci_disable_device;
	}
	XDX_DEV_INFO("Base hw val 0x%llx\n", (unsigned long long)base_hdwr);

	base_len = pci_resource_len(dev, PCI_H2_BAR);
	XDX_DEV_INFO("Base hw len 0x%x\n", (unsigned int)base_len);

	base_ddr_hw = pci_resource_start(dev, PCI_DDR_BAR);
	if (base_ddr_hw == 0) {
		XDX_DEV_INFO("PcieInit: Base Address not set.\n");
		goto out_pci_disable_device;
	}
	XDX_DEV_INFO("Base memory val 0x%llx\n", (unsigned long long)base_ddr_hw);

	ddr_len = pci_resource_len(dev, PCI_DDR_BAR);
	XDX_DEV_INFO("Base memory len 0x%llx\n", (unsigned long long)ddr_len);

	pci_par->pci_base_reg_hw = base_hdwr;
	pci_par->pci_base_reg_len = base_len;
	pci_par->pci_base_ddr_hw = base_ddr_hw;
	pci_par->pci_base_ddr_len = ddr_len;


	return 0;

out_pci_disable_device:
	pci_disable_device(dev);
out:
	return -1;
#else
	pci_par->pci_base_reg_hw = xdev->rmmio_base;
	pci_par->pci_base_reg_len = xdev->rmmio_size;
	pci_par->pci_base_ddr_hw = xdev->aper_base;
	pci_par->pci_base_ddr_len = xdev->aper_size;

	return 0;
#endif

}

u64 vpu_alloc_gddr(struct xdx_device *xdev, struct xdx_bo **bo, unsigned long size)
{
	u64 paddr;
	void *vaddr;
	int ret;

	ret = xdx_bo_create_reserved(xdev, size, PAGE_SIZE, XDXGPU_PL_VRAM,
		bo, &paddr, (void **)&vaddr);
	if (ret) {
		dev_err(xdev->dev, "Failed to alloc memory\n");
		return 0;
	}

	paddr = paddr - PANGU_VRAM_DEV_PHYADDR_BASE + pci_resource_start(xdev->pdev, 2);

	return paddr;
}

